/* 
 * File:   Perft.cpp
 * Author: peter
 * 
 * Created on 01. Februar 2012, 12:15
 */

#include "Perft.h"
#include <cstdlib>
#include <time.h>
#include <pthread.h>
#include <iomanip>
#include "Position.h"
#include <sys/time.h>

#include <boost/ptr_container/ptr_list.hpp>

using namespace std;

typedef boost::ptr_list<string> PositionList;
typedef boost::ptr_list<string>::iterator PositionIterator;

Perft::Perft() {
}

void Perft::doPerftList(Position &pos, int max_depth, long *results) {
    result = new Perftres(max_depth);
    pos.getPerftInfo(1,max_depth,result);
    
    for (int i=0;i<max_depth;i++) {
        if (result->perftres[7*i]==results[i])
            cout << "OK  ";
        else
            cout << "ERR(" << result->perftres[7*i] << "," << results[i] << ")  ";
    }
    delete result;
}

void Perft::doPerft(Position &pos, int max_depth) {
    timeval start,end;
    gettimeofday(&start,0);

    result = new Perftres(max_depth);
    // doLevel1(pos,1,max_depth);
    pos.getPerftInfo(1,max_depth,result);
    long sum=result->print_info();
    cout << "================================================" << endl << "total: " << sum << " nodes" << endl;
    
    gettimeofday(&end,0);
   
    double seconds = (end.tv_sec-start.tv_sec) + (end.tv_usec-start.tv_usec)*0.000001;
    
    cout.setf(ios::fixed, ios::floatfield);
    cout << "time: " << setprecision(3) << seconds << " seconds, nps: " << setprecision(0) << double(sum/seconds) << endl << endl;
    delete result;
}

void Perft::doSpeedPerft(Position &pos, int max_depth) {
    timeval start,end;
    gettimeofday(&start,0);    
    
    long sum=0;
    pos.getChildsSpeed(0,1,max_depth,sum);
    
    gettimeofday(&end,0);
    double seconds = (end.tv_sec-start.tv_sec) + (end.tv_usec-start.tv_usec)*0.000001;
    
    cout.setf(ios::fixed, ios::floatfield);
    cout << endl << "perft result (" << max_depth << "):" << endl;
    cout << "nodes: " << sum << endl;
    cout << "time: " << setprecision(3) << seconds << " seconds, nps: " << setprecision(0) << double(sum/seconds) << endl << endl;    
}




// rekursion fuer pseudo legal moves
void Perft::getChildsPseudo(Position &pos, int depth, long &sum) {
    pos.cd=depth;
    pos.calc_moves(true);
    if (depth>0) {
        pos.doBackup(depth);
        for (int i=0;i<pos.mvSize[depth];i++) {
            pos.move_figure(pos.mvD[depth][i][0],pos.mvD[depth][i][1],pos.mvI[depth][i]);
            getChildsPseudo(pos, depth-1, sum);
            pos.undoMove(depth);
        }
    } else
        sum+=pos.mvSize[depth];      
}

// pseudo legal moves
void Perft::doPseudoPerft(Position &pos, int max_depth) {
    timeval start,end;
    gettimeofday(&start,0);    
    
    long sum=0;
    // pos.getChildsPseudo(max_depth-1,sum);
    getChildsPseudo(pos,max_depth-1,sum);
    
    gettimeofday(&end,0);
    double seconds = (end.tv_sec-start.tv_sec) + (end.tv_usec-start.tv_usec)*0.000001;
    
    cout.setf(ios::fixed, ios::floatfield);
    cout << endl << "perft result (" << max_depth << "):" << endl;
    cout << "nodes: " << sum << endl;
    cout << "time: " << setprecision(3) << seconds << " seconds, nps: " << setprecision(0) << double(sum/seconds) << endl << endl;    
}



Perft::~Perft() {
}

/*
int Perft::doLevel1(Position &pos, int curr_depth, int max_depth) {
    childlist = pos.getMMChilds();
    
    if (childlist.size()>0) { // new nodes found
        result->perftres[(curr_depth-1)*7]+=childlist.size(); // update found node-count

        if (curr_depth<max_depth) {
            
            int numThreads=1;
            
            Perftres perftsi[numThreads];
            for (int i=0;i<numThreads;i++)
                perftsi[i].setSize(max_depth);
            
            pthread_t threads[numThreads];
            int rc;
            
            threadData myThreadData[numThreads];

            
            
            for (int i=0;i<numThreads;i++) {
                int from=i*childlist.size()/numThreads;
                int to=(i+1)*childlist.size()/numThreads;
                
                myThreadData[i].pos=Position(pos);
                myThreadData[i].curr_depth=1;
                myThreadData[i].max_depth=max_depth;
                myThreadData[i].childlist=childlist;
                myThreadData[i].result= &(perftsi[i]);
                myThreadData[i].fromIndex=from;
                myThreadData[i].toIndex=to;
                
                rc = pthread_create( &threads[i], NULL, doThread, (void *) &myThreadData[i]);
                if( rc != 0 ) {
                    cout << "Konnte Thread " << i << " nicht erzeugen" << endl;
                    return EXIT_FAILURE;
                }
            }

            for (int i=0;i<numThreads;i++)
                pthread_join(threads[i], NULL);
          
            for (int i=0;i<numThreads;i++) {
                for (int k=0;k<result->getSize();k++) {
                    result->perftres[k]+=perftsi[i].perftres[k];
                }
            }
        }
        
    } else { // if childlist is empty, we've found a mate position of the previous level
        cout << "There's no move. Color is checkmate.";
    }
    return 0;
}

void *Perft::doThread(void *threadarg) {
    struct threadData *myData;
    myData = (struct threadData *) threadarg;

    int curr=myData->curr_depth;
    int max=myData->max_depth;
    Perftres *result=myData->result;
    
    int index=0;
    for (PositionIterator i=myData->childlist.begin();i!=myData->childlist.end();++i,++index) {
        cout << ".";
        if (index>=myData->fromIndex && index<myData->toIndex) {
            i->getPerftInfo(1,(curr+1),max,result);
            
        }
    }
    return NULL;
}
 * */