/* 
 * File:   Position.h
 * Author: peter
 *
 * Created on 31. Jänner 2012, 18:23
 */

#ifndef POSITION_H
#define	POSITION_H

#include <string>
#include <list>
#include <iostream>
#include <stdio.h>
#include <sstream>
#include <cstdlib>
#include <boost/ptr_container/ptr_list.hpp>
#include "Perftres.h"


using namespace std;

struct backup {
    unsigned long aw;
    unsigned long ab;
    unsigned long all;
    unsigned long kings[2];
    unsigned long pawns[2];
    unsigned long bishops[2];
    unsigned long knights[2];
    unsigned long queens[2];
    unsigned long rooks[2];    
    unsigned long targetmap;    
    short int ep;
    bool white_castling_short;
    bool white_castling_long;
    bool black_castling_short;
    bool black_castling_long;  
    bool onMove;
};

class Position {
public:
    Position();
    Position(const Position &pos);
    Position(const string &fen);
    string getFen();
    
    boost::ptr_list<string> getMMChilds();
    void getChildsSpeed(int color, int curr_depth, int depth, long &sum);
    void getPerftInfo(int curr_depth, int depth, Perftres *perft);
    void getDivide(int color, int curr_depth, const int &max_depth, const bool &modus, long &sum, long &total);

    int alphaBetaMax(bool side2move, int alpha, int beta, int curr_depth, int depth, int &sum, string &move);
    int evaluate(bool side2move);
    
    void sortMoves(unsigned short int sortedIndizes[], int cd);
    
    int negaMax(int curr_depth, int depth);

    void egdb();
    void rekEgdb(int depth);
    
    void printBoard();

    unsigned long aw;
    unsigned long ab;
    unsigned long all;
    unsigned long kings[2];
    unsigned long pawns[2];
    unsigned long bishops[2];
    unsigned long knights[2];
    unsigned long queens[2];
    unsigned long rooks[2];    
    unsigned long targetmap;
    
    short int ep;
    bool white_castling_short;
    bool white_castling_long;
    bool black_castling_short;
    bool black_castling_long;      
    string last_move;
    
    int cd;
    backup myBackup[15];
    unsigned long mvD[15][256][2];
    short int mvI[15][256];
    unsigned short int mvSize[15];
    // unsigned short int sortedIndizes[256];    
    
    bool onMove;
    
    
    bool move_figure (const unsigned long &from, const unsigned long &to, const int &promotion);
    void getInfo();
    
    static string intToString(int value) {
        string s;
        stringstream out;
        out << value;
        s = out.str();
        return s;
    }  
    
    static const int index64[64];
    static const unsigned int RUN_FROM[];
    static const unsigned int RUN_TO[];    
    
    static const unsigned long TARGET_Nx[];
    static const unsigned int TARGET_Nx_FROM[];
    static const unsigned int TARGET_Nx_TO[];

    static const unsigned long TARGET_Ex[];
    static const unsigned int TARGET_Ex_FROM[];
    static const unsigned int TARGET_Ex_TO[];    

    static const unsigned long TARGET_Sx[];
    static const unsigned int TARGET_Sx_FROM[];
    static const unsigned int TARGET_Sx_TO[];        

    static const unsigned long TARGET_Wx[];
    static const unsigned int TARGET_Wx_FROM[];
    static const unsigned int TARGET_Wx_TO[];        

    static const unsigned long TARGET_NEx[];
    static const unsigned int TARGET_NEx_FROM[];
    static const unsigned int TARGET_NEx_TO[];       

    static const unsigned long TARGET_SEx[];
    static const unsigned int TARGET_SEx_FROM[];
    static const unsigned int TARGET_SEx_TO[];       

    static const unsigned long TARGET_SWx[];
    static const unsigned int TARGET_SWx_FROM[];
    static const unsigned int TARGET_SWx_TO[];       

    static const unsigned long TARGET_NWx[];
    static const unsigned int TARGET_NWx_FROM[];
    static const unsigned int TARGET_NWx_TO[];       
    
    // static const unsigned long OLD_KING_POS[];
    // static const unsigned long NEW_KING_POS[];
    // static const unsigned long OLD_ROOK_POS[];
    // static const unsigned long NEW_ROOK_POS[];
    
    static const unsigned long BIN[];
    
    static const unsigned long K_TARGET2x[];
    static const unsigned int K_TARGET2x_FROM[];
    static const unsigned int K_TARGET2x_TO[];
    
    static const unsigned long KN_TARGET3x[];
    static const unsigned int KN_TARGET3x_FROM[];
    static const unsigned int KN_TARGET3x_TO[];
    
    
    
    static const unsigned int KN_RUN_FROM[];
    static const unsigned int KN_RUN_TO[];

    static const unsigned long TARGET_BLACK_V1[];
    static const unsigned long TARGET_BLACK_V2[];
    static const unsigned long TARGET_BLACK_V3[];
    static const unsigned long TARGET_BLACK_V4[];
    static const unsigned long TARGET_WHITE_V1[];
    static const unsigned long TARGET_WHITE_V2[];
    static const unsigned long TARGET_WHITE_V3[];
    static const unsigned long TARGET_WHITE_V4[];    

    void calc_moves(const bool &check);
    void doBackup(int depth);
    void undoMove(int depth);
    bool figOnPos(long figure, int position);
    
private:

    Position new_ep_child(const int &from, const int &to, const int &kill);
    inline void promote(const unsigned long &pos, const int &promotype, const int &color);
    inline void rochade(const unsigned long &old_king_pos, const unsigned long &new_king_pos, const unsigned long &old_rook_pos, const unsigned long &new_rook_pos);
    inline void updateBitboardLongs();    
    

    inline void split_moves_direct(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &myFigures);
    inline void split_moves__(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op);
    inline void addPawnSingles_black(const int &z);
    inline void addPawnSingles_white(const int &z);
    inline int bitScanForward(unsigned long &bb);
    inline void addKingMoves(const int &color);
    inline bool opponent_target_bitboard2(const unsigned long &bitboard);

    int popCount(unsigned long x);
    string printBoardRow(int row);

    string getMoveString(int i);
    string getBoolString(bool value);
    
    string get_hum(int k);



};

#endif	/* POSITION_H */