/* 
 * File:   egdb.cpp
 * Author: peter
 * 
 * Created on 28. Februar 2012, 13:10
 */

#include <mysql/mysql.h>
#include <stdio.h>
#include <iostream>
#include <string>
#include "Egdb.h"
#include "Position.h"
#include "globals.h"

using namespace std;

Egdb::Egdb() {

   /* Handle initialisieren */
   my = mysql_init(NULL);
   if(my == NULL) {
      fprintf(stderr, " Initialisierung fehlgeschlagen\n");
      return;
   }

   /* mit dem Server verbinden */
   if( mysql_real_connect (
        my,   /* Zeiger auf MYSQL-Handler */
        "127.0.0.1", /* Host-Name */
        "root", /* User-Name */
        "", /* Passwort für user_name */
           "",0,NULL,0
           ) == NULL) {
      fprintf (stderr, "Fehler mysql_real_connect():"
        "%u (%s)\n",mysql_errno (my), mysql_error (my));
   }
   else {
      printf("Erfolgreich mit dem MySQL-Server verbunden\n");
      mysql_query(my,"DROP DATABASE chess");
      mysql_query(my,"CREATE database chess");
      mysql_select_db(my,"chess");
   }

   /* Hier befindet sich der Code für die Arbeit mit MySQL. */

   /* Verbindung trennen */

}

Egdb::~Egdb() {
    mysql_close (my);
}

void Egdb::build() {
    cout << "start creating egdb" << endl;

//    buildTable("Kkb");
//    buildTable("Kkn");
//    buildTable("Kkr");
//    buildTable("Kkq");
    buildTable("Kkp");
    
    cout << "egdb build successful" << endl;
    
}

void Egdb::buildTable(string figs) {
 
    figures=figs;
    sumFigures=figures.length();
    
    
    string query = "CREATE TABLE IF NOT EXISTS `";
    query+=figures;
    query+="` (`id` int(11) NOT NULL AUTO_INCREMENT,`fen` varchar(110) COLLATE utf8_bin NOT NULL,`black` TINYINT NOT NULL ,`white` TINYINT NOT NULL,PRIMARY KEY (`id`)) ENGINE=MYISAM DEFAULT CHARSET=utf8 COLLATE=utf8_bin AUTO_INCREMENT=1";
    mysql_query(my, query.c_str());	     
    query.clear();
    
    query="TRUNCATE ";
    query+=figures;
    mysql_query(my, query.c_str());
    

    query="CREATE TABLE IF NOT EXISTS `"+figures+"_positions` (`id` int(11) NOT NULL AUTO_INCREMENT,`fen` varchar(110) COLLATE utf8_bin NOT NULL, PRIMARY KEY(`id`)) ENGINE=MYISAM DEFAULT CHARSET=utf8 COLLATE=utf8_bin AUTO_INCREMENT=1";
    mysql_query(my, query.c_str());
    query="TRUNCATE "+figures+"_positions";
    // ALTER TABLE `chess`.`Kkp_positions` ADD INDEX `fen` ( `fen` ) 
    mysql_query(my, query.c_str());
    

    query="CREATE TABLE IF NOT EXISTS `"+figures+"_movesblack` (`fen1` varchar(110) COLLATE utf8_bin NOT NULL,`fen2` varchar(110) COLLATE utf8_bin NOT NULL ) ENGINE=MYISAM DEFAULT CHARSET=utf8 COLLATE=utf8_bin";
    mysql_query(my, query.c_str());
    query="TRUNCATE "+figures+"_movesblack";
    mysql_query(my, query.c_str());    
    
    
    
    sumPositions=0;
    rekursive(sumFigures-1);
    cout << "Positions (" << figures << "): " << sumPositions << endl;    
        
    
}

void Egdb::rekursive(int depth) {
    for (int i=1;i<=64;i++) {
        egdbDepth[depth]=i;
        if (depth>0)
            rekursive(depth-1);
        else {
            bool check=false;
            for (int k=0;k<sumFigures;k++) {
                for (int m=k+1;m<sumFigures;m++) {
                    if (egdbDepth[k]==egdbDepth[m]) {
                        check=!check;
                        break;
                    }
                }
                if (check)
                    break;
            }
            if (!check)
                buildFen();
        }
    }
}

bool Egdb::isPositionLegal(string fen) {
    Position pos;
    
    pos = Position(fen+" b - -");
    if (pos.inCheck()) // König nicht bedroht
        return false;
    if ((pos.pawns[pos.onMove]&255)!=0L) // Bauer nicht hinter Startlinie
        return false;
    if ((pos.pawns[pos.onMove]&18374686479671623680)!=0L) // Bauer nicht in vordester Zeile
        return false;
    
    pos = Position(fen+" w - -");
    if (pos.inCheck()) // König nicht bedroht
        return false;
    if ((pos.pawns[pos.onMove]&18374686479671623680)!=0L) // Bauer nicht hinter Startlinie
        return false;
    if ((pos.pawns[pos.onMove]&255)!=0L) // Bauer nicht in vordester Zeile
        return false;
    
    return true;
}

bool Egdb::isBlackMate(string fen) {
    Position pos = Position(fen+" b - -");
    return pos.isCheckMate();
}

bool Egdb::isWhiteMate(string fen) {
    Position pos = Position(fen+" w - -");
    return pos.isCheckMate();
}

void Egdb::buildFen() {
    sumPositions++;
    
    string fen="";
    
    int spaces=0;
    for (int i=1;i<=64;i++) {
        
        bool found=false;
        for (int k=0;k<sumFigures;k++) {
            if (egdbDepth[k]==i) {
                found=true;
                
                if (spaces>0) {
                    fen+=int2String(spaces);
                    spaces=0;
                }                
                
                fen.append(figures.substr(k,1));
                break;
            }
        }
        if(!found)
            spaces++;

        if (i%8==0 && i<64) {
            if (spaces>0) {
                fen+=int2String(spaces);
                spaces=0;
            }
            fen+="/";
        }
        if (i==64 && spaces>0) {
                fen+=int2String(spaces);
        }        
        
    }
    
    // cout << figures << ": " << egdbDepth[2] << " / " << egdbDepth[1] << " / " << egdbDepth[0] << " --> " << fen << endl;

    if (!isPositionLegal(fen))
        return;
    
    string query;
    query="INSERT INTO `";
    query+=figures;
    query+="` (`fen`,`black`,`white`) VALUES ('";
    query+=fen;
    
    int blackMate=isBlackMate(fen);
    int whiteMate=isWhiteMate(fen);

    query+="','"+int2String(blackMate)+"','"+int2String(whiteMate)+"')";

    mysql_query(my, query.c_str());
    
    query="INSERT INTO `"+figures+"_positions` (`fen`) VALUES ('"+fen+"')";
    mysql_query(my, query.c_str());

    
    
    Position pos = Position(fen+" b - -");

    
    
    list<string> positions;
    
    positions = pos.getMovePositions();
    
    for (list<string>::iterator it=positions.begin();it!=positions.end();it++) {

        query="INSERT INTO `"+figures+"_movesblack` (`fen1`,`fen2`) VALUES ('"+fen+"','"+(*it)+"')";
        mysql_query(my, query.c_str());        
        
    }
    

    
}