/* 
 * File:   Position.h
 * Author: peter
 *
 * Created on 31. Jänner 2012, 18:23
 */

#ifndef POSITION_H
#define	POSITION_H

#include <string>
#include <list>
#include <iostream>
#include <stdio.h>
#include <cstdlib>
#include "globals.h"

struct backup {
    unsigned long aw;
    unsigned long ab;
    unsigned long all;
    unsigned long kings[2];
    unsigned long pawns[2];
    unsigned long bishops[2];
    unsigned long knights[2];
    unsigned long queens[2];
    unsigned long rooks[2];    
    unsigned long targetmap;    
    short int ep;
    char castlingRules;
    bool onMove;
};

using namespace std;

class Position {
public:
    Position();
    Position(const string &fen);
    string getFen();
    string getFenShort();
    void testHash(int curr_depth, int depth, long &sum);
    void getChildsPseudo(int depth, long &sum);
    int getMMChilds(string moveArray[], int moveIndex[]);
    void getPerftInfo(int curr_depth, int depth, Perftres *perft);
    void getDivide(int color, int curr_depth, const int &max_depth, const bool &modus, long &sum, long &total);
    int negaMax(int alpha, int beta, int curr_depth, int depth, bool isPv); // , bool canNull);
    void printBoard();
    string getBoardString();
    void calc_moves();
    void calcCaptures();
    void doBackup();
    void undoMove(const unsigned long &from, const unsigned long &to, const int &promotion, const unsigned short int &capture);
    unsigned short int move_figure (const unsigned long &from, const unsigned long &to, const int &promotion);
    bool isEndgame();
    void getInfo();    
    string getMoveString(int i);
    string getMoveStringAlgShort(int i);
    
    unsigned long aw;
    unsigned long ab;
    unsigned long all;
    unsigned long kings[2];
    unsigned long pawns[2];
    unsigned long bishops[2];
    unsigned long knights[2];
    unsigned long queens[2];
    unsigned long rooks[2];    
    short int ep;
    bool onMove;
    char castlingRules;
    unsigned long hashKey;

    // Variablen für Suchbaum
    int cd;
    backup myBackup[128];
    unsigned long mvD[128][256][2]; // 256*2*8=4096 bytes = 4 kilobytes => bei maxdepth=64: 256 kilobytes = 0,25 megabyte
    unsigned short int mvSize[128];
    
    // Index vom besten Zug (negaMax), sofern zuvor berechnet
    int bestMoveIndex;
    
    
    unsigned int sum;
    
    
    // statische bitboards
    static const int index64[64];
    static const unsigned short int castlingR1[];
    static const unsigned long promotionRow[];
    static const unsigned long TARGET_Nx[];
    static const unsigned int TARGET_Nx_FROM[];
    static const unsigned int TARGET_Nx_TO[];
    static const unsigned long TARGET_Ex[];
    static const unsigned int TARGET_Ex_FROM[];
    static const unsigned int TARGET_Ex_TO[];    
    static const unsigned long TARGET_Sx[];
    static const unsigned int TARGET_Sx_FROM[];
    static const unsigned int TARGET_Sx_TO[];        
    static const unsigned long TARGET_Wx[];
    static const unsigned int TARGET_Wx_FROM[];
    static const unsigned int TARGET_Wx_TO[];        
    static const unsigned long TARGET_NEx[];
    static const unsigned int TARGET_NEx_FROM[];
    static const unsigned int TARGET_NEx_TO[];       
    static const unsigned long TARGET_SEx[];
    static const unsigned int TARGET_SEx_FROM[];
    static const unsigned int TARGET_SEx_TO[];       
    static const unsigned long TARGET_SWx[];
    static const unsigned int TARGET_SWx_FROM[];
    static const unsigned int TARGET_SWx_TO[];       
    static const unsigned long TARGET_NWx[];
    static const unsigned int TARGET_NWx_FROM[];
    static const unsigned int TARGET_NWx_TO[];           
    static const unsigned long BIN[];
    static const unsigned long K_TARGET2x[];
    static const unsigned int K_TARGET2x_FROM[];
    static const unsigned int K_TARGET2x_TO[];
    static const unsigned long KN_TARGET3x[];
    static const unsigned int KN_TARGET3x_FROM[];
    static const unsigned int KN_TARGET3x_TO[];
    static const unsigned int KN_RUN_FROM[];
    static const unsigned int KN_RUN_TO[];
    static const unsigned long TARGET_BLACK_V1[];
    static const unsigned long TARGET_BLACK_V2[];
    static const unsigned long TARGET_BLACK_V3[];
    static const unsigned long TARGET_BLACK_V4[];
    static const unsigned long TARGET_WHITE_V1[];
    static const unsigned long TARGET_WHITE_V2[];
    static const unsigned long TARGET_WHITE_V3[];
    static const unsigned long TARGET_WHITE_V4[];
    static const unsigned long PAWN_ATTACKS[2][64];
    
    bool inCheck() {
        return opponent_target_bitboard2(kings[onMove]);
    }
    
    bool isCheckMate();
    
    list<string> getMovePositions();
    
private:
    /*
     * Start: Einige Konstanten der Eval-Methode
     */
    
    
    static const int knightAdj[9];
    static const int rookAdj[9];
    
    static const int diagNW[64];
    
    static const int diagNE[64];
    
    static const int bonusDiaDistance[14];

    static const int multiPawnPenalty[7];
    
    static const unsigned long cols[8];

    static const unsigned long d4e4bb;
    static const unsigned long d4c4bb;
    static const unsigned long d4e3bb;
    static const unsigned long e4f4bb;
    static const unsigned long e4d3bb;
    
    static const unsigned long d5e5bb;
    static const unsigned long d5c5bb;
    static const unsigned long d5e6bb;
    static const unsigned long e5f5bb;
    static const unsigned long e5d6bb;
    
    static const unsigned short int d4e4bonus; // d5e5
    static const unsigned short int d4c4bonus; // d5c5
    static const unsigned short int d4e3bonus; // d5e6
    static const unsigned short int e4f4bonus; // e5f5
    static const unsigned short int e4d3bonus; // e5d6

    static const int passedPawnsBonus[2][64];
    
    static const int passedPawnsProtectedBonus[2][64];
    
    static const unsigned long passedPawnsBitboards[2][64];

    static const unsigned long passedPawnsBitboardsAttacks[2][64];

    static const unsigned long passedPawnsProtectedBitboards[2][64];
    
    static const int weakPawnsPenalty[2][64];
    
    static const unsigned long weakPawnsBitboards[2][64];

    static const int knightA8Penalty;
    static const int knightH8Penalty;
    static const int knightA7Penalty;
    static const int knightH7Penalty;
    static const int knightC3Penalty;

    static const int bishopA7Penalty;
    static const int bishopH7Penalty;
    static const int bishopB8Penalty;
    static const int bishopG8Penalty;
    static const int bishopA6Penalty;
    static const int bishopH6Penalty;
    static const int bishopF1Bonus;
    static const int bishopC1Bonus;

    static const int rookOpenBonus;
    static const int rookHalfOpenBonus;

    static const int shield1Bonus;
    static const int shield2Bonus;
    static const int noShieldPenalty;

    static const int centerPawnPenalty;
    
    static const int kingBlockRookPenalty;    
    
    static const int slaveMistakePenalty;
    
    static const int fianchettoBonus;
    static const int fianchettoPenalty;

    static const int pawnPieceSquare[2][64];

    static const int knightPieceSquare[2][64];

    static const int bishopPieceSquare[2][64];
    
    static const int rookPieceSquare[2][64];
    
    static const int queenPieceSquare[2][64];
    
    static const int kingPieceSquare[2][64];
    
    static const int kingPieceSquareEndgame[2][64];
    
/*
    static const int knightAdj[9] = { -20, -16, -12, -8, -4,  0,  4,  8, 12};
    static const int rookAdj[9] =   {  15,  12,   9,  6,  3,  0, -3, -6, -9};
    
    static const int diagNW[64] = {
        0,  1,  2,  3,  4,  5,  6,  7,
        1,  2,  3,  4,  5,  6,  7,  8,
        2,  3,  4,  5,  6,  7,  8,  9,
        3,  4,  5,  6,  7,  8,  9, 10,
        4,  5,  6,  7,  8,  9, 10, 11,
        5,  6,  7,  8,  9, 10, 11, 12,
        6,  7,  8,  9, 10, 11, 12, 13,
        7,  8,  9, 10, 11, 12, 13, 14
    };    
    
    static const int diagNE[64] = {
         7,  6,  5,  4,  3,  2,  1,  0,
         8,  7,  6,  5,  4,  3,  2,  1,
         9,  8,  7,  6,  5,  4,  3,  2,
        10,  9,  8,  7,  6,  5,  4,  3,
        11, 10,  9,  8,  7,  6,  5,  4,
        12, 11, 10,  9,  8,  7,  6,  5,
        13, 12, 11, 10,  9,  8,  7,  6,
        14, 13, 12, 11, 10,  9,  8,  7
    };       
    
    static const int bonusDiaDistance[14] = {3, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

    static const int multiPawnPenalty[7] = {0, 0, 25, 60, 100, 100, 100};
    
    static const unsigned long cols[8] = {
        72340172838076673,
        144680345676153346,
        289360691352306692,
        578721382704613384,
        1157442765409226768,
        2314885530818453536,
        4629771061636907072,
        9259542123273814144UL
    };

    static const unsigned long d4e4bb=103079215104;
    static const unsigned long d4c4bb=51539607552;
    static const unsigned long d4e3bb=17626545782784;
    static const unsigned long e4f4bb=206158430208;
    static const unsigned long e4d3bb=8864812498944;
    
    static const unsigned long d5e5bb=402653184;
    static const unsigned long d5c5bb=201326592;
    static const unsigned long d5e6bb=135266304;
    static const unsigned long e5f5bb=805306368;
    static const unsigned long e5d6bb=268959744;
    
    static const unsigned short int d4e4bonus=10; // d5e5
    static const unsigned short int d4c4bonus=5; // d5c5
    static const unsigned short int d4e3bonus=5; // d5e6
    static const unsigned short int e4f4bonus=5; // e5f5
    static const unsigned short int e4d3bonus=5; // e5d6

    static const int passedPawnsBonus[2][64] = {{
          0,   0,   0,   0,   0,   0,   0,   0,
        100, 100, 100, 100, 100, 100, 100, 100,
         80,  80,  80,  80,  80,  80,  80,  80,
         60,  60,  60,  60,  60,  60,  60,  60,
         40,  40,  40,  40,  40,  40,  40,  40,
         20,  20,  20,  20,  20,  20,  20,  20,
         20,  20,  20,  20,  20,  20,  20,  20,
          0,   0,   0,   0,   0,   0,   0,   0
    },{
          0,   0,   0,   0,   0,   0,   0,   0,
         20,  20,  20,  20,  20,  20,  20,  20,
         20,  20,  20,  20,  20,  20,  20,  20,                  
         40,  40,  40,  40,  40,  40,  40,  40,
         60,  60,  60,  60,  60,  60,  60,  60,
         80,  80,  80,  80,  80,  80,  80,  80,
        100, 100, 100, 100, 100, 100, 100, 100,                  
          0,   0,   0,   0,   0,   0,   0,   0                  
    }};
    
    static const int passedPawnsProtectedBonus[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        125, 125, 125, 125, 125, 125, 125, 125,
        100, 100, 100, 100, 100, 100, 100, 100,
        75, 75, 75, 75, 75, 75, 75, 75,
        50, 50, 50, 50, 50, 50, 50, 50,
        25, 25, 25, 25, 25, 25, 25, 25,
        25, 25, 25, 25, 25, 25, 25, 25,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,
        25, 25, 25, 25, 25, 25, 25, 25,
        25, 25, 25, 25, 25, 25, 25, 25,
        50, 50, 50, 50, 50, 50, 50, 50,                
        75, 75, 75, 75, 75, 75, 75, 75,
        100, 100, 100, 100, 100, 100, 100, 100,
        125, 125, 125, 125, 125, 125, 125, 125,
        0, 0, 0, 0, 0, 0, 0, 0                
    }};
    
    static const unsigned long passedPawnsBitboards[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 4, 8, 16, 32, 64, 128,
        257, 514, 1028, 2056, 4112, 8224, 16448, 32896,                
        65793, 131586, 263172, 526344, 1052688, 2105376, 4210752, 8421504, 
        16843009, 33686018, 67372036, 134744072, 269488144, 538976288, 1077952576, 2155905152, 
        4311810305, 8623620610, 17247241220, 34494482440, 68988964880, 137977929760, 275955859520, 551911719040, 
        1103823438081, 2207646876162, 4415293752324, 8830587504648, 17661175009296, 35322350018592, 70644700037184, 141289400074368,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,                
        72340172838076416UL, 144680345676152832UL, 289360691352305664UL, 578721382704611328UL, 1157442765409222656UL, 2314885530818445312UL, 4629771061636890624UL, 9259542123273781248UL,
        72340172838010880UL, 144680345676021760UL, 289360691352043520UL, 578721382704087040UL, 1157442765408174080UL, 2314885530816348160UL, 4629771061632696320UL, 9259542123265392640UL,
        72340172821233664UL, 144680345642467328UL, 289360691284934656UL, 578721382569869312UL, 1157442765139738624UL, 2314885530279477248UL, 4629771060558954496UL, 9259542121117908992UL,
        72340168526266368UL, 144680337052532736UL, 289360674105065472UL, 578721348210130944UL, 1157442696420261888UL, 2314885392840523776UL, 4629770785681047552UL, 9259541571362095104UL,
        72339069014638592UL, 144678138029277184UL, 289356276058554368UL, 578712552117108736UL, 1157425104234217472UL, 2314850208468434944UL, 4629700416936869888UL, 9259400833873739776UL,
        72057594037927936UL, 144115188075855872UL, 288230376151711744UL, 576460752303423488UL, 1152921504606846976UL, 2305843009213693952UL, 4611686018427387904UL, 9223372036854775808UL,
        0, 0, 0, 0, 0, 0, 0, 0
    }};

    static const unsigned long passedPawnsBitboardsAttacks[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0,
        514,1285, 2570, 5140, 10280, 20560, 41120, 16448,
        131586, 328965, 657930, 1315860, 2631720, 5263440, 10526880, 4210752,
        33686018, 84215045, 168430090, 336860180, 673720360, 1347440720, 2694881440, 1077952576,
        8623620610, 21559051525, 43118103050, 86236206100, 172472412200, 344944824400, 689889648800, 275955859520,
        2207646876162, 5519117190405, 11038234380810, 22076468761620, 44152937523240, 88305875046480, 176611750092960, 70644700037184,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,
        144680345676152832UL, 361700864190382080UL, 723401728380764160UL, 1446803456761528320UL, 2893606913523056640UL, 5787213827046113280UL, 11574427654092226560UL, 4629771061636890624UL,
        144680345676021760UL, 361700864190054400UL, 723401728380108800UL, 1446803456760217600UL, 2893606913520435200UL, 5787213827040870400UL, 11574427654081740800UL, 4629771061632696320UL,
        144680345642467328UL, 361700864106168320UL, 723401728212336640UL, 1446803456424673280UL, 2893606912849346560UL, 5787213825698693120UL, 11574427651397386240UL, 4629771060558954496UL,
        144680337052532736UL, 361700842631331840UL, 723401685262663680UL, 1446803370525327360UL, 2893606741050654720UL, 5787213482101309440UL, 11574426964202618880UL, 4629770785681047552UL,
        144678138029277184UL, 361695345073192960UL, 723390690146385920UL, 1446781380292771840UL, 2893562760585543680UL, 5787125521171087360UL, 11574251042342174720UL, 4629700416936869888UL,
        0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0
    }};

    static const unsigned long passedPawnsProtectedBitboards[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        131584, 328960, 657920, 1315840, 2631680, 5263360, 10526720, 4210688,
        33685504, 84213760, 168427520, 336855040, 673710080, 1347420160, 2694840320, 1077936128,
        8623489024, 21558722560, 43117445120, 86234890240, 172469780480, 344939560960, 689879121920, 275951648768,
        2207613190144, 5519032975360, 11038065950720, 22076131901440, 44152263802880, 88304527605760, 176609055211520, 70643622084608,
        565148976676864, 1412872441692160, 2825744883384320, 5651489766768640, 11302979533537280, 22605959067074560, 45211918134149120, 18084767253659648,
        562949953421312, 1407374883553280, 2814749767106560, 5629499534213120, 11258999068426240, 22517998136852480, 45035996273704960, 18014398509481984,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,
        512, 1280, 2560, 5120, 10240, 20480, 40960, 16384,
        131584, 328960, 657920, 1315840, 2631680, 5263360, 10526720, 4210688,
        33685504, 84213760, 168427520, 336855040, 673710080, 1347420160, 2694840320, 1077936128,
        8623489024, 21558722560, 43117445120, 86234890240, 172469780480, 344939560960, 689879121920, 275951648768,
        2207613190144, 5519032975360, 11038065950720, 22076131901440, 44152263802880, 88304527605760, 176609055211520, 70643622084608,
        565148976676864, 1412872441692160, 2825744883384320, 5651489766768640, 11302979533537280, 22605959067074560, 45211918134149120, 18084767253659648,
        0, 0, 0, 0, 0, 0, 0, 0
    }};    
    
    static const int weakPawnsPenalty[2][64] = {{
          0,    0,   0,   0,    0,   0,   0,   0,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
          -8, -12,  -14, -16, -16, -14, -12, -10,
          -8, -12,  -14, -16, -16, -14, -12, -10,
           0,   0,   0,    0,   0,   0,   0,   0
    },{
           0,   0,   0,    0,   0,   0,   0,   0,
          -8, -12,  -14, -16, -16, -14, -12, -10,
          -8, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
          0,    0,   0,   0,    0,   0,   0,   0,                   
    }};
    
    static const unsigned long weakPawnsBitboards[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        144680345676152832UL, 361700864190382080UL, 723401728380764160UL, 1446803456761528320UL, 2893606913523056640UL, 5787213827046113280UL, 11574427654092226560UL, 4629771061636890624UL,
        144680345676021760UL, 361700864190054400UL, 723401728380108800UL, 1446803456760217600UL, 2893606913520435200UL, 5787213827040870400UL, 11574427654081740800UL, 4629771061632696320UL,
        144680345642467328UL, 361700864106168320UL, 723401728212336640UL, 1446803456424673280UL, 2893606912849346560UL, 5787213825698693120UL, 11574427651397386240UL, 4629771060558954496UL,
        144680337052532736UL, 361700842631331840UL, 723401685262663680UL, 1446803370525327360UL, 2893606741050654720UL, 5787213482101309440UL, 11574426964202618880UL, 4629770785681047552UL,
        144678138029277184UL, 361695345073192960UL, 723390690146385920UL, 1446781380292771840UL, 2893562760585543680UL, 5787125521171087360UL, 11574251042342174720UL, 4629700416936869888UL,
        0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0,
        514,1285, 2570, 5140, 10280, 20560, 41120, 16448,
        131586, 328965, 657930, 1315860, 2631720, 5263440, 10526880, 4210752,
        33686018, 84215045, 168430090, 336860180, 673720360, 1347440720, 2694881440, 1077952576,
        8623620610, 21559051525, 43118103050, 86236206100, 172472412200, 344944824400, 689889648800, 275955859520,
        2207646876162, 5519117190405, 11038234380810, 22076468761620, 44152937523240, 88305875046480, 176611750092960, 70644700037184,
        0, 0, 0, 0, 0, 0, 0, 0
    }};

    static const int knightA8Penalty=150;
    static const int knightH8Penalty=150;
    static const int knightA7Penalty=100;
    static const int knightH7Penalty=100;
    static const int knightC3Penalty=5;

    static const int bishopA7Penalty=150;
    static const int bishopH7Penalty=150;
    static const int bishopB8Penalty=150;
    static const int bishopG8Penalty=150;
    static const int bishopA6Penalty=50;
    static const int bishopH6Penalty=50;
    static const int bishopF1Bonus=20;
    static const int bishopC1Bonus=20;

    static const int rookOpenBonus=10;
    static const int rookHalfOpenBonus=5;

    static const int shield1Bonus=10;
    static const int shield2Bonus=5;
    static const int noShieldPenalty=10;

    static const int centerPawnPenalty=24;
    
    static const int kingBlockRookPenalty=24;    
    
    static const int slaveMistakePenalty=10;
    
    static const int fianchettoBonus=4;
    static const int fianchettoPenalty=4;

    static const int pawnPieceSquare[2][64] = {{        
         0,  0,  0,  0,  0,  0,  0,  0,
        50, 50, 50, 50, 50, 50, 50, 50,
        10, 10, 20, 30, 30, 20, 10, 10,
         5,  5, 10, 25, 25, 10,  5,  5,
         0,  0,  0, 20, 20,  0,  0,  0,
         5, -5,-10,  0,  0,-10, -5,  5,
         5, 10, 10,-20,-20, 10, 10,  5,
         0,  0,  0,  0,  0,  0,  0,  0        
    },{
         0,  0,  0,  0,  0,  0,  0,  0,        
         5, 10, 10,-20,-20, 10, 10,  5,
         5, -5,-10,  0,  0,-10, -5,  5,         
         0,  0,  0, 20, 20,  0,  0,  0,
         5,  5, 10, 25, 25, 10,  5,  5,
        10, 10, 20, 30, 30, 20, 10, 10,         
        50, 50, 50, 50, 50, 50, 50, 50,         
         0,  0,  0,  0,  0,  0,  0,  0         
    }};

    static const int knightPieceSquare[2][64] = {{
        -50,-40,-30,-30,-30,-30,-40,-50,
        -40,-20,  0,  0,  0,  0,-20,-40,
        -30,  0, 10, 15, 15, 10,  0,-30,
        -30,  5, 15, 20, 20, 15,  5,-30,
        -30,  0, 15, 20, 20, 15,  0,-30,
        -30,  5, 10, 15, 15, 10,  5,-30,
        -40,-20,  0,  5,  5,  0,-20,-40,
        -50,-40,-30,-30,-30,-30,-40,-50        
    },{
        -50,-40,-30,-30,-30,-30,-40,-50,
        -40,-20,  0,  5,  5,  0,-20,-40,
        -30,  5, 10, 15, 15, 10,  5,-30,
        -30,  0, 15, 20, 20, 15,  0,-30,
        -30,  5, 15, 20, 20, 15,  5,-30,
        -30,  0, 10, 15, 15, 10,  0,-30,
        -40,-20,  0,  0,  0,  0,-20,-40,
        -50,-40,-30,-30,-30,-30,-40,-50        
    }};

    static const int bishopPieceSquare[2][64] = {{
        -20,-10,-10,-10,-10,-10,-10,-20,
        -10,  0,  0,  0,  0,  0,  0,-10,
        -10,  0,  5, 10, 10,  5,  0,-10,
        -10,  5,  5, 10, 10,  5,  5,-10,
        -10,  0, 10, 10, 10, 10,  0,-10,
        -10, 10, 10, 10, 10, 10, 10,-10,
        -10,  5,  0,  0,  0,  0,  5,-10,
        -20,-10,-10,-10,-10,-10,-10,-20        
    },{
        -20,-10,-10,-10,-10,-10,-10,-20,
        -10,  5,  0,  0,  0,  0,  5,-10,
        -10, 10, 10, 10, 10, 10, 10,-10,
        -10,  0, 10, 10, 10, 10,  0,-10,
        -10,  5,  5, 10, 10,  5,  5,-10,
        -10,  0,  5, 10, 10,  5,  0,-10,
        -10,  0,  0,  0,  0,  0,  0,-10,
        -20,-10,-10,-10,-10,-10,-10,-20        
    }};
    
    static const int rookPieceSquare[2][64] = {{
          0,  0,  0,  0,  0,  0,  0,  0,
          5, 10, 10, 10, 10, 10, 10,  5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
          0,  0,  0,  5,  5,  0,  0,  0                
    },{
          0,  0,  0,  5,  5,  0,  0,  0        
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
          5, 10, 10, 10, 10, 10, 10,  5,
          0,  0,  0,  0,  0,  0,  0,  0,
    }}; 
    
    static const int queenPieceSquare[2][64] = {{    
        -20,-10,-10, -5, -5,-10,-10,-20,
        -10,  0,  0,  0,  0,  0,  0,-10,
        -10,  0,  5,  5,  5,  5,  0,-10,
         -5,  0,  5,  5,  5,  5,  0, -5,
          0,  0,  5,  5,  5,  5,  0, -5,
        -10,  5,  5,  5,  5,  5,  0,-10,
        -10,  0,  5,  0,  0,  0,  0,-10,
        -20,-10,-10, -5, -5,-10,-10,-20        
    },{
        -20,-10,-10, -5, -5,-10,-10,-20,
        -10,  0,  5,  0,  0,  0,  0,-10,
        -10,  5,  5,  5,  5,  5,  0,-10,
          0,  0,  5,  5,  5,  5,  0, -5,
         -5,  0,  5,  5,  5,  5,  0, -5,        
        -10,  0,  5,  5,  5,  5,  0,-10,
        -10,  0,  0,  0,  0,  0,  0,-10,
        -20,-10,-10, -5, -5,-10,-10,-20        
    }};
    
    static const int kingPieceSquare[2][64] = {{
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -20,-30,-30,-40,-40,-30,-30,-20,
        -10,-20,-20,-20,-20,-20,-20,-10,
         20, 20,  0,  0,  0,  0, 20, 20,
         20, 30, 10,  0,  0, 10, 30, 20        
    },{
         20, 30, 10,  0,  0, 10, 30, 20,
         20, 20,  0,  0,  0,  0, 20, 20,         
        -10,-20,-20,-20,-20,-20,-20,-10,
        -20,-30,-30,-40,-40,-30,-30,-20,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30         
    }};
    
    static const int kingPieceSquareEndgame[2][64] = {{
        -50,-40,-30,-20,-20,-30,-40,-50,
        -30,-20,-10,  0,  0,-10,-20,-30,
        -30,-10, 20, 30, 30, 20,-10,-30,
        -30,-10, 30, 40, 40, 30,-10,-30,
        -30,-10, 30, 40, 40, 30,-10,-30,
        -30,-10, 20, 30, 30, 20,-10,-30,
        -30,-30,  0,  0,  0,  0,-30,-30,
        -50,-30,-30,-30,-30,-30,-30,-50
    },{
        -50,-30,-30,-30,-30,-30,-30,-50,
        -30,-30,  0,  0,  0,  0,-30,-30,
        -30,-10, 20, 30, 30, 20,-10,-30,
        -30,-10, 30, 40, 40, 30,-10,-30,
        -30,-10, 30, 40, 40, 30,-10,-30,
        -30,-10, 20, 30, 30, 20,-10,-30,
        -30,-20,-10,  0,  0,-10,-20,-30,
        -50,-40,-30,-20,-20,-30,-40,-50                
    }};
*/

    
    
    
    /*
     * Ende: Einige Konstanten der Eval-Methode
     */
    
    int quiesce(int alpha, int beta, int curr_depth);
    int evaluateFast(bool side2move);
    int evaluate(bool side2move);
    
    void sortMoves(unsigned short int sortedIndizes[], int cd, int bestmoveHash);
    unsigned short int sortMovesQuiesce(unsigned short int sortedIndizes[], int cd);
    void getMoveScores(short int moveScores[], int cd);
    
    
    inline void rochade(const unsigned long &old_king_pos, const unsigned long &new_king_pos, const unsigned long &old_rook_pos, const unsigned long &new_rook_pos);
    inline void updateBitboardLongs();    
    inline void split_moves_direct(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &myFigures);
    inline void split_moves_direct_captures(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op);
    inline void split_moves__(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op);
    inline void split_moves_captures__(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op);
    inline void addPawnSingles_black(const int &z);
    inline void addPawnSingles_white(const int &z);
    inline void addPawnSingles_black_captures(const int &z);
    inline void addPawnSingles_white_captures(const int &z);
    string get_hum(int k);
    inline int bitScanForward(const unsigned long &bb) {
        if (bb)
                return __builtin_ffsll(bb)-1;
        else
            return 0;
//        const unsigned long debruijn64 = 0x07EDD5E59A4E28C2;
//        return Position::index64[((bb & -bb) * debruijn64) >> 58];
        
//        const int index64[64] = {
//            0, 47,  1, 56, 48, 27,  2, 60,
//           57, 49, 41, 37, 28, 16,  3, 61,
//           54, 58, 35, 52, 50, 42, 21, 44,
//           38, 32, 29, 23, 17, 11,  4, 62,
//           46, 55, 26, 59, 40, 36, 15, 53,
//           34, 51, 20, 43, 31, 22, 10, 45,
//           25, 39, 14, 33, 19, 30,  9, 24,
//           13, 18,  8, 12,  7,  6,  5, 63
//        };
//
//        
//        const unsigned long debruijn64 = 0x03f79d71b4cb0a89;
//           return index64[((bb ^ (bb-1)) * debruijn64) >> 58];        

    };
    inline int popCount(unsigned long x) { // too slow
//        int count = 0;
//        for (int i = 0; i < 64; i++, x >>= 1)
//                count += (int)x & 1;
//        return count;
        
      int c = 0;

      while (x) {
        c++;
        x &= x - 1;
      }
      return (c);        
        
    }
    bool opponent_target_bitboard2(const unsigned long &bitboard);
};

#endif	/* POSITION_H */