#include <iostream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cstdint>
#include <ctime>
#include "Position.h"
#include "HashT.h"
#include "MoveList.h"
#include "Timer.h"
#include "globals.h"

using namespace std;

extern HashT hashTable;
extern MoveList moveList;
extern Timer timer;

extern unsigned short int distances[64][64];

/***********************************************************************
 * Die evaluate-Funktion bewertet eine gegebene Position               *
 * nach verschiedenen heuristischen Methoden für jeweils beide Spieler *
 * Die Werte der Spieler zu jeder Methode werden subtrahiert,          *
 * nach der Priorität gewichtet und zu einem Gesamt-Score              *
 * zusammengefasst.                                                    *
 ***********************************************************************/

int Position::evaluateFast(bool side2move) {
    int materialScore0, materialScore1;

    materialScore0  =   100 *   popCount(pawns[0]) +
                        275 *   popCount(knights[0]) +
                        325 *   popCount(bishops[0]) +
                        465 *   popCount(rooks[0]) +
                        900 *   popCount(queens[0]) +
                        5000 *  popCount(kings[0]);    // sollte der score hier so hoch sein (50000)? oder kommt ein overflow zustande?
    
    materialScore1  =   100 *   popCount(pawns[1]) +
                        275 *   popCount(knights[1]) +
                        325 *   popCount(bishops[1]) +
                        465 *   popCount(rooks[1]) +
                        900 *   popCount(queens[1]) +
                        5000 *  popCount(kings[1]);    
    
    int score = (materialScore0 - materialScore1);    
     
    if (side2move)
        score*=-1;        
    return score;    
}

int Position::evaluate(bool side2move) {
    
    /******************
     * Material Score *
     * ****************/
    
    int materialScore0, materialScore1;

    materialScore0  =   100 *   popCount(pawns[0]) +
                        275 *   popCount(knights[0]) +
                        325 *   popCount(bishops[0]) +
                        465 *   popCount(rooks[0]) +
                        900 *   popCount(queens[0]) +
                        5000 *  popCount(kings[0]);    
    
    materialScore1  =   100 *   popCount(pawns[1]) +
                        275 *   popCount(knights[1]) +
                        325 *   popCount(bishops[1]) +
                        465 *   popCount(rooks[1]) +
                        900 *   popCount(queens[1]) +
                        5000 *  popCount(kings[1]);        

    /********************************************************************
     * Anpassung des Material Scores abhängig von der Anzahl der Bauern *
     ********************************************************************/

    
    materialScore0 += popCount(knights[0]) * knightAdj[popCount(pawns[0])];
    materialScore0 += popCount(rooks[0]) * rookAdj[popCount(pawns[0])];
    materialScore1 += popCount(knights[1]) * knightAdj[popCount(pawns[1])];
    materialScore1 += popCount(rooks[1]) * rookAdj[popCount(pawns[1])];
    
    /*****************************
     * Keine Bauern mehr am Feld *
     *****************************/
    
    if (popCount(pawns[0])==0)  materialScore0-=16;
    if (popCount(pawns[1])==0)  materialScore1-=16;
    
    /*********
     * Tempo *
     *********/
    
    if (side2move) materialScore1+=10;
    else           materialScore0+=10;     

    /************************************************
     * King Tropism                                 *
     * (Abstand der Figuren zum gegnerischen König) *
     ************************************************/

    unsigned long bitboard;
    
    int kingTropism0=0, kingTropism1=0;

    unsigned short int posKingWhite = bitScanForward(kings[0]);
    unsigned short int posKingBlack = bitScanForward(kings[1]);
    
    

    
    /*****************
     * weiße Figuren *
     *****************/

    bitboard = knights[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism0+=distances[x][posKingBlack];
        bitboard-=Position::BIN[x];
    }

    bitboard = rooks[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism0+=distances[x][posKingBlack]/2;
        bitboard-=Position::BIN[x];
    }    

    bitboard = queens[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism0+=distances[x][posKingBlack]*5/2;
        bitboard-=Position::BIN[x];
    }        
    
    bitboard = bishops[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism0+=distances[x][posKingBlack]/2;
        kingTropism0+=bonusDiaDistance[abs(diagNE[x]-diagNE[posKingBlack])];
        kingTropism0+=bonusDiaDistance[abs(diagNW[x]-diagNW[posKingBlack])];
        bitboard-=Position::BIN[x];
    }        
    
    /********************
     * schwarze Figuren *
     ********************/
    
    bitboard = knights[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism1+=distances[x][posKingWhite];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = rooks[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism1+=distances[x][posKingWhite]/2;
        bitboard-=Position::BIN[x];
    }    
    
    bitboard = queens[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism1+=distances[x][posKingWhite]*5/2;
        bitboard-=Position::BIN[x];
    }
    
    bitboard = bishops[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism1+=distances[x][posKingWhite]/2;
        kingTropism1+=bonusDiaDistance[abs(diagNE[x]-diagNE[posKingWhite])];
        kingTropism1+=bonusDiaDistance[abs(diagNW[x]-diagNW[posKingWhite])];        
        bitboard-=Position::BIN[x];
    }    
    
    
    /******************
     * Pawn Structure *
     ******************/
    
    /********************************************
     * Doppelte Bauern in einer Linie (Penalty) *
     ********************************************/
    
    int pawnScore0=0, pawnScore1=0;
    
    for (unsigned short int i=0;i<8;i++) {
        pawnScore0-=multiPawnPenalty[popCount(cols[i]&pawns[0])];
        pawnScore1-=multiPawnPenalty[popCount(cols[i]&pawns[1])];
    }
    
    /*****************************
     * Bauern im Zentrum (Bonus) *
     *****************************/
    
    if ((d4e4bb&pawns[0])==d4e4bb)        pawnScore0+=d4e4bonus;
    if ((d4c4bb&pawns[0])==d4c4bb)        pawnScore0+=d4c4bonus;
    if ((d4e3bb&pawns[0])==d4e3bb)        pawnScore0+=d4e3bonus;
    if ((e4f4bb&pawns[0])==e4f4bb)        pawnScore0+=e4f4bonus;
    if ((e4d3bb&pawns[0])==e4d3bb)        pawnScore0+=e4d3bonus;
    
    if ((d5e5bb&pawns[1])==d5e5bb)        pawnScore1+=d4e4bonus;
    if ((d5c5bb&pawns[1])==d5c5bb)        pawnScore1+=d4c4bonus;
    if ((d5e6bb&pawns[1])==d5e6bb)        pawnScore1+=d4e3bonus;
    if ((e5f5bb&pawns[1])==e5f5bb)        pawnScore1+=e4f4bonus;
    if ((e5d6bb&pawns[1])==e5d6bb)        pawnScore1+=e4d3bonus;    
    
    /****************
     * Passed Pawns *
     ****************/
       
    bitboard = pawns[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if (    (passedPawnsBitboards[0][x]&pawns[0])==0 && (passedPawnsBitboards[0][x]&pawns[1])==0 // Felder vor Bauern sind frei
                && (passedPawnsBitboardsAttacks[0][x]&pawns[1])==0 // Felder vor Bauern werden nicht vom Gegner durch Bauern bedroht
                ) {
            
            if ((passedPawnsProtectedBitboards[0][x]&pawns[0])!=0)
                pawnScore0+=passedPawnsProtectedBonus[0][x];
            else
                pawnScore0+=passedPawnsBonus[0][x];
        
        }
        
        bitboard-=Position::BIN[x];
    }
    
    bitboard = pawns[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if (    (passedPawnsBitboards[1][x]&pawns[0])==0 && (passedPawnsBitboards[1][x]&pawns[1])==0 // Felder vor Bauern sind frei
                && (passedPawnsBitboardsAttacks[1][x]&pawns[0])==0 // Felder vor Bauern werden nicht vom Gegner durch Bauern bedroht                
                ) {
            
            if ((passedPawnsProtectedBitboards[1][x]&pawns[1])!=0)
                pawnScore1+=passedPawnsProtectedBonus[1][x];
            else
                pawnScore1+=passedPawnsBonus[1][x];
            
        }
        
        bitboard-=Position::BIN[x];
    }    
    
    /**************
     * Weak Pawns *
     **************/
    
    bitboard = pawns[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if (    (weakPawnsBitboards[0][x]&pawns[0])==0 )
                pawnScore0+=weakPawnsPenalty[0][x];
        
        bitboard-=Position::BIN[x];
    }    
    
    bitboard = pawns[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if (    (weakPawnsBitboards[1][x]&pawns[1])==0 )
                pawnScore1+=weakPawnsPenalty[1][x];
        
        bitboard-=Position::BIN[x];
    }    
        
    
    
    
    
    
    
    
    
    
    
    /********************
     * Piece Evaluation *
     ********************/
    
    int pieceScore0=0, pieceScore1=0;
    
    /************
     * Springer *
     ************/
    

    
    // A8
    if ((knights[0]&BIN[0])!=0) {
        if ((pawns[1]&BIN[8])!=0 || (pawns[1]&BIN[10])!=0)
            pieceScore0-=knightA8Penalty;
    }
    if ((knights[1]&BIN[56])!=0) {
        if ((pawns[0]&BIN[48])!=0 || (pawns[0]&BIN[50])!=0)
            pieceScore1-=knightA8Penalty;
    }
    
    // H8
    if ((knights[0]&BIN[7])!=0) {
        if ((pawns[1]&BIN[13])!=0 || (pawns[1]&BIN[15])!=0)
            pieceScore0-=knightH8Penalty;
    }
    if ((knights[1]&BIN[63])!=0) {
        if ((pawns[0]&BIN[55])!=0 || (pawns[0]&BIN[53])!=0)
            pieceScore1-=knightH8Penalty;
    }
    
    // A7
    if ((knights[0]&BIN[8])!=0) {
        if ((pawns[1]&BIN[16])!=0 && (pawns[1]&BIN[9])!=0)
            pieceScore0-=knightA7Penalty;
    }
    if ((knights[1]&BIN[48])!=0) {
        if ((pawns[0]&BIN[40])!=0 && (pawns[0]&BIN[49])!=0)
            pieceScore1-=knightA7Penalty;
    }
    
    // H7
    if ((knights[0]&BIN[15])!=0) {
        if ((pawns[1]&BIN[23])!=0 && (pawns[1]&BIN[14])!=0)
            pieceScore0-=knightH7Penalty;
    }
    if ((knights[1]&BIN[55])!=0) {
        if ((pawns[0]&BIN[47])!=0 && (pawns[0]&BIN[54])!=0)
            pieceScore1-=knightH7Penalty;
    }
    
    // C3
    if ((knights[0]&BIN[42])!=0) {
        if ((pawns[0]&BIN[50])!=0 && (pawns[0]&BIN[35])!=0 && (pawns[0]&BIN[36])==0)
            pieceScore0-=knightC3Penalty;            
    }
    if ((knights[1]&BIN[18])!=0) {
        if ((pawns[1]&BIN[10])!=0 && (pawns[1]&BIN[27])!=0 && (pawns[1]&BIN[28])==0)
            pieceScore1-=knightC3Penalty;
    }
        
    /**********
     * Läufer *
     **********/
    


    // A7
    if ((rooks[0]&BIN[8])!=0) {
        if ((pawns[1]&BIN[17])!=0)
            pieceScore0-=bishopA7Penalty;
    }
    if ((rooks[1]&BIN[48])!=0) {
        if ((pawns[0]&BIN[41])!=0)
            pieceScore1-=bishopA7Penalty;
    }
    
    // H7
    if ((rooks[0]&BIN[15])!=0) {
        if ((pawns[1]&BIN[22])!=0)
            pieceScore0-=bishopH7Penalty;
    }
    if ((rooks[1]&BIN[55])!=0) {
        if ((pawns[0]&BIN[46])!=0)
            pieceScore1-=bishopH7Penalty;
    }
    
    // B8
    if ((rooks[0]&BIN[1])!=0) {
        if ((pawns[1]&BIN[10])!=0)
            pieceScore0-=bishopB8Penalty;
    }
    if ((rooks[1]&BIN[57])!=0) {
        if ((pawns[0]&BIN[50])!=0)
            pieceScore1-=bishopB8Penalty;
    }
    
    // G8
    if ((rooks[0]&BIN[6])!=0) {
        if ((pawns[1]&BIN[13])!=0)
            pieceScore0-=bishopG8Penalty;
    }
    if ((rooks[1]&BIN[62])!=0) {
        if ((pawns[0]&BIN[53])!=0)
            pieceScore1-=bishopG8Penalty;
    }
    
    // A6
    if ((rooks[0]&BIN[16])!=0) {
        if ((pawns[1]&BIN[25])!=0)
            pieceScore0-=bishopA6Penalty;
    }
    if ((rooks[1]&BIN[40])!=0) {
        if ((pawns[0]&BIN[33])!=0)
            pieceScore1-=bishopA6Penalty;
    }
    
    // H6
    if ((rooks[0]&BIN[23])!=0) {
        if ((pawns[1]&BIN[30])!=0)
            pieceScore0-=bishopH6Penalty;
    }
    if ((rooks[1]&BIN[47])!=0) {
        if ((pawns[0]&BIN[38])!=0)
            pieceScore1-=bishopH6Penalty;
    }
    
    // F1
    if ((rooks[0]&BIN[61])!=0) {
        if ((kings[0]&BIN[62])!=0)
            pieceScore0+=bishopF1Bonus;
    }
    if ((rooks[1]&BIN[5])!=0) {
        if ((kings[1]&BIN[6])!=0)
            pieceScore1+=bishopF1Bonus;
    }
    
    // C1
    if ((rooks[0]&BIN[58])!=0) {
        if ((kings[0]&BIN[57])!=0)
            pieceScore0+=bishopC1Bonus;
    }
    if ((rooks[1]&BIN[2])!=0) {
        if ((kings[1]&BIN[1])!=0)
            pieceScore1+=bishopC1Bonus;
    }    
        
    /********************************************
     * Turm                                     *
     * - Bonus für offene und halboffene Linien *
     ********************************************/


    
    for (unsigned short int i=0;i<8;i++) {
        if ((rooks[0]&cols[i])!=0 && (cols[i]&pawns[0])==0) {
            if ((cols[i]&pawns[1])==0)
                pieceScore0+=rookOpenBonus;
            else
                pieceScore0+=rookHalfOpenBonus;
        }
    }

    for (unsigned short int i=0;i<8;i++) {
        if ((rooks[1]&cols[i])!=0 && (cols[i]&pawns[1])==0) {
            if ((cols[i]&pawns[0])==0)
                pieceScore1+=rookOpenBonus;
            else
                pieceScore1+=rookHalfOpenBonus;
        }
    }    

    /**************************************
     * Dame                               *
     * - Penalty für zu frühe Entwicklung *
     **************************************/

    bitboard = queens[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if ((BIN[x]&281474976710655)!=0) {
            if ((knights[0]&BIN[57])!=0)  pieceScore0-=2;
            if ((knights[0]&BIN[62])!=0)  pieceScore0-=2;
            if ((bishops[0]&BIN[58])!=0)  pieceScore0-=2;
            if ((bishops[0]&BIN[61])!=0)  pieceScore0-=2;
        }

        bitboard-=Position::BIN[x];
    }
    
    bitboard = queens[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if ((BIN[x]&18446744073709486080UL)!=0) {
            if ((knights[1]&BIN[1])!=0)  pieceScore1-=2;
            if ((knights[1]&BIN[6])!=0)  pieceScore1-=2;
            if ((bishops[1]&BIN[2])!=0)  pieceScore1-=2;
            if ((bishops[1]&BIN[5])!=0)  pieceScore1-=2;
        }

        bitboard-=Position::BIN[x];
    }    
    
    
    /*********
     * König *
     *********/

    unsigned short int kCol;    
    
    /****************
     * Weißer König *
     ****************/
    
    for (kCol=0;kCol<8;kCol++) {
        if ((kings[0]&cols[kCol])!=0)
            break;
    }
    
    // king side
    if (kCol>4) {
        
        if ((pawns[0]&BIN[53])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[45])!=0)   pieceScore0+=shield2Bonus;
        
        if ((pawns[0]&BIN[54])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[46])!=0)   pieceScore0+=shield2Bonus;
        else if ((pawns[0]&cols[6])==0)  pieceScore0-=noShieldPenalty;
        
        if ((pawns[0]&BIN[55])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[47])!=0)   pieceScore0+=shield2Bonus;
        else if ((pawns[0]&cols[7])==0)  pieceScore0+=noShieldPenalty;
        
    }
    
    // queenside
    if (kCol<3) {
        
        if ((pawns[0]&BIN[48])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[40])!=0)   pieceScore0+=shield2Bonus;
        else if ((pawns[0]&cols[0])==0)  pieceScore0-=noShieldPenalty;
        
        if ((pawns[0]&BIN[49])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[41])!=0)   pieceScore0+=shield2Bonus;
        else if ((pawns[0]&cols[1])==0)  pieceScore0+=noShieldPenalty;        

        if ((pawns[0]&BIN[50])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[42])!=0)   pieceScore0+=shield2Bonus;        
        
    }
    
    /*******************
     * Schwarzer König *
     *******************/
    
    for (kCol=0;kCol<8;kCol++) {
        if ((kings[1]&cols[kCol])!=0)
            break;
    }
    
    // king side
    if (kCol>4) {
        
        if ((pawns[1]&BIN[13])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[21])!=0)   pieceScore1+=shield2Bonus;
        
        if ((pawns[1]&BIN[14])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[22])!=0)   pieceScore1+=shield2Bonus;
        else if ((pawns[1]&cols[6])==0)  pieceScore1-=noShieldPenalty;
        
        if ((pawns[1]&BIN[15])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[23])!=0)   pieceScore1+=shield2Bonus;
        else if ((pawns[1]&cols[7])==0)  pieceScore1+=noShieldPenalty;
        
    }
    
    // queenside
    if (kCol<3) {
        
        if ((pawns[1]&BIN[8])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[16])!=0)   pieceScore1+=shield2Bonus;
        else if ((pawns[1]&cols[0])==0)  pieceScore1-=noShieldPenalty;
        
        if ((pawns[1]&BIN[9])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[17])!=0)   pieceScore1+=shield2Bonus;
        else if ((pawns[1]&cols[1])==0)  pieceScore1+=noShieldPenalty;        

        if ((pawns[1]&BIN[10])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[18])!=0)   pieceScore1+=shield2Bonus;
        
    }

    /**************************
     * Diverse Pattern-Matches *
     **************************/
    
    int patternScore0=0, patternScore1=0;
    
    /**********************************
     * Blockierte Bauern in der Mitte *
     **********************************/
    
    // D2
    if ((pawns[0]&BIN[51])!=0 && (all&BIN[43])!=0)
        patternScore0-=centerPawnPenalty;
    
    // E2
    if ((pawns[0]&BIN[52])!=0 && (all&BIN[44])!=0)
        patternScore0-=centerPawnPenalty;
    
    // D7
    if ((pawns[1]&BIN[11])!=0 && (all&BIN[19])!=0)
        patternScore1-=centerPawnPenalty;
    
    // E7
    if ((pawns[1]&BIN[12])!=0 && (all&BIN[20])!=0)
        patternScore1-=centerPawnPenalty;
    
    /********************
     * Blockierte Türme *
     ********************/
    
    if(         ((kings[0]&BIN[61])!=0 || (kings[0]&BIN[62])!=0)
            &&  ((rooks[0]&BIN[63])!=0 || (rooks[0]&BIN[62])!=0))
        patternScore0-=kingBlockRookPenalty;
    
    if(         ((kings[0]&BIN[58])!=0 || (kings[0]&BIN[57])!=0)
            &&  ((rooks[0]&BIN[56])!=0 || (rooks[0]&BIN[57])!=0))
        patternScore0-=kingBlockRookPenalty;
    
    if(         ((kings[1]&BIN[5])!=0 || (kings[1]&BIN[6])!=0)
            &&  ((rooks[1]&BIN[7])!=0 || (rooks[1]&BIN[6])!=0))
        patternScore1-=kingBlockRookPenalty;
    
    if(         ((kings[1]&BIN[2])!=0 || (kings[1]&BIN[1])!=0)
            &&  ((rooks[1]&BIN[0])!=0 || (rooks[1]&BIN[1])!=0))
        patternScore1-=kingBlockRookPenalty;    
    
    /****************
     * Slave Fehler *
     ****************/
    
    if ((pawns[0]&BIN[35])!=0 && (pawns[0]&BIN[26])!=0 &&
            (pawns[1]&BIN[27])!=0 && (pawns[1]&BIN[18])!=0)
        patternScore0-=slaveMistakePenalty;
    
    if ((pawns[1]&BIN[27])!=0 && (pawns[1]&BIN[34])!=0 &&
            (pawns[0]&BIN[35])!=0 && (pawns[0]&BIN[42])!=0)
        patternScore1-=slaveMistakePenalty;
    
    /**************
     * Fianchetto *
     **************/
    
    if ((pawns[0]&BIN[46])!=0) {
        if ((bishops[0]&BIN[54])!=0)
            patternScore0+=fianchettoBonus;
        else {
            if (    (bishops[0]&BIN[45])==0 && 
                    (bishops[0]&BIN[47])==0 && 
                    (bishops[0]&BIN[63])==0)
                patternScore0-=fianchettoPenalty;
        }
    }

    if ((pawns[0]&BIN[41])!=0) {
        if ((bishops[0]&BIN[49])!=0)
            patternScore0+=fianchettoBonus;
        else {
            if (    (bishops[0]&BIN[42])==0 && 
                    (bishops[0]&BIN[40])==0 && 
                    (bishops[0]&BIN[56])==0)
                patternScore0-=fianchettoPenalty;
        }
    }
    
    if ((pawns[1]&BIN[22])!=0) {
        if ((bishops[1]&BIN[14])!=0)
            patternScore1+=fianchettoBonus;
        else {
            if (    (bishops[1]&BIN[21])==0 && 
                    (bishops[1]&BIN[23])==0 && 
                    (bishops[1]&BIN[7])==0)
                patternScore1-=fianchettoPenalty;
        }
    }
    
    if ((pawns[1]&BIN[17])!=0) {
        if ((bishops[1]&BIN[9])!=0)
            patternScore1+=fianchettoBonus;
        else {
            if (    (bishops[1]&BIN[18])==0 && 
                    (bishops[1]&BIN[16])==0 && 
                    (bishops[1]&BIN[0])==0)
                patternScore1-=fianchettoPenalty;
        }
    }    
    
    /************************************************************************
     * berechne targetmaps und mobility score                               *
     * targetmaps werden nachfolgend zb. für den capture score benötigt     *
     * zur berechnung des mobility score wird die gesamte zugliste benötigt *
     * daher werden diese beiden elemente hier zusammen berechnet           *
     * **********************************************************************/

//    long targetmap0, targetmap1;
//    int mobilityScore0 = 0, mobilityScore1 = 0;
//    int moveSize0, moveSize1;
//    
//    /***************
//     * weiß am zug *
//     * *************/
//
//    onMove=false;
////    targetmap=0L;
//    calc_moves(false); // true=with castling
//    moveSize0=mvSize[cd];
////    targetmap0=targetmap;
//    
//    for (unsigned int i=0;i<mvSize[cd];i++) {
//        // if ((mvD[cd][i][0]&kings[0])!=0L) mobilityScore0+=20;
//        // if ((mvD[cd][i][0]&queens[0])!=0L) mobilityScore0+=9;
//        if ((mvD[cd][i][0]&rooks[0])!=0L) mobilityScore0+=5;
//        else if ((mvD[cd][i][0]&bishops[0])!=0L) mobilityScore0+=3;
//        else if ((mvD[cd][i][0]&knights[0])!=0L) mobilityScore0+=3;
//        else if ((mvD[cd][i][0]&pawns[0])!=0L) mobilityScore0+=1;
//    }
//
//    /******************
//     * schwarz am zug *
//     * ****************/    
//    
//    onMove=true;
////    targetmap=0L;
//    calc_moves(false); // true=with castling
//    moveSize1=mvSize[cd];
////    targetmap1=targetmap;    
//
//    for (unsigned int i=0;i<mvSize[cd];i++) {
//        // if ((mvD[cd][i][0]&kings[1])!=0L) mobilityScore1+=20;
//        // if ((mvD[cd][i][0]&queens[1])!=0L) mobilityScore1+=9;
//        if ((mvD[cd][i][0]&rooks[1])!=0L) mobilityScore1+=5;
//        else if ((mvD[cd][i][0]&bishops[1])!=0L) mobilityScore1+=3;
//        else if ((mvD[cd][i][0]&knights[1])!=0L) mobilityScore1+=3;
//        else if ((mvD[cd][i][0]&pawns[1])!=0L) mobilityScore1+=1;
//    }    
//
//     mobilityScore0/=moveSize0;
//     mobilityScore1/=moveSize1;
        
    
    /*************************
     * Anzahl möglicher Züge *
     *************************/
    
//    int legalMoves0, legalMoves1;
//    int legalMovesWeight=70;
    
//    legalMoves0 = popCount(targetmap0);
//    legalMoves1 = popCount(targetmap1);
    
    
    /****************************************************************
     * berechne capture score                                       *
     * in abhängigkeit der figur-typen die geschlagen werden können *
     * **************************************************************/    
    
    // sollte captureScore bei idealer quiescence-suche nicht 0 sein?
    
//    int captureScore0 = 0, captureScore1 = 0;
//    int captureWeight = 80;
    
//    if ((targetmap0&kings[1])!=0)       captureScore0+=20;
//    if ((targetmap0&queens[1])!=0)      captureScore0+=9*popCount(targetmap0&queens[1]);
//    if ((targetmap0&rooks[1])!=0)       captureScore0+=5*popCount(targetmap0&rooks[1]);
//    if ((targetmap0&bishops[1])!=0)     captureScore0+=3*popCount(targetmap0&bishops[1]);
//    if ((targetmap0&knights[1])!=0)     captureScore0+=3*popCount(targetmap0&knights[1]);
//    if ((targetmap0&pawns[1])!=0)       captureScore0+=popCount(targetmap0&pawns[1]);
//
//    if ((targetmap1&kings[0])!=0)       captureScore1+=20;
//    if ((targetmap1&queens[0])!=0)      captureScore1+=9*popCount(targetmap1&queens[0]);
//    if ((targetmap1&rooks[0])!=0)       captureScore1+=5*popCount(targetmap1&rooks[0]);
//    if ((targetmap1&bishops[0])!=0)     captureScore1+=3*popCount(targetmap1&bishops[0]);
//    if ((targetmap1&knights[0])!=0)     captureScore1+=3*popCount(targetmap1&knights[0]);
//    if ((targetmap1&pawns[0])!=0)       captureScore1+=popCount(targetmap1&pawns[0]);    
//    
        
    
    
    
    /*********************************************************************
     * Piece-Square Tabellen                                             *
     * Berechnet Punkt-Gewinn/Abzug für Figuren an bestimmten Positionen *
     *********************************************************************/
    
    int pieceSquare0=0, pieceSquare1=0;
    
    /********
     * weiß *
     ********/
    
    bitboard = pawns[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=pawnPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = knights[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=knightPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = bishops[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=bishopPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = rooks[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=rookPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = queens[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=queenPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }    

    if (isEndgame()) {
        bitboard = kings[0];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            pieceSquare0+=kingPieceSquareEndgame[0][x];
            bitboard-=Position::BIN[x];
        }
    } else {
        bitboard = kings[0];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            pieceSquare0+=kingPieceSquare[0][x];
            bitboard-=Position::BIN[x];
        }
    }
    
    /***********
     * schwarz *
     ***********/
    
    bitboard = pawns[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=pawnPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = knights[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=knightPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = bishops[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=bishopPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }    

    bitboard = rooks[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=rookPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }    

    bitboard = queens[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=queenPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }        

    if (isEndgame()) {
        bitboard = kings[1];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            pieceSquare1+=kingPieceSquareEndgame[1][x];
            bitboard-=Position::BIN[x];
        }
    } else {
        bitboard = kings[1];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            pieceSquare1+=kingPieceSquare[1][x];
            bitboard-=Position::BIN[x];
        }        
    }
    
    /*****************************************************************************************************************
     * berechne gesamt score                                                                                         *
     * die gewichtung der einzelnen werte sollte später automatisch durch ein skript ermittelt werden.               *
     * mit entsprechend vielen tournaments sollte so eine verbesserung bzw. verschlechterung der ELO erkennbar sein. *
     * die dadurch gewonnen daten können auch für weitere analysen verwendet werden um so zu ermitteln,              *
     * ob es werte gibt, die eigentlich gar nicht zwingend berechnet werden müssten, da sie ohnehin                  *
     * mit anderen werten korrelieren                                                                                *
     *****************************************************************************************************************/    
    
    int score = (materialScore0 - materialScore1) +
                (kingTropism0 - kingTropism1) +
                (patternScore0 - patternScore1) +
                (pawnScore0 - pawnScore1) +
                (pieceScore0 - pieceScore1) +
                (pieceSquare0 - pieceSquare1);

//                (mobilityScore0 - mobilityScore1) +                
                
   // cout << (legalMoves0 - legalMoves1) * legalMovesWeight << endl;
    
    /******************
     * we're ready :) *
     * ****************/        
    
    if (side2move)
        score*=-1;
        
    return score;
}


