#include <iostream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cstdint>
#include <ctime>
#include "Position.h"
#include "HashT.h"
#include "MoveList.h"
#include "Timer.h"
#include "globals.h"

using namespace std;

extern HashT hashTable;
extern MoveList moveList;
extern Timer timer;

extern unsigned short int distances[64][64];

/**************************************************************
 * NegaMax ist die Kernfunktion zur Suche nach dem besten Zug *
 * Aktuell implementiert ist dabei:                           *
 *   - eine einfache Zugsortierung                            *
 *   - Transposition/Hash-Table                               *
 *   - Quiescence Search                                      *
 **************************************************************/
int Position::negaMax(int alpha, int beta, int curr_depth, int depth, bool isPv) { //, bool canNull) {

    sum++;
    
    cd = curr_depth;
    
    if (!timer.isTimeout() && (sum&1023)==0)
        timer.checkTimeout();    
    if (timer.isTimeout())
        return 0;    
    

    if (curr_depth==depth)
//        return evaluate(onMove);
        return quiesce(alpha,beta,curr_depth);

    
    // unsigned long hashKey = hashTable.getHashKey(this);
    moveList.setMoveAt(curr_depth,hashKey,"",0);
    
    /*********************************************************************************
     * Überprüfe ob die aktuelle Position bereits im Spiel- bzw. Suchverlauf vorkam. *
     * Ggf. wird ein Score für ein Draw bzw. ein entsprechender Contempt-Faktor      *
     * zurück gegeben.                                                               *
     *********************************************************************************/

//    if (cd>1 && moveList.isRepetition(cd)) {
//        int contemptOpening=-50;
//        int contemptMiddlegame=-25;
//        int contempt=0;
//        if (moveList.getNumHalfMoves()<20)
//            contempt=contemptOpening;
//        else if (!isEndgame())
//            contempt=contemptMiddlegame;
//        if (cd%2==1)
//            contempt=-contempt;
//        return contempt;
//    }
    
    /****************************************************************************
     * Suche in der HashTable nach der aktuellen Position und prüfe ggf. ob der *
     * gespeicherte Score verwendet werden kann.                                *
     ****************************************************************************/
    int score2;
    int bestmoveHash=-1;
    
    if (cd>1 && hashTable.contains(hashKey,depth-cd,score2,alpha,beta,bestmoveHash)) { // cd>1 oder cd>0 ???
//        cout << "deub msg: score from hashtable = " << score2 << endl;
        return score2;
    }
        
//    if (bestmoveHash>=0)
//        cout << "debug msg: bestmove index from hashtable = " << bestmoveHash << endl;

    int score=0;  
    


    /************************************************************************
     * Züge der aktuellen Position generieren und zwecks Pruning sortieren. *
     ************************************************************************/
    calc_moves();
    unsigned short int sortedIndizes[256];
    sortMoves(sortedIndizes,cd,bestmoveHash);
    doBackup();

    unsigned int nodeType=HASHT_ALPHA;
    int maxWert=alpha;
    
    int i;
    int bmIndex=sortedIndizes[0];

    for (i=mvSize[cd]-1;i>=0;i--) {
        unsigned short int ret = move_figure(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4);        
        
        if (curr_depth==0) {
            onMove=!onMove;
            bool catchKing=opponent_target_bitboard2(kings[onMove]); // TODO wollen wir das in jeder tiefe?
            onMove=!onMove;
            if (catchKing) {
                undoMove(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4,ret);
                continue;
            }
        }
    

        score = -negaMax(-beta,-maxWert,(curr_depth+1),depth,false);
        cd=curr_depth;
        
        /*****************************************************
         * Zug rückgängig machen und den Score des Zuges     *
         * ggf. speichern, bzw. überprüfen ob die restlichen *
         * Züge noch durchgeführt werden müssen.             *
         *****************************************************/
        undoMove(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4,ret);        

//        if (score==0 && timer.isTimeout())
//            return 0;
        
        if (score > maxWert) {
            bmIndex=sortedIndizes[i];
            maxWert = score;
            if (curr_depth==0)
                bestMoveIndex=sortedIndizes[i];
            if (maxWert>=beta) {
                nodeType=HASHT_BETA;
                maxWert=beta;
//                hashTable.insert(hashKey,maxWert,2,bmIndex,depth-cd); // beta
//                moveList.setMoveAt(curr_depth,hashKey,getMoveString(bmIndex),(depth-curr_depth));
//                return beta;
                break;
            }
                       
            nodeType=HASHT_EXACT;
//            hashTable.insert(hashKey,maxWert,3,bmIndex,depth-cd); // alpha
        }
    }

    /**************************************************
     * Den berechneten Score in der HashTable und den *
     * besten Zug in der GameHistory speichern.       *
     **************************************************/
    hashTable.insert(hashKey,maxWert,nodeType,bmIndex,depth-cd); // exact
//    moveList.setMoveAt(curr_depth,hashKey,getMoveString(bmIndex),(depth-curr_depth));
    return maxWert;
}

/************************************************************************
 * Die Quiescence-Suche wird an jedem Leaf-Node des NegaMax aufgerufen. *
 * Betrachtet werden hierbei nur noch jene Züge die Figuren schlagen.   *
 * Im Idealfall wird die Suche solange fortgesetzt, bis eine Position   *
 * erreicht wurde, in der keine Figur geschlagen werden kann oder nicht *
 * mehr tiefer evaluiert werden muss.                                   *
 * StandingPat wurde bereits implementiert. Es fehlt allerdings noch    *
 * die Implementierung der Static Exchange Evaluation bzw. auch         *
 * Delta-Pruning.                                                       *
 ************************************************************************/
int Position::quiesce(int alpha, int beta, int curr_depth) {
       
    cd = curr_depth;

    sum++;
//    
//    cd = curr_depth;
//    
    if (!timer.isTimeout() && (sum&1023)==0)
        timer.checkTimeout();    
    if (timer.isTimeout())
        return 0;    
    
    
    int standPat = evaluate(onMove);
    if (standPat>=beta)
        return beta;
    if(standPat>alpha)
        alpha=standPat;

    calcCaptures();
    if (!mvSize[cd])
        return alpha;

    unsigned short int sortedIndizes[256];
    sortMovesQuiesce(sortedIndizes,cd);

//    unsigned short int moveScores[256];
//    getMoveScores(moveScores,cd);
    
    doBackup();
    for (int i=mvSize[cd]-1;i>=0;i--) {    
        
//        if ((mvD[cd][sortedIndizes[i]][1]&kings[!onMove])!=0)
//            return 20000;
        
        unsigned short int ret=move_figure(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4);
        int score = -quiesce(-beta,-alpha,(curr_depth+1));
        cd--;
        undoMove(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4,ret);
        if (score > alpha) {
            if (score>=beta) 
                return beta;
            alpha=score;
        }
    }
    return alpha;
}
