/* 
 * File:   Position.cpp
 * Author: peter
 * 
 * Created on 31. Jänner 2012, 18:23
 */

#include <iostream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cstdint>
#include <ctime>
#include "Position.h"
#include "Heap.h"
#include "HashT.h"
#include "globals.h"

using namespace std;

/************************************
 * Initialisiert eine Startposition *
 ************************************/
Position::Position() {
    
    ep=0;
    targetmap=0L;
    
    kings[1]=Position::BIN[4];
    queens[1]=Position::BIN[3];
    bishops[1]=Position::BIN[2] | Position::BIN[5];
    knights[1]=Position::BIN[1] | Position::BIN[6];
    rooks[1]=Position::BIN[0] | Position::BIN[7];
    pawns[1]=Position::BIN[8] | Position::BIN[9] | Position::BIN[10] | Position::BIN[11] | Position::BIN[12] | Position::BIN[13] | Position::BIN[14] | Position::BIN[15];

    kings[0]=Position::BIN[60];
    queens[0]=Position::BIN[59];
    bishops[0]=Position::BIN[58] | Position::BIN[61];
    knights[0]=Position::BIN[57] | Position::BIN[62];
    rooks[0]=Position::BIN[56] | Position::BIN[63];
    pawns[0]=Position::BIN[48] | Position::BIN[49] | Position::BIN[50] | Position::BIN[51] | Position::BIN[52] | Position::BIN[53] | Position::BIN[54] | Position::BIN[55];

    updateBitboardLongs();

    castlingRules=15;
    
    onMove=false;
    
}

/*****************************************************************
 * Initialisiert eine Position durch den übergebenen FEN-String. *
 *****************************************************************/
Position::Position(const string &fen) {
    
    castlingRules=0;
    
    ep=0;
    targetmap=0L;

    aw=0L;
    ab=0L;
    all=0L;
    
    pawns[0]=0L;
    pawns[1]=0L;
    rooks[0]=0L;
    rooks[1]=0L;
    bishops[0]=0L;
    bishops[1]=0L;
    rooks[0]=0L;
    rooks[1]=0L;
    queens[0]=0L;
    queens[1]=0L;
    knights[0]=0L;
    knights[1]=0L;
    kings[0]=0L;
    kings[1]=0L;
    
    size_t found;
    found=fen.find(" ");
    string fengrp0 = fen.substr(0,int(found));
    string rest = fen.substr(int(found)+1,fen.size()-int(found));
    string fengrp1 = rest.substr(0,1);
    rest = rest.substr(2,rest.size()-2);
    found=rest.find(" ");
    string fengrp2 = rest.substr(0,int(found));
    rest = rest.substr(int(found)+1,rest.size()-int(found));
    string fengrp3 = rest;

    // 1. group: figures
    int pos=0;
    for (unsigned int i=0;i<fengrp0.length();i++) {
        if (fengrp0.at(i)=='/') continue;
        else if (fengrp0.at(i)=='1') pos=pos+1;
        else if (fengrp0.at(i)=='2') pos=pos+2;
        else if (fengrp0.at(i)=='3') pos=pos+3;
        else if (fengrp0.at(i)=='4') pos=pos+4;
        else if (fengrp0.at(i)=='5') pos=pos+5;
        else if (fengrp0.at(i)=='6') pos=pos+6;
        else if (fengrp0.at(i)=='7') pos=pos+7;
        else if (fengrp0.at(i)=='8') pos=pos+8;
        else if (fengrp0.at(i)=='k') { kings[1]=Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='q') { queens[1]=queens[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='n') { knights[1]=knights[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='b') { bishops[1]=bishops[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='r') { rooks[1]=rooks[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='p') { pawns[1]=pawns[1]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='K') { kings[0]=Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='Q') { queens[0]=queens[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='N') { knights[0]=knights[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='B') { bishops[0]=bishops[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='R') { rooks[0]=rooks[0]|Position::BIN[pos]; pos++;}
        else if (fengrp0.at(i)=='P') { pawns[0]=pawns[0]|Position::BIN[pos]; pos++;}
    }
    
    // 2. group: color on move, fengrp1
    if (fengrp1=="b")
        onMove=true;
    else
        onMove=false;

    // 3. group: castles
    if (fengrp2.at(0)!='-') {
        for (unsigned int i=0;i<fengrp2.size();i++) {
            if (fengrp2.at(i)=='K')   castlingRules^=1; // white_castling_short=true;
            if (fengrp2.at(i)=='Q')   castlingRules^=2; // white_castling_long=true;
            if (fengrp2.at(i)=='k')   castlingRules^=4; // black_castling_short=true;
            if (fengrp2.at(i)=='q')   castlingRules^=8; // black_castling_long=true;            
        }
    }
    
    // 4. group: en passant
    int tmp_ep=0;
    if (fengrp3.length()==2) {
        if (fengrp3.at(0)=='a' && fengrp3.at(1)=='6') tmp_ep=1;
        if (fengrp3.at(0)=='b' && fengrp3.at(1)=='6') tmp_ep=2;
        if (fengrp3.at(0)=='c' && fengrp3.at(1)=='6') tmp_ep=3;
        if (fengrp3.at(0)=='d' && fengrp3.at(1)=='6') tmp_ep=4;
        if (fengrp3.at(0)=='e' && fengrp3.at(1)=='6') tmp_ep=5;
        if (fengrp3.at(0)=='f' && fengrp3.at(1)=='6') tmp_ep=6;
        if (fengrp3.at(0)=='g' && fengrp3.at(1)=='6') tmp_ep=7;
        if (fengrp3.at(0)=='h' && fengrp3.at(1)=='6') tmp_ep=8;
        if (fengrp3.at(0)=='a' && fengrp3.at(1)=='3') tmp_ep=9;
        if (fengrp3.at(0)=='b' && fengrp3.at(1)=='3') tmp_ep=10;
        if (fengrp3.at(0)=='c' && fengrp3.at(1)=='3') tmp_ep=11;
        if (fengrp3.at(0)=='d' && fengrp3.at(1)=='3') tmp_ep=12;
        if (fengrp3.at(0)=='e' && fengrp3.at(1)=='3') tmp_ep=13;
        if (fengrp3.at(0)=='f' && fengrp3.at(1)=='3') tmp_ep=14;
        if (fengrp3.at(0)=='g' && fengrp3.at(1)=='3') tmp_ep=15;
        if (fengrp3.at(0)=='h' && fengrp3.at(1)=='3') tmp_ep=16;
    }
    ep=tmp_ep;
    
    updateBitboardLongs();
}

/**************************************************************
 * Wird ein UCI-Command oder ein Cmd des Benutzers            *
 * nicht verstanden, so wird diese Funktion aufgerufen,       *
 * um zu prüfen, ob es sich bei dem Cmd um einen zur Position *
 * gültigen Zug handelt (siehe main.cpp)                      *
 **************************************************************/
int Position::getMMChilds(string moveArray[], int moveIndex[]) {
    cd=0;
    doBackup(cd);
    calc_moves(true);
    int cntMoves=0;
    for (int i=0;i<mvSize[cd];i++) {
        move_figure(mvD[cd][i][0],mvD[cd][i][1],mvP[cd*256+i]);
        onMove=!onMove;
        bool catchKing=opponent_target_bitboard2(kings[onMove]);
        onMove=!onMove;
        if (!catchKing) {
            moveArray[cntMoves]=getMoveString(i);
            moveIndex[cntMoves]=i;
            cntMoves++;
        }
        undoMove(cd);
    }
    return cntMoves;
}

/**********************************************************************
 * Gibt zur gegebenen Position und Tiefe die Anzahl der Nodes         *
 * für jeden möglichen Zug an.                                        *
 * divide [n] ... Anzahl der Blätter/Nodes in Tiefe n                 *
 * divide2 [n] ... Summe aller Nodes bis inkl. Tiefe n                *
 * Diese Funktionen sind speziel zum Debugging des Zug-Generators     *
 * wertvoll. An dieser Stelle einen großen Dank an                    *
 * Roman Hartmann (http://www.rocechess.ch/roce.html),dessen Programm *
 * ich selbst verwendet hatte um meine Engine zu debuggen             *
 **********************************************************************/
void Position::getDivide(int color, int curr_depth, const int &depth, const bool &modus, long &sum, long &total) {
    cd = curr_depth-1;
    calc_moves(true);  
    doBackup(cd);
    if (curr_depth<depth+1) {
        for (int i=0;i<mvSize[cd];i++) {
            move_figure(mvD[cd][i][0],mvD[cd][i][1],mvP[cd*256+i]);
            if (curr_depth==1)
                sum=0;
            onMove=!onMove;
            bool catchKing=opponent_target_bitboard2(kings[onMove]);
            onMove=!onMove;
            if (!catchKing) {
                getDivide(abs(color-1), (curr_depth+1),depth, modus, sum, total);
                cd--;
                if (curr_depth==1)
                    cout << getMoveString(i) << " " << sum << "     " << getFen() << endl;                                
            } else {
                if (    (curr_depth<=depth && modus==true) || (curr_depth==depth && modus==false)   ) {
                    sum--;
                    total--;
                }                
            }
            undoMove(cd);
        }
    }
    if (    (curr_depth<=depth && modus==true) || (curr_depth==depth && modus==false)   ) {
        sum+=mvSize[cd];
        total+=mvSize[cd];
    }
}

/*************************************************************************
 * Führt einen Performance-Test zur aktuellen Position aus und           *
 * liefert ausführliche Daten zu Captures, En-Passants, Promotions, etc. *
 *************************************************************************/
void Position::getPerftInfo(int curr_depth, int depth, Perftres *perft) {
    
    cd = curr_depth-1;
    doBackup(cd);
    calc_moves(true);
    
    if (curr_depth<depth+1) {
        for (int i=0;i<mvSize[cd];i++) {
            int ret=move_figure(mvD[cd][i][0],mvD[cd][i][1],mvP[cd*256+i]);
            
            onMove=!onMove;
            bool catchKing=opponent_target_bitboard2(kings[onMove]);
            onMove=!onMove;
            
            if (!catchKing) {
                getPerftInfo((curr_depth+1),depth, perft);
                cd--;

                // count captures
                if (ret)
                    perft->perftres[1+(curr_depth-1)*7]++;                    

                // count ep-moves
                // if (mvD[cd][i][2]!=0L) { // <-- funktioniert nicht mehr, kill / [2] nicht mehr vorhanden
                if (mvD[cd][i][1]!=0L) {                
                    perft->perftres[4+(curr_depth-1)*7]++;
                  //  perft->perftres[1+(curr_depth-1)*7]++;
                } 

                // count promotions
                if (mvP[cd*256+i]>0)
                    perft->perftres[5+(curr_depth-1)*7]++;                                
                
                // count castlings
                if ((mvD[cd][i][1]&kings[!onMove])!=0L && (mvD[cd][i][0]<<2==mvD[cd][i][1] || mvD[cd][i][0]>>2==mvD[cd][i][1]))
                    perft->perftres[6+(curr_depth-1)*7]++;                     
            
            } else
                perft->perftres[(curr_depth-1)*7]--;                
            undoMove(cd);
            
        }
    }    

    perft->perftres[(curr_depth-1)*7]+=mvSize[cd];
    
    if (perft!=NULL) {
        /*        


        // count checks
        Position tmp_pos3(*tmp_pos);
        if (opponent_target_bitboard(tmp_pos3,tmp_pos->kings[abs(color-1)],abs(color-1)))
            perft->perftres[2+(curr_depth-1)*7]++;                

        */
    }    
}

/*****************************************************************************
 * Sortiert das Array mit den möglichen Zügen einer Position.                *
 * Züge mit Captures werden bevorzugt.                                       *
 * Wert eines Zuges ist abhängig welcher Figuren-Typ welchen gegnerischen    *
 * Figuren-Typ schlägt, dh. Bauer schlägt Dame ist ein wertvollerer Zug als  *
 * Dame schlägt Bauer, usw.                                                  *
 * Die Sortierung dient dazu, beim NegaMax früher bzw. bessere Cutoffs       *
 * zu bekommen, somit sollte sich die Zeit der Zugsuche deutlich reduzieren. *
 *****************************************************************************/
void Position::sortMoves(unsigned short int sortedIndizes[], int cd, int bestmoveHash) {
    unsigned short int cntMoves=mvSize[cd];
    short int moveValues[cntMoves];

    /*************************************************************
     * Jedem Zug einen Score zuweisen. Züge ohne Captures        *
     * werden mit dem Wert -2000 bewertet. Züge mit Captures mit *
     * einem Wert abhängig davon, welcher Figurentyp welchen     *
     * gegnerischen Figurentyp schlägt.                          *
     *************************************************************/
    for (int i=0;i<cntMoves;i++) {
        
        if ((mvD[cd][i][1]&all)==0L) {
            moveValues[i]=-2000;
            continue;
        }
        
        unsigned short int fromValue, toValue;
        
        if ((kings[onMove]&mvD[cd][i][0])!=0L)          fromValue=1000;
        else if ((queens[onMove]&mvD[cd][i][0])!=0L)    fromValue=9;
        else if ((rooks[onMove]&mvD[cd][i][0])!=0L)     fromValue=5;
        else if ((bishops[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else if ((knights[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else /* if ((pawns[onMove]&mvD[cd][i][0])!=0L) */    fromValue=1;

        bool notOnMove=!onMove;
        if ((kings[notOnMove]&mvD[cd][i][1])!=0L)          toValue=1000;
        else if ((queens[notOnMove]&mvD[cd][i][1])!=0L)    toValue=9;
        else if ((rooks[notOnMove]&mvD[cd][i][1])!=0L)     toValue=5;
        else if ((bishops[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else if ((knights[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else /* if ((pawns[notOnMove]&mvD[cd][i][1])!=0L) */     toValue=1;        

        moveValues[i]=toValue-fromValue;
        
    }
    
    /**************************************************
     * Index des besten Zuges aus dem Hash markieren. *
     **************************************************/
    if (bestmoveHash>=0 && bestmoveHash<cntMoves)
        moveValues[bestmoveHash]=10000;
    
    /********************
     * Werte sortieren. *
     ********************/
    
    for (int i=0;i<cntMoves;i++)
        sortedIndizes[i]=i;
    
    Heap::buildHeap(sortedIndizes,moveValues,cntMoves);    
}

unsigned short int Position::sortMovesQuiesce(unsigned short int sortedIndizes[], int cd) {
    unsigned short int cntMoves=mvSize[cd];
    short int moveValues[cntMoves];
    unsigned short int cntCaptures=0;

    // calculate a value for every move
    for (int i=0;i<cntMoves;i++) {
        
        if ((mvD[cd][i][1]&all)==0L) {
            moveValues[i]=-2000;
            continue;
        }
        cntCaptures++;
        
        unsigned short int fromValue, toValue;
        
        if ((kings[onMove]&mvD[cd][i][0])!=0L)          fromValue=1000;
        else if ((queens[onMove]&mvD[cd][i][0])!=0L)    fromValue=9;
        else if ((rooks[onMove]&mvD[cd][i][0])!=0L)     fromValue=5;
        else if ((bishops[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else if ((knights[onMove]&mvD[cd][i][0])!=0L)   fromValue=3;
        else /* if ((pawns[onMove]&mvD[cd][i][0])!=0L) */    fromValue=1;

        bool notOnMove=!onMove;
        if ((kings[notOnMove]&mvD[cd][i][1])!=0L)          toValue=1000;
        else if ((queens[notOnMove]&mvD[cd][i][1])!=0L)    toValue=9;
        else if ((rooks[notOnMove]&mvD[cd][i][1])!=0L)     toValue=5;
        else if ((bishops[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else if ((knights[notOnMove]&mvD[cd][i][1])!=0L)   toValue=3;
        else /* if ((pawns[notOnMove]&mvD[cd][i][1])!=0L) */     toValue=1;        

        moveValues[i]=toValue-fromValue;
        
    }
    
    for (int i=0;i<cntMoves;i++)
        sortedIndizes[i]=i;
    
    Heap::buildHeap(sortedIndizes,moveValues,cntMoves);
    
    return cntCaptures;
}

/**************************************************************
 * NegaMax ist die Kernfunktion zur Suche nach dem besten Zug *
 * Aktuell implementiert ist dabei:                           *
 *   - eine einfache Zugsortierung                            *
 *   - Transposition/Hash-Table                               *
 *   - Quiescence Search                                      *
 **************************************************************/
int Position::negaMax(bool side2move, int alpha, int beta, int curr_depth, int depth, int &sum, bool isPv, bool canNull) {

    sum++;

    
    if (cd>2 && sum%1000==0)
        timer.checkTimeout();
    
    if (cd>2 && timer.isTimeout())
        return 0;
    
    
    cd = curr_depth;
    
    unsigned long hashKey = hashTable.getHashKey(this);
    moveList.setMoveAt(curr_depth,hashKey,"",0);
    
    /*********************************************************************************
     * Überprüfe ob die aktuelle Position bereits im Spiel- bzw. Suchverlauf vorkam. *
     * Ggf. wird ein Score für ein Draw bzw. ein entsprechender Contempt-Faktor      *
     * zurück gegeben.                                                               *
     *********************************************************************************/
    if (cd>0 && moveList.isRepetition(cd)) {
        int contemptOpening=-50;
        int contemptMiddlegame=-25;
        int contempt=0;
        if (moveList.getNumHalfMoves()<20)
            contempt=contemptOpening;
        else if (!isEndgame())
            contempt=contemptMiddlegame;
        if (cd%2==1)
            contempt=-contempt;
        return contempt;
    }
    
    /****************************************************************************
     * Suche in der HashTable nach der aktuellen Position und prüfe ggf. ob der *
     * gespeicherte Score verwendet werden kann.                                *
     ****************************************************************************/
    int score2;
    int bestmoveHash=-1;
    if (cd>0 && hashTable.contains(hashKey,depth-cd,&score2,alpha,beta,&bestmoveHash)) // cd>1 oder cd>0 ???
        return score2;
    
    /*********************************************************************
     * Wenn der König im Schach steht, soll die Suchtiefe erhöht werden. *
     *********************************************************************/
    bool inCheck = false;

//    if (curr_depth==0 || curr_depth!=depth) { // && depth-curr_depth>1) {
//    if (depth<2) {

        inCheck=opponent_target_bitboard2(kings[onMove]);
        /*
        if (inCheck)
                depth++;
         * */
//    }

    
    /*********************
     * Null-Move-Pruning *
     *********************/
      
    int score=0;  

//    if (depth-curr_depth>3 && canNull && evaluate(onMove)>beta && !isEndgame() && !inCheck) {
//        
////        makeNullMove();
//        
//        int r=2;
//        if (depth-curr_depth>6)
//            r=3;
//        
//        score = -negaMax(!side2move,-beta,-beta+1,(curr_depth+1),depth-r-1,sum,false,false);
//        cd=curr_depth;
//        
////        undoNullMove();
//        
//        if (timer.isTimeout())
//            return 0;
//        if (score>=beta) {// save hash here
//            // logger.log("nullmove cutoff");
//            return beta;
//        }
//    }
    
    /********************************************************************************
     * Sofern wir in der maximalen Tiefe angekommen sind, soll die Quiescence-Suche *
     * aufgerufen werden.                                                           *
     ********************************************************************************/
    if (curr_depth==depth) {
        // hashScore hier speichern
        return quiesce(side2move,alpha,beta,curr_depth,depth+2);        
    }

    /************************************************************************
     * Züge der aktuellen Position generieren und zwecks Pruning sortieren. *
     ************************************************************************/
//    calc_moves(curr_depth==0); // true nur bei curr_depth==1
    calc_moves(true);
    unsigned short int sortedIndizes[256];
    sortMoves(sortedIndizes,cd,bestmoveHash);
    doBackup(cd);

    unsigned int nodeType=3;
    int maxWert=alpha;
    bool raisedAlpha=false;
    
    int i;
    int bmIndex=-1;

    for (i=mvSize[cd]-1;i>=0;i--) {
        bool ret = move_figure(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],mvP[cd*256+sortedIndizes[i]]);
        
        /***********************************************************************************
         * Prüft ob der Zug legal ist. Das ist dann der Fall, wenn der gegnerische Spieler *
         * anschließend nicht den König schlagen kann. Da diese Überprüfung sehr aufwendig *
         * ist, wird sie nur in der untersten Ebene durchgeführt, damit die Engine nur     *
         * einen legalen Zug als besten Zug wählen kann.                                   *
         ***********************************************************************************/
//        if (curr_depth==0) {
            onMove=!onMove;
            bool catchKing=opponent_target_bitboard2(kings[onMove]);
            onMove=!onMove;
            if (catchKing) {
                undoMove(cd);
                continue;
            }
//        }
    
        
//        moveList.setMoveAt(curr_depth,hashKey,getMoveString(sortedIndizes[i]),curr_depth);
            
        /*******************
         * Check-Extension *
         *******************/
        unsigned short int ext=0;
//
//        if (depth>1 && opponent_target_bitboard2(kings[onMove])) {
////            logger.log("gamelist: "+moveList.getGameList(curr_depth));
//            ext=1;
//        }

        /*****************************
         * Principal-Variation-Suche *
         *****************************/
//        if (depth==1) {
//            if (i==mvSize[cd]-1 || -negaMax(!side2move,-maxWert-1,-maxWert,(curr_depth+1),depth+ext,sum,false,true)>maxWert) // {
//                score = -negaMax(!side2move,-beta,-maxWert,(curr_depth+1),depth+ext,sum,true,true);
//        } else {
//            // score = -negaMax(!side2move,-beta,-maxWert,(curr_depth+1),depth,sum,true);
//            if (!isPv || (depth-curr_depth)<3)
//                score = -negaMax(!side2move,-beta,-maxWert,(curr_depth+1),depth+ext,sum,false,true);
//            else {
//                if (!raisedAlpha)
//                    score = -negaMax(!side2move,-beta,-maxWert,(curr_depth+1),depth+ext,sum,true,true);
//                else {
//                    if (-negaMax(!side2move,-maxWert-1,-maxWert,(curr_depth+1),depth+ext,sum,false,true)>maxWert)
//                        score = -negaMax(!side2move,-beta,-maxWert,(curr_depth+1),depth+ext,sum,true,true);
//                }
//            }
//        }

        score = -negaMax(!side2move,-beta,-maxWert,(curr_depth+1),depth+ext,sum,true,true);        
        cd=curr_depth;
        
        /*****************************************************
         * Zug rückgängig machen und den Score des Zuges     *
         * ggf. speichern, bzw. überprüfen ob die restlichen *
         * Züge noch durchgeführt werden müssen.             *
         *****************************************************/
        undoMove(cd);

        if (score==0 && timer.isTimeout())
            return 0;
        
        if (score > maxWert) {
            raisedAlpha=true;
            bmIndex=sortedIndizes[i];
            maxWert = score;
            if (curr_depth==0)
                bestMoveIndex=sortedIndizes[i];
            if (maxWert>=beta) {
                nodeType=2;
                maxWert=beta;
                break;
            }
            nodeType=1;
        }
    }

    /**************************************************
     * Den berechneten Score in der HashTable und den *
     * besten Zug in der GameHistory speichern.       *
     **************************************************/
    hashTable.insert(hashKey,maxWert,nodeType,bmIndex,depth-cd);
    moveList.setMoveAt(curr_depth,hashKey,getMoveString(bmIndex),(depth-curr_depth));
    return maxWert;
}

/******************************************************
 * Gibt zurück, ob es sich bei der aktuellen Position *
 * um eine Position im Endspiel handelt oder nicht.   *
 * Ein Endspiel ist gegeben, wenn:                    *
 * - keine der Damen mehr im Spiel sind,              *
 * - einer der Spieler weniger als 8 Figuren hat      *
 ******************************************************/
bool Position::isEndgame() {
    
    if (queens[0]==0UL && queens[1]==0UL)
        return true;
    
    if (popCount(aw)<8 || popCount(ab)<8)
        return true;
    
    return false;
}

/************************************************************************
 * Die Quiescence-Suche wird an jedem Leaf-Node des NegaMax aufgerufen. *
 * Betrachtet werden hierbei nur noch jene Züge die Figuren schlagen.   *
 * Im Idealfall wird die Suche solange fortgesetzt, bis eine Position   *
 * erreicht wurde, in der keine Figur geschlagen werden kann oder nicht *
 * mehr tiefer evaluiert werden muss.                                   *
 * StandingPat wurde bereits implementiert. Es fehlt allerdings noch    *
 * die Implementierung der Static Exchange Evaluation bzw. auch         *
 * Delta-Pruning.                                                       *
 ************************************************************************/
int Position::quiesce(bool side2move, int alpha, int beta, int curr_depth, int max_depth) {
       
    cd = curr_depth;
    
    if (curr_depth==max_depth)
        return evaluate(side2move);    
    
    int standPat = evaluate(side2move);
    if (standPat>=beta)
        return beta;
    if (alpha<standPat)
        alpha=standPat;
   
    calc_moves(false);
    unsigned short int sortedIndizes[256];
    unsigned short int cntCaptures=sortMovesQuiesce(sortedIndizes,cd);
    doBackup(cd);
    
    int maxWert=alpha;
    
    for (int i=mvSize[cd]-1;i>=(mvSize[cd]-cntCaptures);i--) {
        
        move_figure(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],mvP[cd*256+sortedIndizes[i]]);
        
        int score = -quiesce(!side2move,-beta,-maxWert,(curr_depth+1),max_depth);
        cd--;
        undoMove(cd);

        if (score > maxWert) {
            maxWert = score;
            if (maxWert>=beta) {
                maxWert=beta;
                break;
            }
        }
    }
    if (!cntCaptures)
        maxWert=evaluate(side2move);
    return maxWert;
}

/***********************************************************************
 * Die evaluate-Funktion bewertet eine gegebene Position               *
 * nach verschiedenen heuristischen Methoden für jeweils beide Spieler *
 * Die Werte der Spieler zu jeder Methode werden subtrahiert,          *
 * nach der Priorität gewichtet und zu einem Gesamt-Score              *
 * zusammengefasst.                                                    *
 ***********************************************************************/
int Position::evaluate(bool side2move) {
    
    /***************************
     * berechne material score *
     * *************************/
    
    int materialScore0, materialScore1;

    materialScore0  =   100 *   popCount(pawns[0]) +
                        275 *   popCount(knights[0]) +
                        325 *   popCount(bishops[0]) +
                        465 *   popCount(rooks[0]) +
                        900 *   popCount(queens[0]) +
                        50000 *  popCount(kings[0]);    
    
    materialScore1  =   100 *   popCount(pawns[1]) +
                        275 *   popCount(knights[1]) +
                        325 *   popCount(bishops[1]) +
                        465 *   popCount(rooks[1]) +
                        900 *   popCount(queens[1]) +
                        50000 *  popCount(kings[1]);        
/*  
    materialScore0  =   100 *   popCount(pawns[0]) +
                        320 *   popCount(knights[0]) +
                        330 *   popCount(bishops[0]) +
                        500 *   popCount(rooks[0]) +
                        900 *   popCount(queens[0]) +
                        50000 *  popCount(kings[0]);
    
    materialScore1  =   100 *   popCount(pawns[1]) +
                        320 *   popCount(knights[1]) +
                        330 *   popCount(bishops[1]) +
                        500 *   popCount(rooks[1]) +
                        900 *   popCount(queens[1]) +
                        50000 *  popCount(kings[1]);    
*/  
    
    
    /************************************************************************
     * berechne targetmaps und mobility score                               *
     * targetmaps werden nachfolgend zb. für den capture score benötigt     *
     * zur berechnung des mobility score wird die gesamte zugliste benötigt *
     * daher werden diese beiden elemente hier zusammen berechnet           *
     * **********************************************************************/

//    long targetmap0, targetmap1;
//    int mobilityScore0 = 0, mobilityScore1 = 0;
//    // int moveSize0, moveSize1;
//    int mobilityWeight = 20;
//    
//    
//    /***************
//     * weiß am zug *
//     * *************/
//
//    onMove=false;
//    targetmap=0L;
//    calc_moves(false); // true=with castling
//    // moveSize0=mvSize[cd];
//    targetmap0=targetmap;
//    
//    for (unsigned int i=0;i<mvSize[cd];i++) {
//        // if ((mvD[cd][i][0]&kings[0])!=0L) mobilityScore0+=20;
//        // if ((mvD[cd][i][0]&queens[0])!=0L) mobilityScore0+=9;
//        if ((mvD[cd][i][0]&rooks[0])!=0L) mobilityScore0+=5;
//        else if ((mvD[cd][i][0]&bishops[0])!=0L) mobilityScore0+=3;
//        else if ((mvD[cd][i][0]&knights[0])!=0L) mobilityScore0+=3;
//        else if ((mvD[cd][i][0]&pawns[0])!=0L) mobilityScore0+=1;
//    }
//
//    /******************
//     * schwarz am zug *
//     * ****************/    
//    
//    onMove=true;
//    targetmap=0L;
//    calc_moves(false); // true=with castling
//    // moveSize1=mvSize[cd];
//    targetmap1=targetmap;    
//
//    for (unsigned int i=0;i<mvSize[cd];i++) {
//        // if ((mvD[cd][i][0]&kings[1])!=0L) mobilityScore1+=20;
//        // if ((mvD[cd][i][0]&queens[1])!=0L) mobilityScore1+=9;
//        if ((mvD[cd][i][0]&rooks[1])!=0L) mobilityScore1+=5;
//        else if ((mvD[cd][i][0]&bishops[1])!=0L) mobilityScore1+=3;
//        else if ((mvD[cd][i][0]&knights[1])!=0L) mobilityScore1+=3;
//        else if ((mvD[cd][i][0]&pawns[1])!=0L) mobilityScore1+=1;
//    }    

    // mobilityScore0/=moveSize0;
    // mobilityScore1/=moveSize1;
    
    /*************************
     * Anzahl möglicher Züge *
     *************************/
    
    int legalMoves0, legalMoves1;
    int legalMovesWeight=70;
    
//    legalMoves0 = popCount(targetmap0);
//    legalMoves1 = popCount(targetmap1);
    
    
    /****************************************************************
     * berechne capture score                                       *
     * in abhängigkeit der figur-typen die geschlagen werden können *
     * **************************************************************/    
    
    // sollte captureScore bei idealer quiescence-suche nicht 0 sein?
    
    int captureScore0 = 0, captureScore1 = 0;
    int captureWeight = 80;
    
//    if ((targetmap0&kings[1])!=0)       captureScore0+=20;
//    if ((targetmap0&queens[1])!=0)      captureScore0+=9*popCount(targetmap0&queens[1]);
//    if ((targetmap0&rooks[1])!=0)       captureScore0+=5*popCount(targetmap0&rooks[1]);
//    if ((targetmap0&bishops[1])!=0)     captureScore0+=3*popCount(targetmap0&bishops[1]);
//    if ((targetmap0&knights[1])!=0)     captureScore0+=3*popCount(targetmap0&knights[1]);
//    if ((targetmap0&pawns[1])!=0)       captureScore0+=popCount(targetmap0&pawns[1]);
//
//    if ((targetmap1&kings[0])!=0)       captureScore1+=20;
//    if ((targetmap1&queens[0])!=0)      captureScore1+=9*popCount(targetmap1&queens[0]);
//    if ((targetmap1&rooks[0])!=0)       captureScore1+=5*popCount(targetmap1&rooks[0]);
//    if ((targetmap1&bishops[0])!=0)     captureScore1+=3*popCount(targetmap1&bishops[0]);
//    if ((targetmap1&knights[0])!=0)     captureScore1+=3*popCount(targetmap1&knights[0]);
//    if ((targetmap1&pawns[0])!=0)       captureScore1+=popCount(targetmap1&pawns[0]);    
//    
    
    /*********************************************************************
     * Piece-Square Tabellen                                             *
     * Berechnet Punkt-Gewinn/Abzug für Figuren an bestimmten Positionen *
     *********************************************************************/
    
    int pieceSquare0=0, pieceSquare1=0;
    int pieceSquareWeight=1;
    
    int pawnPieceSquare[2][64] = {{
/*
        0,   0,   0,   0,   0,   0,   0,   0,                        
        0,   0,   0, -40, -40,   0,   0,   0,
        1,   2,   3, -10, -10,   3,   2,   1,        
        2,   4,   6,   8,   8,   6,   4,   2,
        3,   6,   9,  12,  12,   9,   6,   3,                      
        4,   8,  12,  16,  16,  12,   8,   4,
        5,  10,  15,  20,  20,  15,  10,   5,                      
        0,   0,   0,   0,   0,   0,   0,   0
 * */        
 0,  0,  0,  0,  0,  0,  0,  0,
50, 50, 50, 50, 50, 50, 50, 50,
10, 10, 20, 30, 30, 20, 10, 10,
 5,  5, 10, 25, 25, 10,  5,  5,
 0,  0,  0, 20, 20,  0,  0,  0,
 5, -5,-10,  0,  0,-10, -5,  5,
 5, 10, 10,-20,-20, 10, 10,  5,
 0,  0,  0,  0,  0,  0,  0,  0        
    },{
 0,  0,  0,  0,  0,  0,  0,  0,        
 5, 10, 10,-20,-20, 10, 10,  5,
 5, -5,-10,  0,  0,-10, -5,  5,         
 0,  0,  0, 20, 20,  0,  0,  0,
 5,  5, 10, 25, 25, 10,  5,  5,
10, 10, 20, 30, 30, 20, 10, 10,         
50, 50, 50, 50, 50, 50, 50, 50,         
 0,  0,  0,  0,  0,  0,  0,  0         
        /*
        0,   0,   0,   0,   0,   0,   0,   0,
        5,  10,  15,  20,  20,  15,  10,   5,
        4,   8,  12,  16,  16,  12,   8,   4,
        3,   6,   9,  12,  12,   9,   6,   3,
        2,   4,   6,   8,   8,   6,   4,   2,
        1,   2,   3, -10, -10,   3,   2,   1,
        0,   0,   0, -40, -40,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0
         * */
    }};

    int knightPieceSquare[2][64] = {{
/*
            -10, -10, -10, -10, -10, -10, -10, -10,
            -10,   0,   0,   0,   0,   0,   0, -10,
            -10,   0,   5,   5,   5,   5,   0, -10,
            -10,   0,   5,  10,  10,   5,   0, -10,
            -10,   0,   5,  10,  10,   5,   0, -10,
            -10,   0,   5,   5,   5,   5,   0, -10,
            -10,   0,   0,   0,   0,   0,   0, -10,
            -10, -30, -10, -10, -10, -10, -30, -10
 * */        
-50,-40,-30,-30,-30,-30,-40,-50,
-40,-20,  0,  0,  0,  0,-20,-40,
-30,  0, 10, 15, 15, 10,  0,-30,
-30,  5, 15, 20, 20, 15,  5,-30,
-30,  0, 15, 20, 20, 15,  0,-30,
-30,  5, 10, 15, 15, 10,  5,-30,
-40,-20,  0,  5,  5,  0,-20,-40,
-50,-40,-30,-30,-30,-30,-40,-50        
    },{
-50,-40,-30,-30,-30,-30,-40,-50,
-40,-20,  0,  5,  5,  0,-20,-40,
-30,  5, 10, 15, 15, 10,  5,-30,
-30,  0, 15, 20, 20, 15,  0,-30,
-30,  5, 15, 20, 20, 15,  5,-30,
-30,  0, 10, 15, 15, 10,  0,-30,
-40,-20,  0,  0,  0,  0,-20,-40,
-50,-40,-30,-30,-30,-30,-40,-50        
/*
            -10, -30, -10, -10, -10, -10, -30, -10,
            -10,   0,   0,   0,   0,   0,   0, -10,
            -10,   0,   5,   5,   5,   5,   0, -10,
            -10,   0,   5,  10,  10,   5,   0, -10,
            -10,   0,   5,  10,  10,   5,   0, -10,
            -10,   0,   5,   5,   5,   5,   0, -10,
            -10,   0,   0,   0,   0,   0,   0, -10,
            -10, -10, -10, -10, -10, -10, -10, -10        
 * */        
    }};

    int bishopPieceSquare[2][64] = {{
/*
            -10, -10, -10, -10, -10, -10, -10, -10,
            -10,   0,   0,   0,   0,   0,   0, -10,
            -10,   0,   5,   5,   5,   5,   0, -10,
            -10,   0,   5,  10,  10,   5,   0, -10,
            -10,   0,   5,  10,  10,   5,   0, -10,
            -10,   0,   5,   5,   5,   5,   0, -10,
            -10,   0,   0,   0,   0,   0,   0, -10,
            -10, -10, -20, -10, -10, -20, -10, -10
 * */        
-20,-10,-10,-10,-10,-10,-10,-20,
-10,  0,  0,  0,  0,  0,  0,-10,
-10,  0,  5, 10, 10,  5,  0,-10,
-10,  5,  5, 10, 10,  5,  5,-10,
-10,  0, 10, 10, 10, 10,  0,-10,
-10, 10, 10, 10, 10, 10, 10,-10,
-10,  5,  0,  0,  0,  0,  5,-10,
-20,-10,-10,-10,-10,-10,-10,-20        
    },{
-20,-10,-10,-10,-10,-10,-10,-20,
-10,  5,  0,  0,  0,  0,  5,-10,
-10, 10, 10, 10, 10, 10, 10,-10,
-10,  0, 10, 10, 10, 10,  0,-10,
-10,  5,  5, 10, 10,  5,  5,-10,
-10,  0,  5, 10, 10,  5,  0,-10,
-10,  0,  0,  0,  0,  0,  0,-10,
-20,-10,-10,-10,-10,-10,-10,-20        
        /*
            -10, -10, -20, -10, -10, -20, -10, -10,
            -10,   0,   0,   0,   0,   0,   0, -10,
            -10,   0,   5,   5,   5,   5,   0, -10,
            -10,   0,   5,  10,  10,   5,   0, -10,
            -10,   0,   5,  10,  10,   5,   0, -10,
            -10,   0,   5,   5,   5,   5,   0, -10,
            -10,   0,   0,   0,   0,   0,   0, -10,
            -10, -10, -10, -10, -10, -10, -10, -10
         * */
    }};
    
    int rookPieceSquare[2][64] = {{
              0,  0,  0,  0,  0,  0,  0,  0,
              5, 10, 10, 10, 10, 10, 10,  5,
             -5,  0,  0,  0,  0,  0,  0, -5,
             -5,  0,  0,  0,  0,  0,  0, -5,
             -5,  0,  0,  0,  0,  0,  0, -5,
             -5,  0,  0,  0,  0,  0,  0, -5,
             -5,  0,  0,  0,  0,  0,  0, -5,
              0,  0,  0,  5,  5,  0,  0,  0        
        
    },{
              0,  0,  0,  5,  5,  0,  0,  0        
             -5,  0,  0,  0,  0,  0,  0, -5,
             -5,  0,  0,  0,  0,  0,  0, -5,
             -5,  0,  0,  0,  0,  0,  0, -5,
             -5,  0,  0,  0,  0,  0,  0, -5,
             -5,  0,  0,  0,  0,  0,  0, -5,
              5, 10, 10, 10, 10, 10, 10,  5,
              0,  0,  0,  0,  0,  0,  0,  0,
    }};
    
    int queenPieceSquare[2][64] = {{    
            -20,-10,-10, -5, -5,-10,-10,-20,
            -10,  0,  0,  0,  0,  0,  0,-10,
            -10,  0,  5,  5,  5,  5,  0,-10,
             -5,  0,  5,  5,  5,  5,  0, -5,
              0,  0,  5,  5,  5,  5,  0, -5,
            -10,  5,  5,  5,  5,  5,  0,-10,
            -10,  0,  5,  0,  0,  0,  0,-10,
            -20,-10,-10, -5, -5,-10,-10,-20        
    },{
            -20,-10,-10, -5, -5,-10,-10,-20,
            -10,  0,  5,  0,  0,  0,  0,-10,
            -10,  5,  5,  5,  5,  5,  0,-10,
              0,  0,  5,  5,  5,  5,  0, -5,
             -5,  0,  5,  5,  5,  5,  0, -5,        
            -10,  0,  5,  5,  5,  5,  0,-10,
            -10,  0,  0,  0,  0,  0,  0,-10,
            -20,-10,-10, -5, -5,-10,-10,-20        
    }};
    
    int kingPieceSquare[2][64] = {{
-30,-40,-40,-50,-50,-40,-40,-30,
-30,-40,-40,-50,-50,-40,-40,-30,
-30,-40,-40,-50,-50,-40,-40,-30,
-30,-40,-40,-50,-50,-40,-40,-30,
-20,-30,-30,-40,-40,-30,-30,-20,
-10,-20,-20,-20,-20,-20,-20,-10,
 20, 20,  0,  0,  0,  0, 20, 20,
 20, 30, 10,  0,  0, 10, 30, 20        
    },{
 20, 30, 10,  0,  0, 10, 30, 20,
 20, 20,  0,  0,  0,  0, 20, 20,         
-10,-20,-20,-20,-20,-20,-20,-10,
-20,-30,-30,-40,-40,-30,-30,-20,
-30,-40,-40,-50,-50,-40,-40,-30,
-30,-40,-40,-50,-50,-40,-40,-30,
-30,-40,-40,-50,-50,-40,-40,-30,
-30,-40,-40,-50,-50,-40,-40,-30         
    }};
    
    /********
     * weiß *
     ********/
    
    unsigned long bitboard = pawns[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=pawnPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = knights[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=knightPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = bishops[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=bishopPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = rooks[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=rookPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = queens[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=queenPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }    

    /*
    bitboard = kings[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=kingPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }        
     * */
    
    /***********
     * schwarz *
     ***********/
    
    bitboard = pawns[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=pawnPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = knights[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=knightPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = bishops[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=bishopPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }    

    bitboard = rooks[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=rookPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }    

    bitboard = queens[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=queenPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }        
    /*
    bitboard = kings[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=kingPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }            
     * */
    
    /*****************************************************************************************************************
     * berechne gesamt score                                                                                         *
     * die gewichtung der einzelnen werte sollte später automatisch durch ein skript ermittelt werden.               *
     * mit entsprechend vielen tournaments sollte so eine verbesserung bzw. verschlechterung der ELO erkennbar sein. *
     * die dadurch gewonnen daten können auch für weitere analysen verwendet werden um so zu ermitteln,              *
     * ob es werte gibt, die eigentlich gar nicht zwingend berechnet werden müssten, da sie ohnehin                  *
     * mit anderen werten korrelieren                                                                                *
     * ***************************************************************************************************************/    
    
    int score = (materialScore0 - materialScore1) +
                // (mobilityScore0 - mobilityScore1) * mobilityWeight +
//                (legalMoves0 - legalMoves1) * legalMovesWeight +
//                (captureScore0 - captureScore1) * captureWeight + 
                (pieceSquare0 - pieceSquare1) * pieceSquareWeight;
                                                   
   // cout << (legalMoves0 - legalMoves1) * legalMovesWeight << endl;
    
    /******************
     * we're ready :) *
     * ****************/        
    
    if (side2move)
        score*=-1;
    
//    evalstat.logFenScore(getFen(),      score, 
//                                        (materialScore0 - materialScore1),
//                                        materialScore0, materialScore1,
//                                        (mobilityScore0 - mobilityScore1) * mobilityWeight,                                        
//                                        mobilityScore0, mobilityScore1,
//                                        (legalMoves0 - legalMoves1) * legalMovesWeight,
//                                        legalMoves0, legalMoves1,
//                                        (captureScore0 - captureScore1) * captureWeight,
//                                        captureScore0, captureScore1,
//                                        (pieceSquare0 - pieceSquare1) * pieceSquareWeight,
//                                        pieceSquare0, pieceSquare1);
        
    return score;
}

/************************************
 * Speichert die aktuelle Position. *
 ************************************/
void Position::doBackup(int depth) {
    myBackup[depth].aw=aw;
    myBackup[depth].ab=ab;
    myBackup[depth].all=all;
    myBackup[depth].kings[0]=kings[0];
    myBackup[depth].kings[1]=kings[1];
    myBackup[depth].pawns[0]=pawns[0];
    myBackup[depth].pawns[1]=pawns[1];
    myBackup[depth].bishops[0]=bishops[0];
    myBackup[depth].bishops[1]=bishops[1];
    myBackup[depth].knights[0]=knights[0];
    myBackup[depth].knights[1]=knights[1];
    myBackup[depth].queens[0]=queens[0];
    myBackup[depth].queens[1]=queens[1];
    myBackup[depth].rooks[0]=rooks[0];
    myBackup[depth].rooks[1]=rooks[1];
    myBackup[depth].targetmap=targetmap;
    myBackup[depth].ep=ep;
    myBackup[depth].castlingRules=castlingRules;
    myBackup[depth].onMove=onMove;
}

/***********************************
 * Lädt eine gespeicherte Position. *
 ***********************************/
void Position::undoMove(int depth) {
    aw=myBackup[depth].aw;
    ab=myBackup[depth].ab;
    all=myBackup[depth].all;
    kings[0]=myBackup[depth].kings[0];
    kings[1]=myBackup[depth].kings[1];
    pawns[0]=myBackup[depth].pawns[0];
    pawns[1]=myBackup[depth].pawns[1];
    bishops[0]=myBackup[depth].bishops[0];
    bishops[1]=myBackup[depth].bishops[1];
    knights[0]=myBackup[depth].knights[0];
    knights[1]=myBackup[depth].knights[1];
    queens[0]=myBackup[depth].queens[0];
    queens[1]=myBackup[depth].queens[1];
    rooks[0]=myBackup[depth].rooks[0];
    rooks[1]=myBackup[depth].rooks[1];
    targetmap=myBackup[depth].targetmap;
    ep=myBackup[depth].ep;
    castlingRules=myBackup[depth].castlingRules;
    onMove=myBackup[depth].onMove;
}

/************************************************************
 * Führt einen Zug aus und updated entsprechend den Angaben *
 * des Zuges die notwendigen Bitboards.                     *
 ************************************************************/
bool Position::move_figure (const unsigned long &from, const unsigned long &to, const int &promotion) {
    bool capture=false;

    ep=0;
    
    // do castling
    // update king- and rook-position
    if ((from&kings[onMove])!=0L && (from&1152921504606846992UL)!=0L && (to&4899916394579099716UL)!=0L) {
        
        if (to==BIN[2]) rochade(from,to,BIN[0],BIN[3]);
        else if (to==BIN[6]) rochade(from,to,BIN[7],BIN[5]);
        else if (to==BIN[58]) rochade(from,to,BIN[56],BIN[59]);
        else if (to==BIN[62]) rochade(from,to,BIN[63],BIN[61]);

        // remove rochade possibilities
        castlingRules&=castlingR1[onMove];

        updateBitboardLongs();
        onMove=!onMove;
        return capture;
    }

    bool notOnMove=!onMove;
        
    if ((to & all)!=0L)
        capture=true;    
    
    // kill target figure (if there's one)
    if ((pawns[notOnMove]&to)!=0L) pawns[notOnMove]^=to;
    else if ((knights[notOnMove]&to)!=0L) knights[notOnMove]^=to;
    else if ((bishops[notOnMove]&to)!=0L) bishops[notOnMove]^=to;
    else if ((queens[notOnMove]&to)!=0L) queens[notOnMove]^=to;

    else if ((rooks[notOnMove]&to)!=0L) {
        rooks[notOnMove]^=to;
        // Castling Möglichkeit entfernen
        if (to==72057594037927936L) castlingRules&=13; // white_castling_long=false;
        else if (to==9223372036854775808UL) castlingRules&=14; // white_castling_short=false;
        else if (to==1L) castlingRules&=7; // black_castling_long=false;
        else if (to==128L) castlingRules&=11; // black_castling_short=false;                
    }
    else if ((kings[notOnMove]&to)!=0L) kings[notOnMove]^=to;
    

    
    // move figure
    if ((pawns[onMove]&from)!=0L) {
        // promote figure
        if ((to&promotionRow[onMove])!=0L) {            
            pawns[onMove] ^= from; // remove pawn
            // add promoted figure
            if      (promotion==1) queens[onMove]    |= to;
            else if (promotion==2) rooks[onMove]     |= to;
            else if (promotion==3) bishops[onMove]   |= to;
            else if (promotion==4) knights[onMove]   |= to;                

        } else {
            // ep-move
            if ((all&to)==0L) {
                if (to==from>>9) pawns[notOnMove]^=(from>>1);
                if (to==from>>7) pawns[notOnMove]^=(from<<1);
                if (to==from<<9) pawns[notOnMove]^=(from<<1);
                if (to==from<<7) pawns[notOnMove]^=(from>>1);                
            }            
            pawns[onMove]^=(from|to);
            if (from==Position::BIN[48] && to==Position::BIN[32])   ep=9;
            else if (from==Position::BIN[49] && to==Position::BIN[33])   ep=10;
            else if (from==Position::BIN[50] && to==Position::BIN[34])   ep=11;
            else if (from==Position::BIN[51] && to==Position::BIN[35])   ep=12;
            else if (from==Position::BIN[52] && to==Position::BIN[36])   ep=13;
            else if (from==Position::BIN[53] && to==Position::BIN[37])   ep=14;
            else if (from==Position::BIN[54] && to==Position::BIN[38])   ep=15;
            else if (from==Position::BIN[55] && to==Position::BIN[39])   ep=16;
            else if (from==Position::BIN[8] && to==Position::BIN[24])    ep=1;
            else if (from==Position::BIN[9] && to==Position::BIN[25])    ep=2;
            else if (from==Position::BIN[10] && to==Position::BIN[26])   ep=3;
            else if (from==Position::BIN[11] && to==Position::BIN[27])   ep=4;
            else if (from==Position::BIN[12] && to==Position::BIN[28])   ep=5;
            else if (from==Position::BIN[13] && to==Position::BIN[29])   ep=6;
            else if (from==Position::BIN[14] && to==Position::BIN[30])   ep=7;
            else if (from==Position::BIN[15] && to==Position::BIN[31])   ep=8;        
        }
    } else if ((queens[onMove]&from)!=0L)       queens[onMove]^=(from|to);
    else if ((bishops[onMove]&from)!=0L)        bishops[onMove]^=(from|to);
    else if ((knights[onMove]&from)!=0L)        knights[onMove]^=(from|to);
    else if ((kings[onMove]&from)!=0L) {
        kings[onMove]^=(from|to);
        if (from==1152921504606846976L) castlingRules&=12; // castling white entfernen
        else if (from==16L) castlingRules&=3; // castling black entfernen
    } else if ((rooks[onMove]&from)!=0L) {
        rooks[onMove]^=(from|to);
        if (from==72057594037927936L) castlingRules&=13; // white_castling_long
        else if (from==9223372036854775808UL) castlingRules&=14; // white_castling_short
        else if (from==1L) castlingRules&=7; // black_castling_long
        else if (from==128L) castlingRules&=11; // black_castling_short
    }
        
    updateBitboardLongs();
    onMove=!onMove;
    return capture;
}

/**********************************************************
 * Ausgabe-Funktion für das "print"-Command. Zeichnet den *
 * aktuellen Zustand des Schachbretts.                    *
 **********************************************************/
void Position::printBoard() {
    cout    << "  a b c d e f g h" << endl;            
    for (int k=0;k<8;k++) {
        cout << (8-k);
        string ret;
        for (int i=0;i<8;i++) {
            int position=k*8+i;
            if (figOnPos(kings[0],position))            ret.append(" K");
            else if (figOnPos(kings[1],position))       ret.append(" k");
            else if (figOnPos(queens[0],position))      ret.append(" Q");
            else if (figOnPos(queens[1],position))      ret.append(" q");
            else if (figOnPos(knights[0],position))     ret.append(" N");
            else if (figOnPos(knights[1],position))     ret.append(" n");
            else if (figOnPos(bishops[0],position))     ret.append(" B");
            else if (figOnPos(bishops[1],position))     ret.append(" b");        
            else if (figOnPos(pawns[0],position))       ret.append(" P");
            else if (figOnPos(pawns[1],position))       ret.append(" p");        
            else if (figOnPos(rooks[0],position))       ret.append(" R");        
            else if (figOnPos(rooks[1],position))       ret.append(" r");
            else                                        ret.append(" .");
        }
        cout << ret << endl;        
    }
}

/*******************************************************
 * Ausgabe-Funktion für das "info"-Command. Diverse    *
 * Angaben zur aktuellen Position: CastlingRules, etc. *
 *******************************************************/
void Position::getInfo() {
    
    cout << getFen() << endl << endl;
    
    cout << "en passant: " << ep << endl << endl;
    
    cout << "white castling short: " << getBoolString(castlingRules&1) << endl;
    cout << "white castling long:  " << getBoolString(castlingRules&2) << endl;
    cout << "black castling short: " << getBoolString(castlingRules&4) << endl;
    cout << "black castling long:  " << getBoolString(castlingRules&8) << endl;    
    
    cout << endl;
    
    cout << "on Move: ";
    if (onMove)
        cout << "black";
    else
        cout << "white";
    
    cout << endl;
}

/***********************************************
 * Eine Rochade durchführen, dh. Bitboards von *
 * König und Turm entsprechend updaten.        *
 ***********************************************/
inline void Position::rochade(const unsigned long &old_king_pos, const unsigned long &new_king_pos, const unsigned long &old_rook_pos, const unsigned long &new_rook_pos) {
    kings[onMove]^=(old_king_pos|new_king_pos);
    rooks[onMove]^=(old_rook_pos|new_rook_pos);
}

/****************************************************
 * Updated die 3 Bitboards mit allen Figuren (all), *
 * weißen Figuren (aw) und schwarzen Figuren (ab).  *
 ****************************************************/
inline void Position::updateBitboardLongs() {
    aw=kings[0] | queens[0] | rooks[0] | bishops[0] | knights[0] | pawns[0];
    ab=kings[1] | queens[1] | rooks[1] | bishops[1] | knights[1] | pawns[1];
    all=ab|aw;    
}

/*******************************************************************
 * Hauptfunktion zur Generierung aller möglichen Züge.             *
 * Dabei wird nicht überprüft, welche Züge auch tatsächlich gültig *
 * sind, da diese Berechnung viel Zeit in Anspruch nimmt und diese *
 * Beachtung bei der Zugsuche nicht immer notwendig ist.           *
 *******************************************************************/
void Position::calc_moves(const bool &check) {
    
    for (unsigned int i=0;i<256;i++)
        mvP[cd*256+i]=0;
    
    mvSize[cd]=0;
    
    unsigned long bitboard;    
    unsigned long myFigures;
    unsigned long opFigures;
    
    if (onMove) {
        // all figures bitboard from moving / opponent color
        myFigures=ab;
        opFigures=aw;
        // pawns
        bitboard = pawns[onMove];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            addPawnSingles_black(x);
            bitboard-=Position::BIN[x];
        }        
    } else {
        // all figures bitboard from moving / opponent color
        myFigures=aw;
        opFigures=ab;        
        // pawns
        bitboard = pawns[onMove];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            addPawnSingles_white(x);
            bitboard-=Position::BIN[x];
        }
    }
    
    // rooks
    bitboard = rooks[onMove];        
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves__(x,Position::TARGET_Nx,Position::TARGET_Nx_FROM[x],Position::TARGET_Nx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Ex,Position::TARGET_Ex_FROM[x],Position::TARGET_Ex_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Sx,Position::TARGET_Sx_FROM[x],Position::TARGET_Sx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Wx,Position::TARGET_Wx_FROM[x],Position::TARGET_Wx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }        

    // bishops
    bitboard = bishops[onMove];        
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves__(x,Position::TARGET_NEx,Position::TARGET_NEx_FROM[x],Position::TARGET_NEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SEx,Position::TARGET_SEx_FROM[x],Position::TARGET_SEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SWx,Position::TARGET_SWx_FROM[x],Position::TARGET_SWx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_NWx,Position::TARGET_NWx_FROM[x],Position::TARGET_NWx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }        

    // queens
    bitboard = queens[onMove];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        split_moves__(x,Position::TARGET_NEx,Position::TARGET_NEx_FROM[x],Position::TARGET_NEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SEx,Position::TARGET_SEx_FROM[x],Position::TARGET_SEx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_SWx,Position::TARGET_SWx_FROM[x],Position::TARGET_SWx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_NWx,Position::TARGET_NWx_FROM[x],Position::TARGET_NWx_TO[x],opFigures);                        
        split_moves__(x,Position::TARGET_Nx,Position::TARGET_Nx_FROM[x],Position::TARGET_Nx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Ex,Position::TARGET_Ex_FROM[x],Position::TARGET_Ex_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Sx,Position::TARGET_Sx_FROM[x],Position::TARGET_Sx_TO[x],opFigures);
        split_moves__(x,Position::TARGET_Wx,Position::TARGET_Wx_FROM[x],Position::TARGET_Wx_TO[x],opFigures);
        bitboard-=Position::BIN[x];
    }         
    
    // knights
    bitboard = knights[onMove];
    while(bitboard != 0L) {
        int x=bitScanForward(bitboard);
        split_moves_direct(x, Position::KN_TARGET3x, Position::KN_TARGET3x_FROM[x], Position::KN_TARGET3x_TO[x], myFigures);
        bitboard-=Position::BIN[x];
    }       
    
    // kings
    int x = bitScanForward(kings[onMove]);
    split_moves_direct(x, Position::K_TARGET2x, Position::K_TARGET2x_FROM[x], Position::K_TARGET2x_TO[x], myFigures);
    if (check) {

        // white short, K
        if (!onMove
                && (castlingRules&1) // castling allowed
                && (all & 6917529027641081856L) == 0L // fields are free between king and rook
                // && opponent_target_bitboard2(8070450532247928832L)==false // check if opponent targets an involved field
                && opponent_target_bitboard2(1152921504606846976)==false
                && opponent_target_bitboard2(2305843009213693952)==false
                && opponent_target_bitboard2(4611686018427387904)==false
                && (pawns[1] & 69805794224242688L) == 0L) { // check if there's a pawn promotion move targeting an involved field

                mvD[cd][mvSize[cd]][0]=kings[onMove];
                mvD[cd][mvSize[cd]][1]=4611686018427387904L;
                mvSize[cd]++;
        }

        // white long, Q
        if (!onMove
                && (castlingRules&2)
                && (all & 1008806316530991104L) == 0L 
                // && opponent_target_bitboard2(2017612633061982208L)==false 
                && opponent_target_bitboard2(288230376151711744)==false 
                && opponent_target_bitboard2(576460752303423488)==false 
                && opponent_target_bitboard2(1152921504606846976)==false 
                && (pawns[1] & 17451448556060672L) == 0L) {

                mvD[cd][mvSize[cd]][0]=kings[onMove];
                mvD[cd][mvSize[cd]][1]=288230376151711744L;
                mvSize[cd]++;
        }

        // black short, k
        if (onMove
                && (castlingRules&4)
                && (all & 96L) == 0L 
                // && opponent_target_bitboard2(112L)==false 
                && opponent_target_bitboard2(16)==false 
                && opponent_target_bitboard2(32)==false 
                && opponent_target_bitboard2(64)==false 
                && (pawns[0] & 63488L) == 0L) {

                mvD[cd][mvSize[cd]][0]=kings[onMove];
                mvD[cd][mvSize[cd]][1]=64L;
                mvSize[cd]++;
        }

        // black long, q
        if (onMove
                && (castlingRules&8)
                && (all & 14L) == 0L 
                // && opponent_target_bitboard2(28L)==false 
                && opponent_target_bitboard2(4)==false 
                && opponent_target_bitboard2(8)==false 
                && opponent_target_bitboard2(16)==false 
                && (pawns[0] & 15872L) == 0L) {

                mvD[cd][mvSize[cd]][0]=kings[onMove];
                mvD[cd][mvSize[cd]][1]=4L;
                mvSize[cd]++;            
        }        
        
    }

    // update targetmap
    for (int i=0;i<mvSize[cd];i++)
        targetmap=targetmap|mvD[cd][i][1];
}

/****************************************************************
 * Generiert Züge für König und Springer aus aktueller Position *
 * und vorgefertigter Bitboards.                                *
 ****************************************************************/
inline void Position::split_moves_direct(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &myFigures) {
    for (unsigned int i=from;i<to;i++) {
        if ((target[i] & myFigures) == 0L) { // don't fight my color figures
            mvD[cd][mvSize[cd]][0]=Position::BIN[z];
            mvD[cd][mvSize[cd]][1]=target[i];
            mvSize[cd]++;
        }
    }
}

/*******************************************************************
 * Generiert Züge für Läufer, Turm und Dame aus aktueller Position *
 * und vorgefertigter Bitboards.                                   *
 *******************************************************************/
inline void Position::split_moves__(const int &z, const unsigned long target[], const unsigned int &from, const unsigned int &to, const unsigned long &op) {
    for (unsigned int i=from;i<to;i++) {
        if ((target[i] & all) == 0L) {
            mvD[cd][mvSize[cd]][0]=Position::BIN[z];
            mvD[cd][mvSize[cd]][1]=target[i];
            mvSize[cd]++;                
        } else {
            if ((target[i] & op) != 0L) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[z];
                mvD[cd][mvSize[cd]][1]=target[i];
                mvSize[cd]++;                    
            }
            break;
        }
    }
}

/**********************************************************
 * Die Züge der schwarzen Bauern zur Zugliste hinzufügen. *
 **********************************************************/
inline void Position::addPawnSingles_black(const int &z) {
    unsigned long startpos_black=65280L;
    unsigned long bitboardpos=Position::BIN[z];

    // usual pawn moves
    if (    (Position::TARGET_BLACK_V1[z] & all) == 0UL) {// one step forward: just go forward if target field is free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V1[z];
                mvSize[cd]++;
    }
    if (    (Position::TARGET_BLACK_V3[z] & aw) != 0UL ) { // fight -> left forward OR right forward: just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V3[z];
                mvSize[cd]++;        
    }
    if (    (Position::TARGET_BLACK_V4[z] & aw) != 0L ) { // fight -> left forward OR right forward: just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V4[z];
                mvSize[cd]++;
    }

    unsigned short int depthIndex=cd<<8;
    // search for promotions and add moves
    int size=mvSize[cd];
    for (int i=0;i<size;i++) {
        if (mvD[cd][i][0]==Position::BIN[z] && (mvD[cd][i][1] & 18374686479671623680UL) != 0L) {
            mvP[depthIndex+i]=1;
            long froml=mvD[cd][i][0];
            long tol=mvD[cd][i][1];
            
            mvD[cd][mvSize[cd]][0]=froml;
            mvD[cd][mvSize[cd]][1]=tol;
            mvP[depthIndex+mvSize[cd]]=2;
            mvSize[cd]++;

            mvD[cd][mvSize[cd]][0]=froml;
            mvD[cd][mvSize[cd]][1]=tol;
            mvP[depthIndex+mvSize[cd]]=3;
            mvSize[cd]++;

            mvD[cd][mvSize[cd]][0]=froml;
            mvD[cd][mvSize[cd]][1]=tol;
            mvP[depthIndex+mvSize[cd]]=4;
            mvSize[cd]++;
        }
    }
      
    if (    (bitboardpos & startpos_black) != 0L && (Position::TARGET_BLACK_V2[z] & all) == 0L && (Position::TARGET_BLACK_V1[z] & all) == 0L) { // two steps forward: both fields forward have to be free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=Position::TARGET_BLACK_V2[z];
                mvSize[cd]++;        
    }    
    
    // en passant moves
    if (ep>8 && ep<=16) { // possible ep for black
        if (ep==9 && bitboardpos==Position::BIN[33]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[33];
                mvD[cd][mvSize[cd]][1]=Position::BIN[40];
                mvSize[cd]++;
        }
        else if (ep==10 && bitboardpos==Position::BIN[32]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[32];
                mvD[cd][mvSize[cd]][1]=Position::BIN[41];
                mvSize[cd]++;            
        }
        else if (ep==10 && bitboardpos==Position::BIN[34]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[34];
                mvD[cd][mvSize[cd]][1]=Position::BIN[41];
                mvSize[cd]++;            
        }
        else if (ep==11 && bitboardpos==Position::BIN[33]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[33];
                mvD[cd][mvSize[cd]][1]=Position::BIN[42];
                mvSize[cd]++;            
        }
        else if (ep==11 && bitboardpos==Position::BIN[35]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[35];
                mvD[cd][mvSize[cd]][1]=Position::BIN[42];
                mvSize[cd]++;            
        }
        else if (ep==12 && bitboardpos==Position::BIN[34]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[34];
                mvD[cd][mvSize[cd]][1]=Position::BIN[43];
                mvSize[cd]++;            
        }
        else if (ep==12 && bitboardpos==Position::BIN[36]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[36];
                mvD[cd][mvSize[cd]][1]=Position::BIN[43];
                mvSize[cd]++;            
        }
        else if (ep==13 && bitboardpos==Position::BIN[35]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[35];
                mvD[cd][mvSize[cd]][1]=Position::BIN[44];
                mvSize[cd]++;            
        }
        else if (ep==13 && bitboardpos==Position::BIN[37]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[37];
                mvD[cd][mvSize[cd]][1]=Position::BIN[44];
                mvSize[cd]++;            
        }
        else if (ep==14 && bitboardpos==Position::BIN[36]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[36];
                mvD[cd][mvSize[cd]][1]=Position::BIN[45];
                mvSize[cd]++;            
        }
        else if (ep==14 && bitboardpos==Position::BIN[38]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[38];
                mvD[cd][mvSize[cd]][1]=Position::BIN[45];
                mvSize[cd]++;
        }
        else if (ep==15 && bitboardpos==Position::BIN[37]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[37];
                mvD[cd][mvSize[cd]][1]=Position::BIN[46];
                mvSize[cd]++;            
        }
        else if (ep==15 && bitboardpos==Position::BIN[39]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[39];
                mvD[cd][mvSize[cd]][1]=Position::BIN[46];
                mvSize[cd]++;            
        }
        else if (ep==16 && bitboardpos==Position::BIN[38]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[38];
                mvD[cd][mvSize[cd]][1]=Position::BIN[47];
                mvSize[cd]++;
        }
    }    
    
}

/*******************************************************
 * Die Züge der weißen Bauern zur Zugliste hinzufügen. *
 *******************************************************/
inline void Position::addPawnSingles_white(const int &z) {
    unsigned long startpos_white=71776119061217280L;
    unsigned long bitboardpos=Position::BIN[z];

    // usual pawn moves
    if (    (TARGET_WHITE_V1[z] & all) == 0L) { // just go forward if target field is free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V1[z];
                mvSize[cd]++;
    }
    if (    (TARGET_WHITE_V3[z] & ab) != 0L ) { // just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V3[z];
                mvSize[cd]++;
    }
    if (    (TARGET_WHITE_V4[z] & ab) != 0L ) {// just fight if there's an opponent on the target field
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V4[z];
                mvSize[cd]++;
    }

    
    unsigned short int depthIndex=cd<<8;
    // search for promotions and add moves
    int size=mvSize[cd];
    for (int i=0;i<size;i++) {
        if (mvD[cd][i][0]==Position::BIN[z] && (mvD[cd][i][1] & 255UL) != 0L) {
            
            // short unsigned int *cntMoves=&mvSize[cd];
            
            mvP[depthIndex+i]=1;
            long froml=mvD[cd][i][0];
            long tol=mvD[cd][i][1];
            
            mvD[cd][mvSize[cd]][0]=froml;
            mvD[cd][mvSize[cd]][1]=tol;
            mvP[depthIndex+mvSize[cd]]=2;
            mvSize[cd]++;

            mvD[cd][mvSize[cd]][0]=froml;
            mvD[cd][mvSize[cd]][1]=tol;
            mvP[depthIndex+mvSize[cd]]=3;
            mvSize[cd]++;

            mvD[cd][mvSize[cd]][0]=froml;
            mvD[cd][mvSize[cd]][1]=tol;
            mvP[depthIndex+mvSize[cd]]=4;
            mvSize[cd]++;
        }
    }    

    if (    (bitboardpos & startpos_white) != 0L && (TARGET_WHITE_V2[z] & all) == 0L && (TARGET_WHITE_V1[z] & all) == 0L) { // both fields forward have to be free
                mvD[cd][mvSize[cd]][0]=bitboardpos;
                mvD[cd][mvSize[cd]][1]=TARGET_WHITE_V2[z];
                mvSize[cd]++;
    }    
    
    // en passant moves
    if (ep>0 && ep<=8) { // possible ep for white
        if (ep==1 && bitboardpos==Position::BIN[25]  ) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[25];
                mvD[cd][mvSize[cd]][1]=Position::BIN[16];
                mvSize[cd]++;            
        }
        else if (ep==2 && bitboardpos==Position::BIN[24]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[24];
                mvD[cd][mvSize[cd]][1]=Position::BIN[17];
                mvSize[cd]++;                        
        }
        else if (ep==2 && bitboardpos==Position::BIN[26]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[26];
                mvD[cd][mvSize[cd]][1]=Position::BIN[17];
                mvSize[cd]++;                        
        }
        else if (ep==3 && bitboardpos==Position::BIN[25]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[25];
                mvD[cd][mvSize[cd]][1]=Position::BIN[18];
                mvSize[cd]++;                        
        }
        else if (ep==3 && bitboardpos==Position::BIN[27]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[27];
                mvD[cd][mvSize[cd]][1]=Position::BIN[18];
                mvSize[cd]++;                        
        }
        else if (ep==4 && bitboardpos==Position::BIN[26]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[26];
                mvD[cd][mvSize[cd]][1]=Position::BIN[19];
                mvSize[cd]++;                        
        }
        else if (ep==4 && bitboardpos==Position::BIN[28]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[28];
                mvD[cd][mvSize[cd]][1]=Position::BIN[19];
                mvSize[cd]++;                        
        }
        else if (ep==5 && bitboardpos==Position::BIN[27]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[27];
                mvD[cd][mvSize[cd]][1]=Position::BIN[20];
                mvSize[cd]++;                        
        }
        else if (ep==5 && bitboardpos==Position::BIN[29]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[29];
                mvD[cd][mvSize[cd]][1]=Position::BIN[20];
                mvSize[cd]++;                        
        }
        else if (ep==6 && bitboardpos==Position::BIN[28]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[28];
                mvD[cd][mvSize[cd]][1]=Position::BIN[21];
                mvSize[cd]++;                        
        }
        else if (ep==6 && bitboardpos==Position::BIN[30]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[30];
                mvD[cd][mvSize[cd]][1]=Position::BIN[21];
                mvSize[cd]++;                        
        }
        else if (ep==7 && bitboardpos==Position::BIN[29]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[29];
                mvD[cd][mvSize[cd]][1]=Position::BIN[22];
                mvSize[cd]++;                        
        }
        else if (ep==7 && bitboardpos==Position::BIN[31]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[31];
                mvD[cd][mvSize[cd]][1]=Position::BIN[22];
                mvSize[cd]++;                        
        }
        else if (ep==8 && bitboardpos==Position::BIN[30]) {
                mvD[cd][mvSize[cd]][0]=Position::BIN[30];
                mvD[cd][mvSize[cd]][1]=Position::BIN[23];
                mvSize[cd]++;                        
        }
    }    
}

/*********************************************************************
 * Prüft ob eine Figur an der übergebenen Position am Schachbrett    *
 * vom Gegner bedroht wird. Primär verwendet um zu überprüfen ob der *
 * König der aktuellen Farbe im Schach steht.                        *
 *********************************************************************/
inline bool Position::opponent_target_bitboard2(const unsigned long &bitboard) {
    
    unsigned long bb=bitboard;
    
    int x=bitScanForward(bb);
    
    bool opponent = !onMove;
    
    // pawns
    if (PAWN_ATTACKS[onMove][x] & pawns[opponent])
        return true;
    
    // knights (direkt)
    for (unsigned int i=KN_TARGET3x_FROM[x];i<KN_TARGET3x_TO[x];i++) {
        if (KN_TARGET3x[i] & knights[opponent])
            return true;
    }
    


    // north (linie)
    for (unsigned int i=TARGET_Nx_FROM[x];i<TARGET_Nx_TO[x];i++) {
        if (TARGET_Nx[i] & all) {
            if (TARGET_Nx[i] & rooks[opponent])
                return true;
            if (TARGET_Nx[i] & queens[opponent])
                return true;
            break;
        }
    }
    
    // east
    for (unsigned int i=TARGET_Ex_FROM[x];i<TARGET_Ex_TO[x];i++) {
        if (TARGET_Ex[i] & all) {
            if (TARGET_Ex[i] & rooks[opponent])
                return true;
            if (TARGET_Ex[i] & queens[opponent])
                return true;
            break;
        }
    }    
    
    // south
    for (unsigned int i=TARGET_Sx_FROM[x];i<TARGET_Sx_TO[x];i++) {
        if (TARGET_Sx[i] & all) {
            if (TARGET_Sx[i] & rooks[opponent])
                return true;
            if (TARGET_Sx[i] & queens[opponent])
                return true;
            break;
        }
    }    
    
    // west
    for (unsigned int i=TARGET_Wx_FROM[x];i<TARGET_Wx_TO[x];i++) {
        if (TARGET_Wx[i] & all) {
            if (TARGET_Wx[i] & rooks[opponent])
                return true;
            if (TARGET_Wx[i] & queens[opponent])
                return true;
            break;
        }
    }    
    
    // north-east
    for (unsigned int i=TARGET_NEx_FROM[x];i<TARGET_NEx_TO[x];i++) {
        if (TARGET_NEx[i] & all) {
            if (TARGET_NEx[i] & bishops[opponent])
                return true;
            if (TARGET_NEx[i] & queens[opponent])
                return true;
            break;
        }
    }    

    // south-east
    for (unsigned int i=TARGET_SEx_FROM[x];i<TARGET_SEx_TO[x];i++) {
        if (TARGET_SEx[i] & all) {
            if (TARGET_SEx[i] & bishops[opponent])
                return true;
            if (TARGET_SEx[i] & queens[opponent])
                return true;
            break;
        }
    }        
    
    // south-west
    for (unsigned int i=TARGET_SWx_FROM[x];i<TARGET_SWx_TO[x];i++) {
        if (TARGET_SWx[i] & all) {
            if (TARGET_SWx[i] & bishops[opponent])
                return true;
            if (TARGET_SWx[i] & queens[opponent])
                return true;
            break;
        }
    }        
    
    // north-west
    for (unsigned int i=TARGET_NWx_FROM[x];i<TARGET_NWx_TO[x];i++) {
        if (TARGET_NWx[i] & all) {
            if (TARGET_NWx[i] & bishops[opponent])
                return true;
            if (TARGET_NWx[i] & queens[opponent])
                return true;
            break;
        }
    }        

    // king (direkt)
    for (unsigned int i=K_TARGET2x_FROM[x];i<K_TARGET2x_TO[x];i++) {
        if (K_TARGET2x[i] & kings[opponent])
            return true;
    }        
    
    return false;
    
/*    
    bool value=false;
    cd++;
    doBackup(cd);
    targetmap=0L;
    onMove=!onMove;
    calc_moves(false);
    onMove=!onMove;
    if ((targetmap & bitboard) != 0L) value=true;
    undoMove(cd);
    cd--;
    return value;
*/    
}

/************************************************************
 * Konvertiert die Daten der aktuell gespeicherten Position *
 * in einen FEN-String.                                     *
 ************************************************************/
string Position::getFen() {
    
    string fen;
    
    // 1. group: figures
    int spaces=0;
    for (int i=1;i<=64;i++) {
        if ((Position::BIN[i-1]&all)!=0L) {
            if (spaces>0) {
                fen+=int2String(spaces);
                spaces=0;
            }
            if ((kings[0]&Position::BIN[i-1])!=0L) fen+="K";
            else if ((queens[0]&Position::BIN[i-1])!=0L) fen+="Q";
            else if ((bishops[0]&Position::BIN[i-1])!=0L) fen+="B";
            else if ((knights[0]&Position::BIN[i-1])!=0L) fen+="N";
            else if ((rooks[0]&Position::BIN[i-1])!=0L) fen+="R";
            else if ((pawns[0]&Position::BIN[i-1])!=0L) fen+="P";
            else if ((kings[1]&Position::BIN[i-1])!=0L) fen+="k";
            else if ((queens[1]&Position::BIN[i-1])!=0L) fen+="q";
            else if ((bishops[1]&Position::BIN[i-1])!=0L) fen+="b";
            else if ((knights[1]&Position::BIN[i-1])!=0L) fen+="n";
            else if ((rooks[1]&Position::BIN[i-1])!=0L) fen+="r";
            else if ((pawns[1]&Position::BIN[i-1])!=0L) fen+="p";
        } else
            spaces++;

        if (i%8==0 && i<64) {
            if (spaces>0) {
                fen+=int2String(spaces);
                spaces=0;
            }
            fen+="/";
        }
        if (i==64 && spaces>0) {
                fen+=int2String(spaces);
        }
            
    }

    // 2. group: color on move
    if (onMove)
        fen+=" b";
    else
        fen+=" w";

    // 3. group: castles
    if (castlingRules==0)
        fen+=" -";
    else {
        fen+=" ";
        if (castlingRules&1)   fen+="K";
        if (castlingRules&2)    fen+="Q";
        if (castlingRules&4)   fen+="k";
        if (castlingRules&8)    fen+="q";        
    }

    // 4. group: en passant
    if (ep==0)  fen+=" -";
    else if(ep==1)  fen+=" a6";
    else if (ep==2) fen+=" b6";
    else if (ep==3) fen+=" c6";
    else if (ep==4) fen+=" d6";
    else if (ep==5) fen+=" e6";
    else if (ep==6) fen+=" f6";
    else if (ep==7) fen+=" g6";
    else if (ep==8) fen+=" h6";
    else if (ep==9)     fen+=" a3";
    else if (ep==10)    fen+=" b3";
    else if (ep==11)    fen+=" c3";
    else if (ep==12)    fen+=" d3";
    else if (ep==13)    fen+=" e3";
    else if (ep==14)    fen+=" f3";
    else if (ep==15)    fen+=" g3";
    else if (ep==16)    fen+=" h3";

    return fen;
}

/*********************************************************
 * konvertiert den Zug mit dem übergebenen Index in eine *
 * lesbare Form, zb: a3a4, g7g8q, o.ä.                   *
 *********************************************************/
string Position::getMoveString(int i) {
    string val;

    val+=get_hum(bitScanForward(mvD[cd][i][0]));
    val+=get_hum(bitScanForward(mvD[cd][i][1]));
    
    if (mvP[cd*256+i]==1)
        val+="q";
    else if (mvP[cd*256+i]==2)
        val+="r";
    else if (mvP[cd*256+i]==3)
        val+="b";
    else if (mvP[cd*256+i]==4)
        val+="n";    
    
    return val;
}

/********************************************************
 * konvertiert den Zug mit dem übergebenen Index in die *
 * algebraische Kurznotation. Noch unvollständig.       *
 ********************************************************/
string Position::getMoveStringAlgShort(int i) {
    string val;
    val+=get_hum(bitScanForward(mvD[cd][i][1]));
    return val;
}

/******************************************
 * Konvertiert einen Index [0-63] in eine *
 * Schachbrett-Koordinate [a-h][1-8]      *
 ******************************************/
string Position::get_hum(int k) {
    string str; // = new String();

    if (k%8==0) str+="a";
    if (k%8==1) str+="b";
    if (k%8==2) str+="c";
    if (k%8==3) str+="d";
    if (k%8==4) str+="e";
    if (k%8==5) str+="f";
    if (k%8==6) str+="g";
    if (k%8==7) str+="h";
    str+=int2String(8-k/8);
    return str;
}