/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   MCTSNode.h
 * Author: peter
 *
 * Created on 10. April 2014, 07:45
 */

#ifndef MCTSNODE_H
#define	MCTSNODE_H

#include <string>
#include <iostream>
#include <stdio.h>
#include <cstdlib>
#include <cmath>
#include "PositionMonteCarloTreeSearch.h"

using namespace std;

//! Node Class for Monte Carlo Tree Search
/*!
 * Node Class for Monte Carlo Tree Search. This class also handles
 * the basic MCTS-functions like select, expand, simulate and update
*/
class MCTSNode {
public:
    MCTSNode();
    MCTSNode(string lastMoveString, unsigned long lastMoveFrom, unsigned long lastMoveTo, int lastMovePromotion, MCTSNode *parent);
    string getMove(); // most visited child
    virtual ~MCTSNode();
    MCTSNode* select(PositionMonteCarloTreeSearch *rootPos);
    MCTSNode* expand(PositionMonteCarloTreeSearch *rootPos);
    double simulate(PositionMonteCarloTreeSearch *rootPos);
    void update(double value);
    bool hasChildren();
    bool isNotFullyExpanded();
    MCTSNode* getChild(int i);
    MCTSNode* createAndAddChild(string lastMoveString, unsigned long lastMoveFrom, unsigned long lastMoveTo, int lastMovePromotion);
private:
    // move data
    unsigned long lastMoveFrom;
    unsigned long lastMoveTo;
    int lastMovePromotion;
    string lastMoveString;
    // node data
    int cntMoves;
    int numChilds;
    double totalValue;
    double visits;
    MCTSNode *parent;
    MCTSNode *childs[256];
};

#endif	/* MCTSNODE_H */

