/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   PositionEval.cpp
 * Author: peter
 * 
 * Created on 20. Februar 2014, 18:23
 */
#include <iostream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cstdint>
#include <ctime>
#include "main.h"
#include "Timer.h"
#include "PositionEval.h"

using namespace std;

extern Timer timer;
extern unsigned short int distances[64][64];

const int PositionEval::knightAdj[9] = { -20, -16, -12, -8, -4,  0,  4,  8, 12};
const int PositionEval::rookAdj[9] =   {  15,  12,   9,  6,  3,  0, -3, -6, -9};
const int PositionEval::diagNW[64] = {
        0,  1,  2,  3,  4,  5,  6,  7,
        1,  2,  3,  4,  5,  6,  7,  8,
        2,  3,  4,  5,  6,  7,  8,  9,
        3,  4,  5,  6,  7,  8,  9, 10,
        4,  5,  6,  7,  8,  9, 10, 11,
        5,  6,  7,  8,  9, 10, 11, 12,
        6,  7,  8,  9, 10, 11, 12, 13,
        7,  8,  9, 10, 11, 12, 13, 14
    };        
const int PositionEval::diagNE[64] = {
         7,  6,  5,  4,  3,  2,  1,  0,
         8,  7,  6,  5,  4,  3,  2,  1,
         9,  8,  7,  6,  5,  4,  3,  2,
        10,  9,  8,  7,  6,  5,  4,  3,
        11, 10,  9,  8,  7,  6,  5,  4,
        12, 11, 10,  9,  8,  7,  6,  5,
        13, 12, 11, 10,  9,  8,  7,  6,
        14, 13, 12, 11, 10,  9,  8,  7
    };       
const int PositionEval::bonusDiaDistance[14] = {3, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
const int PositionEval::multiPawnPenalty[7] = {0, 0, 25, 60, 100, 100, 100};    
const unsigned long PositionEval::cols[8] = {
        72340172838076673,
        144680345676153346,
        289360691352306692,
        578721382704613384,
        1157442765409226768,
        2314885530818453536,
        4629771061636907072,
        9259542123273814144UL
    };
const unsigned long PositionEval::d4e4bb=103079215104;
const unsigned long PositionEval::d4c4bb=51539607552;
const unsigned long PositionEval::d4e3bb=17626545782784;
const unsigned long PositionEval::e4f4bb=206158430208;
const unsigned long PositionEval::e4d3bb=8864812498944;
const unsigned long PositionEval::d5e5bb=402653184;
const unsigned long PositionEval::d5c5bb=201326592;
const unsigned long PositionEval::d5e6bb=135266304;
const unsigned long PositionEval::e5f5bb=805306368;
const unsigned long PositionEval::e5d6bb=268959744;
const unsigned short int PositionEval::d4e4bonus=10; // d5e5
const unsigned short int PositionEval::d4c4bonus=5; // d5c5
const unsigned short int PositionEval::d4e3bonus=5; // d5e6
const unsigned short int PositionEval::e4f4bonus=5; // e5f5
const unsigned short int PositionEval::e4d3bonus=5; // e5d6
const int PositionEval::passedPawnsBonus[2][64] = {{
          0,   0,   0,   0,   0,   0,   0,   0,
        100, 100, 100, 100, 100, 100, 100, 100,
         80,  80,  80,  80,  80,  80,  80,  80,
         60,  60,  60,  60,  60,  60,  60,  60,
         40,  40,  40,  40,  40,  40,  40,  40,
         20,  20,  20,  20,  20,  20,  20,  20,
         20,  20,  20,  20,  20,  20,  20,  20,
          0,   0,   0,   0,   0,   0,   0,   0
    },{
          0,   0,   0,   0,   0,   0,   0,   0,
         20,  20,  20,  20,  20,  20,  20,  20,
         20,  20,  20,  20,  20,  20,  20,  20,                  
         40,  40,  40,  40,  40,  40,  40,  40,
         60,  60,  60,  60,  60,  60,  60,  60,
         80,  80,  80,  80,  80,  80,  80,  80,
        100, 100, 100, 100, 100, 100, 100, 100,                  
          0,   0,   0,   0,   0,   0,   0,   0                  
    }};
const int PositionEval::passedPawnsProtectedBonus[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        125, 125, 125, 125, 125, 125, 125, 125,
        100, 100, 100, 100, 100, 100, 100, 100,
        75, 75, 75, 75, 75, 75, 75, 75,
        50, 50, 50, 50, 50, 50, 50, 50,
        25, 25, 25, 25, 25, 25, 25, 25,
        25, 25, 25, 25, 25, 25, 25, 25,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,
        25, 25, 25, 25, 25, 25, 25, 25,
        25, 25, 25, 25, 25, 25, 25, 25,
        50, 50, 50, 50, 50, 50, 50, 50,                
        75, 75, 75, 75, 75, 75, 75, 75,
        100, 100, 100, 100, 100, 100, 100, 100,
        125, 125, 125, 125, 125, 125, 125, 125,
        0, 0, 0, 0, 0, 0, 0, 0                
    }};
const unsigned long PositionEval::passedPawnsBitboards[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 4, 8, 16, 32, 64, 128,
        257, 514, 1028, 2056, 4112, 8224, 16448, 32896,                
        65793, 131586, 263172, 526344, 1052688, 2105376, 4210752, 8421504, 
        16843009, 33686018, 67372036, 134744072, 269488144, 538976288, 1077952576, 2155905152, 
        4311810305, 8623620610, 17247241220, 34494482440, 68988964880, 137977929760, 275955859520, 551911719040, 
        1103823438081, 2207646876162, 4415293752324, 8830587504648, 17661175009296, 35322350018592, 70644700037184, 141289400074368,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,                
        72340172838076416UL, 144680345676152832UL, 289360691352305664UL, 578721382704611328UL, 1157442765409222656UL, 2314885530818445312UL, 4629771061636890624UL, 9259542123273781248UL,
        72340172838010880UL, 144680345676021760UL, 289360691352043520UL, 578721382704087040UL, 1157442765408174080UL, 2314885530816348160UL, 4629771061632696320UL, 9259542123265392640UL,
        72340172821233664UL, 144680345642467328UL, 289360691284934656UL, 578721382569869312UL, 1157442765139738624UL, 2314885530279477248UL, 4629771060558954496UL, 9259542121117908992UL,
        72340168526266368UL, 144680337052532736UL, 289360674105065472UL, 578721348210130944UL, 1157442696420261888UL, 2314885392840523776UL, 4629770785681047552UL, 9259541571362095104UL,
        72339069014638592UL, 144678138029277184UL, 289356276058554368UL, 578712552117108736UL, 1157425104234217472UL, 2314850208468434944UL, 4629700416936869888UL, 9259400833873739776UL,
        72057594037927936UL, 144115188075855872UL, 288230376151711744UL, 576460752303423488UL, 1152921504606846976UL, 2305843009213693952UL, 4611686018427387904UL, 9223372036854775808UL,
        0, 0, 0, 0, 0, 0, 0, 0
    }};
const unsigned long PositionEval::passedPawnsBitboardsAttacks[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0,
        514,1285, 2570, 5140, 10280, 20560, 41120, 16448,
        131586, 328965, 657930, 1315860, 2631720, 5263440, 10526880, 4210752,
        33686018, 84215045, 168430090, 336860180, 673720360, 1347440720, 2694881440, 1077952576,
        8623620610, 21559051525, 43118103050, 86236206100, 172472412200, 344944824400, 689889648800, 275955859520,
        2207646876162, 5519117190405, 11038234380810, 22076468761620, 44152937523240, 88305875046480, 176611750092960, 70644700037184,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,
        144680345676152832UL, 361700864190382080UL, 723401728380764160UL, 1446803456761528320UL, 2893606913523056640UL, 5787213827046113280UL, 11574427654092226560UL, 4629771061636890624UL,
        144680345676021760UL, 361700864190054400UL, 723401728380108800UL, 1446803456760217600UL, 2893606913520435200UL, 5787213827040870400UL, 11574427654081740800UL, 4629771061632696320UL,
        144680345642467328UL, 361700864106168320UL, 723401728212336640UL, 1446803456424673280UL, 2893606912849346560UL, 5787213825698693120UL, 11574427651397386240UL, 4629771060558954496UL,
        144680337052532736UL, 361700842631331840UL, 723401685262663680UL, 1446803370525327360UL, 2893606741050654720UL, 5787213482101309440UL, 11574426964202618880UL, 4629770785681047552UL,
        144678138029277184UL, 361695345073192960UL, 723390690146385920UL, 1446781380292771840UL, 2893562760585543680UL, 5787125521171087360UL, 11574251042342174720UL, 4629700416936869888UL,
        0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0
    }};
const unsigned long PositionEval::passedPawnsProtectedBitboards[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        131584, 328960, 657920, 1315840, 2631680, 5263360, 10526720, 4210688,
        33685504, 84213760, 168427520, 336855040, 673710080, 1347420160, 2694840320, 1077936128,
        8623489024, 21558722560, 43117445120, 86234890240, 172469780480, 344939560960, 689879121920, 275951648768,
        2207613190144, 5519032975360, 11038065950720, 22076131901440, 44152263802880, 88304527605760, 176609055211520, 70643622084608,
        565148976676864, 1412872441692160, 2825744883384320, 5651489766768640, 11302979533537280, 22605959067074560, 45211918134149120, 18084767253659648,
        562949953421312, 1407374883553280, 2814749767106560, 5629499534213120, 11258999068426240, 22517998136852480, 45035996273704960, 18014398509481984,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,
        512, 1280, 2560, 5120, 10240, 20480, 40960, 16384,
        131584, 328960, 657920, 1315840, 2631680, 5263360, 10526720, 4210688,
        33685504, 84213760, 168427520, 336855040, 673710080, 1347420160, 2694840320, 1077936128,
        8623489024, 21558722560, 43117445120, 86234890240, 172469780480, 344939560960, 689879121920, 275951648768,
        2207613190144, 5519032975360, 11038065950720, 22076131901440, 44152263802880, 88304527605760, 176609055211520, 70643622084608,
        565148976676864, 1412872441692160, 2825744883384320, 5651489766768640, 11302979533537280, 22605959067074560, 45211918134149120, 18084767253659648,
        0, 0, 0, 0, 0, 0, 0, 0
    }};    
const int PositionEval::weakPawnsPenalty[2][64] = {{
          0,    0,   0,   0,    0,   0,   0,   0,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
          -8, -12,  -14, -16, -16, -14, -12, -10,
          -8, -12,  -14, -16, -16, -14, -12, -10,
           0,   0,   0,    0,   0,   0,   0,   0
    },{
           0,   0,   0,    0,   0,   0,   0,   0,
          -8, -12,  -14, -16, -16, -14, -12, -10,
          -8, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
         -10, -12,  -14, -16, -16, -14, -12, -10,
          0,    0,   0,   0,    0,   0,   0,   0,                   
    }};
const unsigned long PositionEval::weakPawnsBitboards[2][64] = {{
        0, 0, 0, 0, 0, 0, 0, 0,
        144680345676152832UL, 361700864190382080UL, 723401728380764160UL, 1446803456761528320UL, 2893606913523056640UL, 5787213827046113280UL, 11574427654092226560UL, 4629771061636890624UL,
        144680345676021760UL, 361700864190054400UL, 723401728380108800UL, 1446803456760217600UL, 2893606913520435200UL, 5787213827040870400UL, 11574427654081740800UL, 4629771061632696320UL,
        144680345642467328UL, 361700864106168320UL, 723401728212336640UL, 1446803456424673280UL, 2893606912849346560UL, 5787213825698693120UL, 11574427651397386240UL, 4629771060558954496UL,
        144680337052532736UL, 361700842631331840UL, 723401685262663680UL, 1446803370525327360UL, 2893606741050654720UL, 5787213482101309440UL, 11574426964202618880UL, 4629770785681047552UL,
        144678138029277184UL, 361695345073192960UL, 723390690146385920UL, 1446781380292771840UL, 2893562760585543680UL, 5787125521171087360UL, 11574251042342174720UL, 4629700416936869888UL,
        0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0
    },{
        0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0,
        514,1285, 2570, 5140, 10280, 20560, 41120, 16448,
        131586, 328965, 657930, 1315860, 2631720, 5263440, 10526880, 4210752,
        33686018, 84215045, 168430090, 336860180, 673720360, 1347440720, 2694881440, 1077952576,
        8623620610, 21559051525, 43118103050, 86236206100, 172472412200, 344944824400, 689889648800, 275955859520,
        2207646876162, 5519117190405, 11038234380810, 22076468761620, 44152937523240, 88305875046480, 176611750092960, 70644700037184,
        0, 0, 0, 0, 0, 0, 0, 0
    }};
const int PositionEval::knightA8Penalty=150;
const int PositionEval::knightH8Penalty=150;
const int PositionEval::knightA7Penalty=100;
const int PositionEval::knightH7Penalty=100;
const int PositionEval::knightC3Penalty=5;
const int PositionEval::bishopA7Penalty=150;
const int PositionEval::bishopH7Penalty=150;
const int PositionEval::bishopB8Penalty=150;
const int PositionEval::bishopG8Penalty=150;
const int PositionEval::bishopA6Penalty=50;
const int PositionEval::bishopH6Penalty=50;
const int PositionEval::bishopF1Bonus=20;
const int PositionEval::bishopC1Bonus=20;
const int PositionEval::rookOpenBonus=10;
const int PositionEval::rookHalfOpenBonus=5;
const int PositionEval::shield1Bonus=10;
const int PositionEval::shield2Bonus=5;
const int PositionEval::noShieldPenalty=10;
const int PositionEval::centerPawnPenalty=24;
const int PositionEval::kingBlockRookPenalty=24;    
const int PositionEval::slaveMistakePenalty=10;
const int PositionEval::fianchettoBonus=4;
const int PositionEval::fianchettoPenalty=4;
const int PositionEval::pawnPieceSquare[2][64] = {{        
         0,  0,  0,  0,  0,  0,  0,  0,
        50, 50, 50, 50, 50, 50, 50, 50,
        10, 10, 20, 30, 30, 20, 10, 10,
         5,  5, 10, 25, 25, 10,  5,  5,
         0,  0,  0, 20, 20,  0,  0,  0,
         5, -5,-10,  0,  0,-10, -5,  5,
         5, 10, 10,-20,-20, 10, 10,  5,
         0,  0,  0,  0,  0,  0,  0,  0        
    },{
         0,  0,  0,  0,  0,  0,  0,  0,        
         5, 10, 10,-20,-20, 10, 10,  5,
         5, -5,-10,  0,  0,-10, -5,  5,         
         0,  0,  0, 20, 20,  0,  0,  0,
         5,  5, 10, 25, 25, 10,  5,  5,
        10, 10, 20, 30, 30, 20, 10, 10,         
        50, 50, 50, 50, 50, 50, 50, 50,         
         0,  0,  0,  0,  0,  0,  0,  0         
    }};
const int PositionEval::knightPieceSquare[2][64] = {{
        -50,-40,-30,-30,-30,-30,-40,-50,
        -40,-20,  0,  0,  0,  0,-20,-40,
        -30,  0, 10, 15, 15, 10,  0,-30,
        -30,  5, 15, 20, 20, 15,  5,-30,
        -30,  0, 15, 20, 20, 15,  0,-30,
        -30,  5, 10, 15, 15, 10,  5,-30,
        -40,-20,  0,  5,  5,  0,-20,-40,
        -50,-40,-30,-30,-30,-30,-40,-50        
    },{
        -50,-40,-30,-30,-30,-30,-40,-50,
        -40,-20,  0,  5,  5,  0,-20,-40,
        -30,  5, 10, 15, 15, 10,  5,-30,
        -30,  0, 15, 20, 20, 15,  0,-30,
        -30,  5, 15, 20, 20, 15,  5,-30,
        -30,  0, 10, 15, 15, 10,  0,-30,
        -40,-20,  0,  0,  0,  0,-20,-40,
        -50,-40,-30,-30,-30,-30,-40,-50        
    }};
const int PositionEval::bishopPieceSquare[2][64] = {{
        -20,-10,-10,-10,-10,-10,-10,-20,
        -10,  0,  0,  0,  0,  0,  0,-10,
        -10,  0,  5, 10, 10,  5,  0,-10,
        -10,  5,  5, 10, 10,  5,  5,-10,
        -10,  0, 10, 10, 10, 10,  0,-10,
        -10, 10, 10, 10, 10, 10, 10,-10,
        -10,  5,  0,  0,  0,  0,  5,-10,
        -20,-10,-10,-10,-10,-10,-10,-20        
    },{
        -20,-10,-10,-10,-10,-10,-10,-20,
        -10,  5,  0,  0,  0,  0,  5,-10,
        -10, 10, 10, 10, 10, 10, 10,-10,
        -10,  0, 10, 10, 10, 10,  0,-10,
        -10,  5,  5, 10, 10,  5,  5,-10,
        -10,  0,  5, 10, 10,  5,  0,-10,
        -10,  0,  0,  0,  0,  0,  0,-10,
        -20,-10,-10,-10,-10,-10,-10,-20        
    }};
const int PositionEval::rookPieceSquare[2][64] = {{
          0,  0,  0,  0,  0,  0,  0,  0,
          5, 10, 10, 10, 10, 10, 10,  5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
          0,  0,  0,  5,  5,  0,  0,  0                
    },{
          0,  0,  0,  5,  5,  0,  0,  0        
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
         -5,  0,  0,  0,  0,  0,  0, -5,
          5, 10, 10, 10, 10, 10, 10,  5,
          0,  0,  0,  0,  0,  0,  0,  0,
    }}; 
const int PositionEval::queenPieceSquare[2][64] = {{    
        -20,-10,-10, -5, -5,-10,-10,-20,
        -10,  0,  0,  0,  0,  0,  0,-10,
        -10,  0,  5,  5,  5,  5,  0,-10,
         -5,  0,  5,  5,  5,  5,  0, -5,
          0,  0,  5,  5,  5,  5,  0, -5,
        -10,  5,  5,  5,  5,  5,  0,-10,
        -10,  0,  5,  0,  0,  0,  0,-10,
        -20,-10,-10, -5, -5,-10,-10,-20        
    },{
        -20,-10,-10, -5, -5,-10,-10,-20,
        -10,  0,  5,  0,  0,  0,  0,-10,
        -10,  5,  5,  5,  5,  5,  0,-10,
          0,  0,  5,  5,  5,  5,  0, -5,
         -5,  0,  5,  5,  5,  5,  0, -5,        
        -10,  0,  5,  5,  5,  5,  0,-10,
        -10,  0,  0,  0,  0,  0,  0,-10,
        -20,-10,-10, -5, -5,-10,-10,-20        
    }};
const int PositionEval::kingPieceSquare[2][64] = {{
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -20,-30,-30,-40,-40,-30,-30,-20,
        -10,-20,-20,-20,-20,-20,-20,-10,
         20, 20,  0,  0,  0,  0, 20, 20,
         20, 30, 10,  0,  0, 10, 30, 20        
    },{
         20, 30, 10,  0,  0, 10, 30, 20,
         20, 20,  0,  0,  0,  0, 20, 20,         
        -10,-20,-20,-20,-20,-20,-20,-10,
        -20,-30,-30,-40,-40,-30,-30,-20,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30,
        -30,-40,-40,-50,-50,-40,-40,-30         
    }};
const int PositionEval::kingPieceSquareEndgame[2][64] = {{
        -50,-40,-30,-20,-20,-30,-40,-50,
        -30,-20,-10,  0,  0,-10,-20,-30,
        -30,-10, 20, 30, 30, 20,-10,-30,
        -30,-10, 30, 40, 40, 30,-10,-30,
        -30,-10, 30, 40, 40, 30,-10,-30,
        -30,-10, 20, 30, 30, 20,-10,-30,
        -30,-30,  0,  0,  0,  0,-30,-30,
        -50,-30,-30,-30,-30,-30,-30,-50
    },{
        -50,-30,-30,-30,-30,-30,-30,-50,
        -30,-30,  0,  0,  0,  0,-30,-30,
        -30,-10, 20, 30, 30, 20,-10,-30,
        -30,-10, 30, 40, 40, 30,-10,-30,
        -30,-10, 30, 40, 40, 30,-10,-30,
        -30,-10, 20, 30, 30, 20,-10,-30,
        -30,-20,-10,  0,  0,-10,-20,-30,
        -50,-40,-30,-20,-20,-30,-40,-50                
    }};

int PositionEval::evaluate(bool side2move) {
    
    /******************
     * Material Score *
     * ****************/
    
    int materialScore0, materialScore1;

    materialScore0  =   100 *   popCount(pawns[0]) +
                        275 *   popCount(knights[0]) +
                        325 *   popCount(bishops[0]) +
                        465 *   popCount(rooks[0]) +
                        900 *   popCount(queens[0]) +
                        5000 *  popCount(kings[0]);    
    
    materialScore1  =   100 *   popCount(pawns[1]) +
                        275 *   popCount(knights[1]) +
                        325 *   popCount(bishops[1]) +
                        465 *   popCount(rooks[1]) +
                        900 *   popCount(queens[1]) +
                        5000 *  popCount(kings[1]);        

    /********************************************************************
     * Anpassung des Material Scores abhängig von der Anzahl der Bauern *
     ********************************************************************/

    
    materialScore0 += popCount(knights[0]) * knightAdj[popCount(pawns[0])];
    materialScore0 += popCount(rooks[0]) * rookAdj[popCount(pawns[0])];
    materialScore1 += popCount(knights[1]) * knightAdj[popCount(pawns[1])];
    materialScore1 += popCount(rooks[1]) * rookAdj[popCount(pawns[1])];
    
    /*****************************
     * Keine Bauern mehr am Feld *
     *****************************/
    
    if (popCount(pawns[0])==0)  materialScore0-=16;
    if (popCount(pawns[1])==0)  materialScore1-=16;
    
    /*********
     * Tempo *
     *********/
    
    if (side2move) materialScore1+=10;
    else           materialScore0+=10;     

    /************************************************
     * King Tropism                                 *
     * (Abstand der Figuren zum gegnerischen König) *
     ************************************************/

    unsigned long bitboard;
    
    int kingTropism0=0, kingTropism1=0;

    unsigned short int posKingWhite = bitScanForward(kings[0]);
    unsigned short int posKingBlack = bitScanForward(kings[1]);
    
    

    
    /*****************
     * weiße Figuren *
     *****************/

    bitboard = knights[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism0+=distances[x][posKingBlack];
        bitboard-=Position::BIN[x];
    }

    bitboard = rooks[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism0+=distances[x][posKingBlack]/2;
        bitboard-=Position::BIN[x];
    }    

    bitboard = queens[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism0+=distances[x][posKingBlack]*5/2;
        bitboard-=Position::BIN[x];
    }        
    
    bitboard = bishops[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism0+=distances[x][posKingBlack]/2;
        kingTropism0+=bonusDiaDistance[abs(diagNE[x]-diagNE[posKingBlack])];
        kingTropism0+=bonusDiaDistance[abs(diagNW[x]-diagNW[posKingBlack])];
        bitboard-=Position::BIN[x];
    }        
    
    /********************
     * schwarze Figuren *
     ********************/
    
    bitboard = knights[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism1+=distances[x][posKingWhite];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = rooks[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism1+=distances[x][posKingWhite]/2;
        bitboard-=Position::BIN[x];
    }    
    
    bitboard = queens[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism1+=distances[x][posKingWhite]*5/2;
        bitboard-=Position::BIN[x];
    }
    
    bitboard = bishops[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        kingTropism1+=distances[x][posKingWhite]/2;
        kingTropism1+=bonusDiaDistance[abs(diagNE[x]-diagNE[posKingWhite])];
        kingTropism1+=bonusDiaDistance[abs(diagNW[x]-diagNW[posKingWhite])];        
        bitboard-=Position::BIN[x];
    }    
    
    
    /******************
     * Pawn Structure *
     ******************/
    
    /********************************************
     * Doppelte Bauern in einer Linie (Penalty) *
     ********************************************/
    
    int pawnScore0=0, pawnScore1=0;
    
    for (unsigned short int i=0;i<8;i++) {
        pawnScore0-=multiPawnPenalty[popCount(cols[i]&pawns[0])];
        pawnScore1-=multiPawnPenalty[popCount(cols[i]&pawns[1])];
    }
    
    /*****************************
     * Bauern im Zentrum (Bonus) *
     *****************************/
    
    if ((d4e4bb&pawns[0])==d4e4bb)        pawnScore0+=d4e4bonus;
    if ((d4c4bb&pawns[0])==d4c4bb)        pawnScore0+=d4c4bonus;
    if ((d4e3bb&pawns[0])==d4e3bb)        pawnScore0+=d4e3bonus;
    if ((e4f4bb&pawns[0])==e4f4bb)        pawnScore0+=e4f4bonus;
    if ((e4d3bb&pawns[0])==e4d3bb)        pawnScore0+=e4d3bonus;
    
    if ((d5e5bb&pawns[1])==d5e5bb)        pawnScore1+=d4e4bonus;
    if ((d5c5bb&pawns[1])==d5c5bb)        pawnScore1+=d4c4bonus;
    if ((d5e6bb&pawns[1])==d5e6bb)        pawnScore1+=d4e3bonus;
    if ((e5f5bb&pawns[1])==e5f5bb)        pawnScore1+=e4f4bonus;
    if ((e5d6bb&pawns[1])==e5d6bb)        pawnScore1+=e4d3bonus;    
    
    /****************
     * Passed Pawns *
     ****************/
       
    bitboard = pawns[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if (    (passedPawnsBitboards[0][x]&pawns[0])==0 && (passedPawnsBitboards[0][x]&pawns[1])==0 // Felder vor Bauern sind frei
                && (passedPawnsBitboardsAttacks[0][x]&pawns[1])==0 // Felder vor Bauern werden nicht vom Gegner durch Bauern bedroht
                ) {
            
            if ((passedPawnsProtectedBitboards[0][x]&pawns[0])!=0)
                pawnScore0+=passedPawnsProtectedBonus[0][x];
            else
                pawnScore0+=passedPawnsBonus[0][x];
        
        }
        
        bitboard-=Position::BIN[x];
    }
    
    bitboard = pawns[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if (    (passedPawnsBitboards[1][x]&pawns[0])==0 && (passedPawnsBitboards[1][x]&pawns[1])==0 // Felder vor Bauern sind frei
                && (passedPawnsBitboardsAttacks[1][x]&pawns[0])==0 // Felder vor Bauern werden nicht vom Gegner durch Bauern bedroht                
                ) {
            
            if ((passedPawnsProtectedBitboards[1][x]&pawns[1])!=0)
                pawnScore1+=passedPawnsProtectedBonus[1][x];
            else
                pawnScore1+=passedPawnsBonus[1][x];
            
        }
        
        bitboard-=Position::BIN[x];
    }    
    
    /**************
     * Weak Pawns *
     **************/
    
    bitboard = pawns[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if (    (weakPawnsBitboards[0][x]&pawns[0])==0 )
                pawnScore0+=weakPawnsPenalty[0][x];
        
        bitboard-=Position::BIN[x];
    }    
    
    bitboard = pawns[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if (    (weakPawnsBitboards[1][x]&pawns[1])==0 )
                pawnScore1+=weakPawnsPenalty[1][x];
        
        bitboard-=Position::BIN[x];
    }    
        
    
    
    
    
    
    
    
    
    
    
    /********************
     * Piece Evaluation *
     ********************/
    
    int pieceScore0=0, pieceScore1=0;
    
    /************
     * Springer *
     ************/
    

    
    // A8
    if ((knights[0]&BIN[0])!=0) {
        if ((pawns[1]&BIN[8])!=0 || (pawns[1]&BIN[10])!=0)
            pieceScore0-=knightA8Penalty;
    }
    if ((knights[1]&BIN[56])!=0) {
        if ((pawns[0]&BIN[48])!=0 || (pawns[0]&BIN[50])!=0)
            pieceScore1-=knightA8Penalty;
    }
    
    // H8
    if ((knights[0]&BIN[7])!=0) {
        if ((pawns[1]&BIN[13])!=0 || (pawns[1]&BIN[15])!=0)
            pieceScore0-=knightH8Penalty;
    }
    if ((knights[1]&BIN[63])!=0) {
        if ((pawns[0]&BIN[55])!=0 || (pawns[0]&BIN[53])!=0)
            pieceScore1-=knightH8Penalty;
    }
    
    // A7
    if ((knights[0]&BIN[8])!=0) {
        if ((pawns[1]&BIN[16])!=0 && (pawns[1]&BIN[9])!=0)
            pieceScore0-=knightA7Penalty;
    }
    if ((knights[1]&BIN[48])!=0) {
        if ((pawns[0]&BIN[40])!=0 && (pawns[0]&BIN[49])!=0)
            pieceScore1-=knightA7Penalty;
    }
    
    // H7
    if ((knights[0]&BIN[15])!=0) {
        if ((pawns[1]&BIN[23])!=0 && (pawns[1]&BIN[14])!=0)
            pieceScore0-=knightH7Penalty;
    }
    if ((knights[1]&BIN[55])!=0) {
        if ((pawns[0]&BIN[47])!=0 && (pawns[0]&BIN[54])!=0)
            pieceScore1-=knightH7Penalty;
    }
    
    // C3
    if ((knights[0]&BIN[42])!=0) {
        if ((pawns[0]&BIN[50])!=0 && (pawns[0]&BIN[35])!=0 && (pawns[0]&BIN[36])==0)
            pieceScore0-=knightC3Penalty;            
    }
    if ((knights[1]&BIN[18])!=0) {
        if ((pawns[1]&BIN[10])!=0 && (pawns[1]&BIN[27])!=0 && (pawns[1]&BIN[28])==0)
            pieceScore1-=knightC3Penalty;
    }
        
    /**********
     * Läufer *
     **********/
    


    // A7
    if ((rooks[0]&BIN[8])!=0) {
        if ((pawns[1]&BIN[17])!=0)
            pieceScore0-=bishopA7Penalty;
    }
    if ((rooks[1]&BIN[48])!=0) {
        if ((pawns[0]&BIN[41])!=0)
            pieceScore1-=bishopA7Penalty;
    }
    
    // H7
    if ((rooks[0]&BIN[15])!=0) {
        if ((pawns[1]&BIN[22])!=0)
            pieceScore0-=bishopH7Penalty;
    }
    if ((rooks[1]&BIN[55])!=0) {
        if ((pawns[0]&BIN[46])!=0)
            pieceScore1-=bishopH7Penalty;
    }
    
    // B8
    if ((rooks[0]&BIN[1])!=0) {
        if ((pawns[1]&BIN[10])!=0)
            pieceScore0-=bishopB8Penalty;
    }
    if ((rooks[1]&BIN[57])!=0) {
        if ((pawns[0]&BIN[50])!=0)
            pieceScore1-=bishopB8Penalty;
    }
    
    // G8
    if ((rooks[0]&BIN[6])!=0) {
        if ((pawns[1]&BIN[13])!=0)
            pieceScore0-=bishopG8Penalty;
    }
    if ((rooks[1]&BIN[62])!=0) {
        if ((pawns[0]&BIN[53])!=0)
            pieceScore1-=bishopG8Penalty;
    }
    
    // A6
    if ((rooks[0]&BIN[16])!=0) {
        if ((pawns[1]&BIN[25])!=0)
            pieceScore0-=bishopA6Penalty;
    }
    if ((rooks[1]&BIN[40])!=0) {
        if ((pawns[0]&BIN[33])!=0)
            pieceScore1-=bishopA6Penalty;
    }
    
    // H6
    if ((rooks[0]&BIN[23])!=0) {
        if ((pawns[1]&BIN[30])!=0)
            pieceScore0-=bishopH6Penalty;
    }
    if ((rooks[1]&BIN[47])!=0) {
        if ((pawns[0]&BIN[38])!=0)
            pieceScore1-=bishopH6Penalty;
    }
    
    // F1
    if ((rooks[0]&BIN[61])!=0) {
        if ((kings[0]&BIN[62])!=0)
            pieceScore0+=bishopF1Bonus;
    }
    if ((rooks[1]&BIN[5])!=0) {
        if ((kings[1]&BIN[6])!=0)
            pieceScore1+=bishopF1Bonus;
    }
    
    // C1
    if ((rooks[0]&BIN[58])!=0) {
        if ((kings[0]&BIN[57])!=0)
            pieceScore0+=bishopC1Bonus;
    }
    if ((rooks[1]&BIN[2])!=0) {
        if ((kings[1]&BIN[1])!=0)
            pieceScore1+=bishopC1Bonus;
    }    
        
    /********************************************
     * Turm                                     *
     * - Bonus für offene und halboffene Linien *
     ********************************************/


    
    for (unsigned short int i=0;i<8;i++) {
        if ((rooks[0]&cols[i])!=0 && (cols[i]&pawns[0])==0) {
            if ((cols[i]&pawns[1])==0)
                pieceScore0+=rookOpenBonus;
            else
                pieceScore0+=rookHalfOpenBonus;
        }
    }

    for (unsigned short int i=0;i<8;i++) {
        if ((rooks[1]&cols[i])!=0 && (cols[i]&pawns[1])==0) {
            if ((cols[i]&pawns[0])==0)
                pieceScore1+=rookOpenBonus;
            else
                pieceScore1+=rookHalfOpenBonus;
        }
    }    

    /**************************************
     * Dame                               *
     * - Penalty für zu frühe Entwicklung *
     **************************************/

    bitboard = queens[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if ((BIN[x]&281474976710655)!=0) {
            if ((knights[0]&BIN[57])!=0)  pieceScore0-=2;
            if ((knights[0]&BIN[62])!=0)  pieceScore0-=2;
            if ((bishops[0]&BIN[58])!=0)  pieceScore0-=2;
            if ((bishops[0]&BIN[61])!=0)  pieceScore0-=2;
        }

        bitboard-=Position::BIN[x];
    }
    
    bitboard = queens[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);

        if ((BIN[x]&18446744073709486080UL)!=0) {
            if ((knights[1]&BIN[1])!=0)  pieceScore1-=2;
            if ((knights[1]&BIN[6])!=0)  pieceScore1-=2;
            if ((bishops[1]&BIN[2])!=0)  pieceScore1-=2;
            if ((bishops[1]&BIN[5])!=0)  pieceScore1-=2;
        }

        bitboard-=Position::BIN[x];
    }    
    
    
    /*********
     * König *
     *********/

    unsigned short int kCol;    
    
    /****************
     * Weißer König *
     ****************/
    
    for (kCol=0;kCol<8;kCol++) {
        if ((kings[0]&cols[kCol])!=0)
            break;
    }
    
    // king side
    if (kCol>4) {
        
        if ((pawns[0]&BIN[53])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[45])!=0)   pieceScore0+=shield2Bonus;
        
        if ((pawns[0]&BIN[54])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[46])!=0)   pieceScore0+=shield2Bonus;
        else if ((pawns[0]&cols[6])==0)  pieceScore0-=noShieldPenalty;
        
        if ((pawns[0]&BIN[55])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[47])!=0)   pieceScore0+=shield2Bonus;
        else if ((pawns[0]&cols[7])==0)  pieceScore0+=noShieldPenalty;
        
    }
    
    // queenside
    if (kCol<3) {
        
        if ((pawns[0]&BIN[48])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[40])!=0)   pieceScore0+=shield2Bonus;
        else if ((pawns[0]&cols[0])==0)  pieceScore0-=noShieldPenalty;
        
        if ((pawns[0]&BIN[49])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[41])!=0)   pieceScore0+=shield2Bonus;
        else if ((pawns[0]&cols[1])==0)  pieceScore0+=noShieldPenalty;        

        if ((pawns[0]&BIN[50])!=0)        pieceScore0+=shield1Bonus;
        else if ((pawns[0]&BIN[42])!=0)   pieceScore0+=shield2Bonus;        
        
    }
    
    /*******************
     * Schwarzer König *
     *******************/
    
    for (kCol=0;kCol<8;kCol++) {
        if ((kings[1]&cols[kCol])!=0)
            break;
    }
    
    // king side
    if (kCol>4) {
        
        if ((pawns[1]&BIN[13])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[21])!=0)   pieceScore1+=shield2Bonus;
        
        if ((pawns[1]&BIN[14])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[22])!=0)   pieceScore1+=shield2Bonus;
        else if ((pawns[1]&cols[6])==0)  pieceScore1-=noShieldPenalty;
        
        if ((pawns[1]&BIN[15])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[23])!=0)   pieceScore1+=shield2Bonus;
        else if ((pawns[1]&cols[7])==0)  pieceScore1+=noShieldPenalty;
        
    }
    
    // queenside
    if (kCol<3) {
        
        if ((pawns[1]&BIN[8])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[16])!=0)   pieceScore1+=shield2Bonus;
        else if ((pawns[1]&cols[0])==0)  pieceScore1-=noShieldPenalty;
        
        if ((pawns[1]&BIN[9])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[17])!=0)   pieceScore1+=shield2Bonus;
        else if ((pawns[1]&cols[1])==0)  pieceScore1+=noShieldPenalty;        

        if ((pawns[1]&BIN[10])!=0)        pieceScore1+=shield1Bonus;
        else if ((pawns[1]&BIN[18])!=0)   pieceScore1+=shield2Bonus;
        
    }

    /**************************
     * Diverse Pattern-Matches *
     **************************/
    
    int patternScore0=0, patternScore1=0;
    
    /**********************************
     * Blockierte Bauern in der Mitte *
     **********************************/
    
    // D2
    if ((pawns[0]&BIN[51])!=0 && (all&BIN[43])!=0)
        patternScore0-=centerPawnPenalty;
    
    // E2
    if ((pawns[0]&BIN[52])!=0 && (all&BIN[44])!=0)
        patternScore0-=centerPawnPenalty;
    
    // D7
    if ((pawns[1]&BIN[11])!=0 && (all&BIN[19])!=0)
        patternScore1-=centerPawnPenalty;
    
    // E7
    if ((pawns[1]&BIN[12])!=0 && (all&BIN[20])!=0)
        patternScore1-=centerPawnPenalty;
    
    /********************
     * Blockierte Türme *
     ********************/
    
    if(         ((kings[0]&BIN[61])!=0 || (kings[0]&BIN[62])!=0)
            &&  ((rooks[0]&BIN[63])!=0 || (rooks[0]&BIN[62])!=0))
        patternScore0-=kingBlockRookPenalty;
    
    if(         ((kings[0]&BIN[58])!=0 || (kings[0]&BIN[57])!=0)
            &&  ((rooks[0]&BIN[56])!=0 || (rooks[0]&BIN[57])!=0))
        patternScore0-=kingBlockRookPenalty;
    
    if(         ((kings[1]&BIN[5])!=0 || (kings[1]&BIN[6])!=0)
            &&  ((rooks[1]&BIN[7])!=0 || (rooks[1]&BIN[6])!=0))
        patternScore1-=kingBlockRookPenalty;
    
    if(         ((kings[1]&BIN[2])!=0 || (kings[1]&BIN[1])!=0)
            &&  ((rooks[1]&BIN[0])!=0 || (rooks[1]&BIN[1])!=0))
        patternScore1-=kingBlockRookPenalty;    
    
    /****************
     * Slave Fehler *
     ****************/
    
    if ((pawns[0]&BIN[35])!=0 && (pawns[0]&BIN[26])!=0 &&
            (pawns[1]&BIN[27])!=0 && (pawns[1]&BIN[18])!=0)
        patternScore0-=slaveMistakePenalty;
    
    if ((pawns[1]&BIN[27])!=0 && (pawns[1]&BIN[34])!=0 &&
            (pawns[0]&BIN[35])!=0 && (pawns[0]&BIN[42])!=0)
        patternScore1-=slaveMistakePenalty;
    
    /**************
     * Fianchetto *
     **************/
    
    if ((pawns[0]&BIN[46])!=0) {
        if ((bishops[0]&BIN[54])!=0)
            patternScore0+=fianchettoBonus;
        else {
            if (    (bishops[0]&BIN[45])==0 && 
                    (bishops[0]&BIN[47])==0 && 
                    (bishops[0]&BIN[63])==0)
                patternScore0-=fianchettoPenalty;
        }
    }

    if ((pawns[0]&BIN[41])!=0) {
        if ((bishops[0]&BIN[49])!=0)
            patternScore0+=fianchettoBonus;
        else {
            if (    (bishops[0]&BIN[42])==0 && 
                    (bishops[0]&BIN[40])==0 && 
                    (bishops[0]&BIN[56])==0)
                patternScore0-=fianchettoPenalty;
        }
    }
    
    if ((pawns[1]&BIN[22])!=0) {
        if ((bishops[1]&BIN[14])!=0)
            patternScore1+=fianchettoBonus;
        else {
            if (    (bishops[1]&BIN[21])==0 && 
                    (bishops[1]&BIN[23])==0 && 
                    (bishops[1]&BIN[7])==0)
                patternScore1-=fianchettoPenalty;
        }
    }
    
    if ((pawns[1]&BIN[17])!=0) {
        if ((bishops[1]&BIN[9])!=0)
            patternScore1+=fianchettoBonus;
        else {
            if (    (bishops[1]&BIN[18])==0 && 
                    (bishops[1]&BIN[16])==0 && 
                    (bishops[1]&BIN[0])==0)
                patternScore1-=fianchettoPenalty;
        }
    }    
    
    /************************************************************************
     * berechne targetmaps und mobility score                               *
     * targetmaps werden nachfolgend zb. für den capture score benötigt     *
     * zur berechnung des mobility score wird die gesamte zugliste benötigt *
     * daher werden diese beiden elemente hier zusammen berechnet           *
     * **********************************************************************/

//    long targetmap0, targetmap1;
//    int mobilityScore0 = 0, mobilityScore1 = 0;
//    int moveSize0, moveSize1;
//    
//    /***************
//     * weiß am zug *
//     * *************/
//
//    onMove=false;
////    targetmap=0L;
//    calc_moves(false); // true=with castling
//    moveSize0=mvSize[cd];
////    targetmap0=targetmap;
//    
//    for (unsigned int i=0;i<mvSize[cd];i++) {
//        // if ((mvD[cd][i][0]&kings[0])!=0L) mobilityScore0+=20;
//        // if ((mvD[cd][i][0]&queens[0])!=0L) mobilityScore0+=9;
//        if ((mvD[cd][i][0]&rooks[0])!=0L) mobilityScore0+=5;
//        else if ((mvD[cd][i][0]&bishops[0])!=0L) mobilityScore0+=3;
//        else if ((mvD[cd][i][0]&knights[0])!=0L) mobilityScore0+=3;
//        else if ((mvD[cd][i][0]&pawns[0])!=0L) mobilityScore0+=1;
//    }
//
//    /******************
//     * schwarz am zug *
//     * ****************/    
//    
//    onMove=true;
////    targetmap=0L;
//    calc_moves(false); // true=with castling
//    moveSize1=mvSize[cd];
////    targetmap1=targetmap;    
//
//    for (unsigned int i=0;i<mvSize[cd];i++) {
//        // if ((mvD[cd][i][0]&kings[1])!=0L) mobilityScore1+=20;
//        // if ((mvD[cd][i][0]&queens[1])!=0L) mobilityScore1+=9;
//        if ((mvD[cd][i][0]&rooks[1])!=0L) mobilityScore1+=5;
//        else if ((mvD[cd][i][0]&bishops[1])!=0L) mobilityScore1+=3;
//        else if ((mvD[cd][i][0]&knights[1])!=0L) mobilityScore1+=3;
//        else if ((mvD[cd][i][0]&pawns[1])!=0L) mobilityScore1+=1;
//    }    
//
//     mobilityScore0/=moveSize0;
//     mobilityScore1/=moveSize1;
        
    
    /*************************
     * Anzahl möglicher Züge *
     *************************/
    
//    int legalMoves0, legalMoves1;
//    int legalMovesWeight=70;
    
//    legalMoves0 = popCount(targetmap0);
//    legalMoves1 = popCount(targetmap1);
    
    
    /****************************************************************
     * berechne capture score                                       *
     * in abhängigkeit der figur-typen die geschlagen werden können *
     * **************************************************************/    
    
    // sollte captureScore bei idealer quiescence-suche nicht 0 sein?
    
//    int captureScore0 = 0, captureScore1 = 0;
//    int captureWeight = 80;
    
//    if ((targetmap0&kings[1])!=0)       captureScore0+=20;
//    if ((targetmap0&queens[1])!=0)      captureScore0+=9*popCount(targetmap0&queens[1]);
//    if ((targetmap0&rooks[1])!=0)       captureScore0+=5*popCount(targetmap0&rooks[1]);
//    if ((targetmap0&bishops[1])!=0)     captureScore0+=3*popCount(targetmap0&bishops[1]);
//    if ((targetmap0&knights[1])!=0)     captureScore0+=3*popCount(targetmap0&knights[1]);
//    if ((targetmap0&pawns[1])!=0)       captureScore0+=popCount(targetmap0&pawns[1]);
//
//    if ((targetmap1&kings[0])!=0)       captureScore1+=20;
//    if ((targetmap1&queens[0])!=0)      captureScore1+=9*popCount(targetmap1&queens[0]);
//    if ((targetmap1&rooks[0])!=0)       captureScore1+=5*popCount(targetmap1&rooks[0]);
//    if ((targetmap1&bishops[0])!=0)     captureScore1+=3*popCount(targetmap1&bishops[0]);
//    if ((targetmap1&knights[0])!=0)     captureScore1+=3*popCount(targetmap1&knights[0]);
//    if ((targetmap1&pawns[0])!=0)       captureScore1+=popCount(targetmap1&pawns[0]);    
//    
        
    
    
    
    /*********************************************************************
     * Piece-Square Tabellen                                             *
     * Berechnet Punkt-Gewinn/Abzug für Figuren an bestimmten Positionen *
     *********************************************************************/
    
    int pieceSquare0=0, pieceSquare1=0;
    
    /********
     * weiß *
     ********/
    
    bitboard = pawns[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=pawnPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = knights[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=knightPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = bishops[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=bishopPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = rooks[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=rookPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = queens[0];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare0+=queenPieceSquare[0][x];
        bitboard-=Position::BIN[x];
    }    

    if (isEndgame()) {
        bitboard = kings[0];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            pieceSquare0+=kingPieceSquareEndgame[0][x];
            bitboard-=Position::BIN[x];
        }
    } else {
        bitboard = kings[0];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            pieceSquare0+=kingPieceSquare[0][x];
            bitboard-=Position::BIN[x];
        }
    }
    
    /***********
     * schwarz *
     ***********/
    
    bitboard = pawns[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=pawnPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = knights[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=knightPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }
    
    bitboard = bishops[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=bishopPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }    

    bitboard = rooks[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=rookPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }    

    bitboard = queens[1];
    while(bitboard != 0UL) {
        int x=bitScanForward(bitboard);
        pieceSquare1+=queenPieceSquare[1][x];
        bitboard-=Position::BIN[x];
    }        

    if (isEndgame()) {
        bitboard = kings[1];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            pieceSquare1+=kingPieceSquareEndgame[1][x];
            bitboard-=Position::BIN[x];
        }
    } else {
        bitboard = kings[1];
        while(bitboard != 0UL) {
            int x=bitScanForward(bitboard);
            pieceSquare1+=kingPieceSquare[1][x];
            bitboard-=Position::BIN[x];
        }        
    }
    
    /*****************************************************************************************************************
     * berechne gesamt score                                                                                         *
     * die gewichtung der einzelnen werte sollte später automatisch durch ein skript ermittelt werden.               *
     * mit entsprechend vielen tournaments sollte so eine verbesserung bzw. verschlechterung der ELO erkennbar sein. *
     * die dadurch gewonnen daten können auch für weitere analysen verwendet werden um so zu ermitteln,              *
     * ob es werte gibt, die eigentlich gar nicht zwingend berechnet werden müssten, da sie ohnehin                  *
     * mit anderen werten korrelieren                                                                                *
     *****************************************************************************************************************/    
    
    int score = (materialScore0 - materialScore1) +
                (kingTropism0 - kingTropism1) +
                (patternScore0 - patternScore1) +
                (pawnScore0 - pawnScore1) +
                (pieceScore0 - pieceScore1) +
                (pieceSquare0 - pieceSquare1);

//                (mobilityScore0 - mobilityScore1) +                
                
   // cout << (legalMoves0 - legalMoves1) * legalMovesWeight << endl;
    
    /******************
     * we're ready :) *
     * ****************/        
    
    if (side2move)
        score*=-1;
        
    return score;
}