/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   PositionEval.h
 * Author: peter
 *
 * Created on 20. Februar 2014, 18:23
 */

#ifndef POSITIONEVAL_H
#define	POSITIONEVAL_H

#include "PositionEvalFast.h"

//! extends PositionEvalFast with an extended evaluation
/*!
 * This class is based on PositionEvalFast and uses the same
 * alpha-beta-search. The extended evaluation contains material-count,
 * pawn-structures, various piece-evaluations, pattern-matches,
 * piece-square-tables and king-tropism.
*/
class PositionEval : public PositionEvalFast {
public:
    string getName() {return "Eval"; };
    
protected:
    int evaluate(bool side2move);        
    
private:
    static const int knightAdj[9];
    static const int rookAdj[9];
    static const int diagNW[64];
    static const int diagNE[64];
    static const int bonusDiaDistance[14];
    static const int multiPawnPenalty[7];
    static const unsigned long cols[8];
    static const unsigned long d4e4bb;
    static const unsigned long d4c4bb;
    static const unsigned long d4e3bb;
    static const unsigned long e4f4bb;
    static const unsigned long e4d3bb;
    static const unsigned long d5e5bb;
    static const unsigned long d5c5bb;
    static const unsigned long d5e6bb;
    static const unsigned long e5f5bb;
    static const unsigned long e5d6bb;
    static const unsigned short int d4e4bonus; // d5e5
    static const unsigned short int d4c4bonus; // d5c5
    static const unsigned short int d4e3bonus; // d5e6
    static const unsigned short int e4f4bonus; // e5f5
    static const unsigned short int e4d3bonus; // e5d6
    static const int passedPawnsBonus[2][64];
    static const int passedPawnsProtectedBonus[2][64];
    static const unsigned long passedPawnsBitboards[2][64];
    static const unsigned long passedPawnsBitboardsAttacks[2][64];
    static const unsigned long passedPawnsProtectedBitboards[2][64];
    static const int weakPawnsPenalty[2][64];
    static const unsigned long weakPawnsBitboards[2][64];
    static const int knightA8Penalty;
    static const int knightH8Penalty;
    static const int knightA7Penalty;
    static const int knightH7Penalty;
    static const int knightC3Penalty;
    static const int bishopA7Penalty;
    static const int bishopH7Penalty;
    static const int bishopB8Penalty;
    static const int bishopG8Penalty;
    static const int bishopA6Penalty;
    static const int bishopH6Penalty;
    static const int bishopF1Bonus;
    static const int bishopC1Bonus;
    static const int rookOpenBonus;
    static const int rookHalfOpenBonus;
    static const int shield1Bonus;
    static const int shield2Bonus;
    static const int noShieldPenalty;
    static const int centerPawnPenalty;
    static const int kingBlockRookPenalty;    
    static const int slaveMistakePenalty;
    static const int fianchettoBonus;
    static const int fianchettoPenalty;
    static const int pawnPieceSquare[2][64];
    static const int knightPieceSquare[2][64];
    static const int bishopPieceSquare[2][64];
    static const int rookPieceSquare[2][64];
    static const int queenPieceSquare[2][64];
    static const int kingPieceSquare[2][64];
    static const int kingPieceSquareEndgame[2][64];    
};

#endif	/* POSITIONEVAL_H */