/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   PositionEvalFast.cpp
 * Author: peter
 * 
 * Created on 20. Februar 2014, 18:32
 */
#include <iostream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cstdint>
#include <ctime>
#include "main.h"
#include "Timer.h"
#include "PositionEvalFast.h"

using namespace std;

extern Timer timer;

int PositionEvalFast::evaluate(bool side2move) {
    int materialScore0, materialScore1;

    materialScore0  =   100 *   popCount(pawns[0]) +
                        275 *   popCount(knights[0]) +
                        325 *   popCount(bishops[0]) +
                        465 *   popCount(rooks[0]) +
                        900 *   popCount(queens[0]) +
                        5000 *  popCount(kings[0]);    // sollte der score hier so hoch sein (50000)? oder kommt ein overflow zustande?
    
    materialScore1  =   100 *   popCount(pawns[1]) +
                        275 *   popCount(knights[1]) +
                        325 *   popCount(bishops[1]) +
                        465 *   popCount(rooks[1]) +
                        900 *   popCount(queens[1]) +
                        5000 *  popCount(kings[1]);    
    
    int score = (materialScore0 - materialScore1);    
     
    if (side2move)
        score*=-1;        
    return score;    
}

int PositionEvalFast::negaMax(int alpha, int beta, int curr_depth, int depth) { //, bool canNull) {
    sum++;   
    cd = curr_depth;
    if (!timer.isTimeout() && (sum&1023)==0)
        timer.checkTimeout();    
    if (timer.isTimeout())
        return 0;
    if (curr_depth==depth)
        return evaluate(onMove);
    
//    moveList.setMoveAt(curr_depth,hashKey,"",0);
    
    /*********************************************************************************
     * Überprüfe ob die aktuelle Position bereits im Spiel- bzw. Suchverlauf vorkam. *
     * Ggf. wird ein Score für ein Draw bzw. ein entsprechender Contempt-Faktor      *
     * zurück gegeben.                                                               *
     *********************************************************************************/

//    if (cd>1 && moveList.isRepetition(cd)) {
//        int contemptOpening=-50;
//        int contemptMiddlegame=-25;
//        int contempt=0;
//        if (moveList.getNumHalfMoves()<20)
//            contempt=contemptOpening;
//        else if (!isEndgame())
//            contempt=contemptMiddlegame;
//        if (cd%2==1)
//            contempt=-contempt;
//        return contempt;
//    }
    
    /****************************************************************************
     * Suche in der HashTable nach der aktuellen Position und prüfe ggf. ob der *
     * gespeicherte Score verwendet werden kann.                                *
     ****************************************************************************/
    int bestmoveHash=-1;
    

    int score=0;  
 
    calc_moves();
    unsigned short int sortedIndizes[256];
    sortMoves(sortedIndizes,cd,bestmoveHash);
    
    doBackup();
    int maxWert=alpha;    
    for (int i=mvSize[cd]-1;i>=0;i--) {
        unsigned short int ret = move_figure(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4);
        if (curr_depth==0) {
            onMove=!onMove;
            bool catchKing=opponent_target_bitboard2(kings[onMove]); // TODO wollen wir das in jeder tiefe?
            onMove=!onMove;
            if (catchKing) {
                undoMove(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4,ret);
                continue;
            }
        }

        score = -negaMax(-beta,-maxWert,(curr_depth+1),depth);
        cd=curr_depth;
        undoMove(mvD[cd][sortedIndizes[i]][0],mvD[cd][sortedIndizes[i]][1],sortedIndizes[i]%4,ret);        
       
        if (score > maxWert) {
            maxWert = score;
            if (curr_depth==0)
                bestMoveIndex=sortedIndizes[i];
            if (maxWert>=beta) {
                maxWert=beta;
                break;
            }
        }
    }
    return maxWert;
}

void PositionEvalFast::go(unsigned int depth) {
    /********************************
     * Iteratives Deepening starten *
     ********************************/
    int score;
    int infinity=20000;
    sum=0;
    unsigned long sumPositions=0;
    score=negaMax(-infinity,infinity,0,1);
    string foundMove=getMoveString(bestMoveIndex);
    sumPositions+=sum;
    cout << "depth: 1, score: " << score << ", nodes: " << sum  << ", bestmove: " << foundMove << endl;
    unsigned long iterationTime=0;
    for (unsigned int currDepth=2;currDepth<=depth;currDepth++) {                
        sum=0;
        iterationTime=timer.getTime();
        score=negaMax(-infinity,infinity,0,currDepth); //,false);
        sumPositions+=sum;
        /**************************************************
         * Iteratives Deepening abbrechen, falls maximale *
         * Zeit für die Zugsuche verstrichen ist.         *
         **************************************************/
        if (timer.isTimeout() || timer.checkTimeout())
            break;                                        
        foundMove=getMoveString(bestMoveIndex);
        cout << "depth: " << currDepth << ", score: " << score << ", nodes: " << sum  << ", bestmove: " << foundMove << endl;
        /***********************************************
         * Iteratives Deepening abbrechen, falls nicht *
         * genug Zeit für eine neue Iteration.         *
         ***********************************************/
        iterationTime=timer.getTime()-iterationTime;
        if ((timer.getEndTime()-timer.getTime())<(iterationTime))
            break;
    }

    /************************************
     * Das Ergebnis der Suche ausgeben. *
     ************************************/
    timer.stopTimer();
    double seconds = timer.getStartEndDiffSeconds();
    cout << "Move: " << foundMove << endl;
    cout << "Score: " << score << endl;
    cout << "Positions: " << sumPositions << endl;
    cout << "Time: " << seconds << " seconds" << endl;
    cout << "bestmove " << foundMove << endl;
}