/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   PositionMonteCarloTreeSearch.cpp
 * Author: peter
 * 
 * Created on 20. Februar 2014, 18:32
 */
#include <iostream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cstdint>
#include <ctime>
#include "main.h"
#include "Timer.h"
#include "PositionMonteCarloTreeSearch.h"
#include "MCTSNode.h"
#include "PositionUCTSearch.h"
#include "UCTNodeSimilarityB.h"

using namespace std;

extern Timer timer;

void PositionMonteCarloTreeSearch::go(unsigned int depth) {
    (void)depth; // depth wird hier nicht benötigt
    srand(time(NULL));
    MCTSNode *rootNode = new MCTSNode();
    int iterations=0;
    while(1) {
        PositionMonteCarloTreeSearch *rootPos = new PositionMonteCarloTreeSearch(*this);
        MCTSNode *selected = rootNode->select(rootPos);
        MCTSNode *expanded = selected->expand(rootPos);
        double result = expanded->simulate(rootPos);
        expanded->update(result);
        iterations++;
        if (iterations%100==0) {
            if (timer.isTimeout() || timer.checkTimeout())
                break;
        }
        delete rootPos;
    }
    string foundMove=rootNode->getMove();
    delete rootNode;
    
    /************************************
     * Das Ergebnis der Suche ausgeben. *
     ************************************/
    timer.stopTimer();
    double seconds = timer.getStartEndDiffSeconds();
    cout << "Move: " << foundMove << endl;
    cout << "Time: " << seconds << " seconds" << endl;
    cout << "bestmove " << foundMove << endl;
}

int PositionMonteCarloTreeSearch::getMCTSChilds(string moveArray[], int moveIndex[], unsigned long moveFrom[], unsigned long moveTo[], int movePromotion[]) {
    cd=0;
    doBackup();
    calc_moves();
    int cntMoves=0;
    for (int i=0;i<mvSize[cd];i++) {
        unsigned short int ret=move_figure(mvD[cd][i][0],mvD[cd][i][1],i%4);
        onMove=!onMove;
        bool catchKing=opponent_target_bitboard2(kings[onMove]);
        onMove=!onMove;
        undoMove(mvD[cd][i][0],mvD[cd][i][1],i%4,ret);        
        if (!catchKing) {
            moveArray[cntMoves]=getMoveString(i);
            moveIndex[cntMoves]=i;
            moveFrom[cntMoves]=mvD[cd][i][0];
            moveTo[cntMoves]=mvD[cd][i][1];
            movePromotion[cntMoves]=i%4;
            cntMoves++;
        }
    }
    return cntMoves;
}

int PositionMonteCarloTreeSearch::doRandomMove() {
    cd=0;
    calc_moves();
    int cntMoves=mvSize[cd];
    if (cntMoves==0)
        return onMove==0?1:2; // nomove for white=1, nomove for black=2;
    int randomMove=rand()%cntMoves;
    move_figure(mvD[0][randomMove][0],mvD[0][randomMove][1],randomMove%4);
    return 0;
}

int PositionMonteCarloTreeSearch::doRandomMoveSimilarity() {
    cd=0;
    calc_moves();
    int cntMoves=mvSize[cd];
    if (cntMoves==0)
        return onMove==0?1:2; // nomove for white=1, nomove for black=2;
    evaluateSkips(0);
    int randomMove=rand()%cntMoves/2;
    move_figure(mvD[0][sortedIndizes[randomMove]][0],mvD[0][sortedIndizes[randomMove]][1],sortedIndizes[randomMove]%4);
    return 0;
}

int PositionMonteCarloTreeSearch::doRandomMoveSimilarityA() {
    cd=0;
    calc_moves();
    int cntMoves=mvSize[cd];
    if (cntMoves==0)
        return onMove==0?1:2; // nomove for white=1, nomove for black=2;
    evaluateSkips(1);
    int randomMove=rand()%cntMoves/2;
    move_figure(mvD[0][sortedIndizes[randomMove]][0],mvD[0][sortedIndizes[randomMove]][1],sortedIndizes[randomMove]%4);
    return 0;
}

int PositionMonteCarloTreeSearch::doRandomMoveSimilarityB() {
    cd=0;
    calc_moves();
    int cntMoves=mvSize[cd];
    if (cntMoves==0)
        return onMove==0?1:2; // nomove for white=1, nomove for black=2;
    evaluateSkips(2);
    int randomMove=rand()%cntMoves/2;
    move_figure(mvD[0][sortedIndizes[randomMove]][0],mvD[0][sortedIndizes[randomMove]][1],sortedIndizes[randomMove]%4);
    return 0;
}

void PositionMonteCarloTreeSearch::evaluateSkips(int mode) {
    bool promotionPossible=false;
    for (int i=0;i<mvSize[0];i++) {
        if (mvD[0][i][0]&getPawnsUnsignedLong(onMove)) { // is Pawn-Move
            if (mvD[0][i][1]&255UL || mvD[0][i][1]&18374686479671623680UL) { // is Promotion
                promotionPossible=true;
                break;
            }
        }
    }
    bool rochadePossible=false;
    for (int i=0;i<mvSize[0];i++) {
        if (mvD[0][i][0] & getKingUnsignedLong(onMove)) { // is King-Move
            if (onMove) {
                if (getCastlingRules()&4 || getCastlingRules()&8) { // castling erlaubt
                    if (mvD[0][i][1]==64L || mvD[0][i][1]==4L) {
                        rochadePossible=true;
                        break;
                    }
                }
            } else {
                if (getCastlingRules()&1 || getCastlingRules()&2) { // castling erlaubt
                    if (mvD[0][i][1]==288230376151711744L || mvD[0][i][1]==4611686018427387904L) {
                        rochadePossible=true;
                        break;
                    }
                }
            }
        }
    }
    bool capturePossible=false;
    for (int i=0;i<mvSize[0];i++) {
        if (mvD[0][i][1] & getAllUnsignedLong()) {
            capturePossible=true;
            break;
        }
    }
    bool isEndgame_=isEndgame();
    bool hasQueen_=hasQueen();
    bool hasBothRooks_=hasBothRooks();
    bool hasBothBishops_=hasBothBishops();
    bool hasBothKnights_=hasBothKnights();

    // stage: 0b
    // endcode booleans in an integer
    int positionHash=0;
    positionHash^=promotionPossible;
    positionHash<<=1;
    positionHash^=rochadePossible;
    positionHash<<=1;
    positionHash^=capturePossible;
    positionHash<<=1;
    positionHash^=isEndgame_;
    positionHash<<=1;
    positionHash^=hasQueen_;
    positionHash<<=1;
    positionHash^=hasBothRooks_;
    positionHash<<=1;
    positionHash^=hasBothBishops_;
    positionHash<<=1;
    positionHash^=hasBothKnights_;
    
    // stage 1: get a moveHash for every move and get a value for every child/move from the stats-array
    for (int i=0;i<mvSize[0];i++) {
        
        unsigned long moveFrom=mvD[0][i][0];
        unsigned long moveTo=mvD[0][i][1];
        
        bool isKingMove=getKingUnsignedLong(onMove)&moveFrom;
        bool isQueenMove=getQueenUnsignedLong(onMove)&moveFrom;
        bool isPawnMove=getPawnsUnsignedLong(onMove)&moveFrom;
        bool isBishopMove=getBishopsUnsignedLong(onMove)&moveFrom;
        bool isRookMove=getRooksUnsignedLong(onMove)&moveFrom;
        bool isKnightMove=getKnightsUnsignedLong(onMove)&moveFrom;
        bool isMoveCenter=moveTo&66229406269440UL;
        bool isMoveCapture=getAllUnsignedLong()&moveTo; // en-passant wird hier nicht betrachtet
//        bool isRochade=false;
//        bool isPromotion=false;
//        bool isMoveForward=false;
//        bool isMoveBackward=false;        
        
        int moveHash=0;
        moveHash^=isKingMove;
        moveHash<<=1;
        moveHash^=isQueenMove;
        moveHash<<=1;
        moveHash^=isPawnMove;
        moveHash<<=1;
        moveHash^=isBishopMove;
        moveHash<<=1;
        moveHash^=isRookMove;
        moveHash<<=1;
        moveHash^=isKnightMove;
        moveHash<<=1;
        moveHash^=isMoveCenter;
        moveHash<<=1;
        moveHash^=isMoveCapture;        

        // score from exactly move
        switch(mode) {
            case 0:
                skipSortValue[i]=PositionUCTSearch::MOVESTATS[positionHash][moveHash];
                break;
            case 1:
                skipSortValue[i]=PositionUCTSearch::MOVESTATS[positionHash][moveHash];
                for (int k=0;k<8;k++)
                    skipSortValue[i]+=PositionUCTSearch::MOVESTATS[positionHash][UCTNodeSimilarityB::similarMoves[moveHash][k]];
                break;                
            case 2:
                skipSortValue[i]=PositionUCTSearch::MOVESTATS[positionHash][moveHash];
                for (int k=0;k<8;k++)
                    skipSortValue[i]+=PositionUCTSearch::MOVESTATS[positionHash][UCTNodeSimilarityB::similarMoves[moveHash][k]];
                for (int k=0;k<28;k++)
                    skipSortValue[i]+=PositionUCTSearch::MOVESTATS[positionHash][UCTNodeSimilarityB::similarMoves2[moveHash][k]]/4;
                break;
        }
    }    
    
    sortChilds();
}

void PositionMonteCarloTreeSearch::sortChilds() {
    // quicksort(0,mvSize[0]-1);
    
    for (int i=0;i<mvSize[0];i++)
        sortedIndizes[i]=i;
    
    /* Bubblesort */
    for (int n=mvSize[0];n>1;n--) {
        for (int k=0;k<n-1;k++) {
            if (skipSortValue[sortedIndizes[k]]<skipSortValue[sortedIndizes[k+1]]) {
                int tmp=sortedIndizes[k];
                sortedIndizes[k]=sortedIndizes[k+1];
                sortedIndizes[k+1]=tmp;
            }
        }
    }    
    
}

//void PositionMonteCarloTreeSearch::quicksort(int left, int right) {
//    if (left<right) {
//        int split=quicksortSplit(left,right);
//        quicksort(left,split-1);
//        quicksort(split+1,right);
//    }
//}
//
//int PositionMonteCarloTreeSearch::quicksortSplit(int left, int right) {
//    int i=left;
//    int j=right-1;
//    UCTNodeSimilarityB *pivot=childs[right];
//    do {
//        while (skipSortValue[i] >= skipSortValue[right] && i<right)
//            i++;
//        
//        while (skipSortValue[j] <= skipSortValue[right] && j>left)
//            j--;
//        
//        if (i<j)
//            swapMoves(childs[i],childs[j]);
//    } while(i<j);    
//    if (skipSortValue[i] < skipSortValue[right])
//        swapMoves(childs[i],childs[right]);
//    return i;
//}
//
//void PositionMonteCarloTreeSearch::swapMoves(int i, int k) {
//    
//    
//}