/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   PositionPureMonteCarlo.h
 * Author: peter
 *
 * Created on 20. Februar 2014, 18:32
 */

#ifndef POSITIONPUREMONTECARLO_H
#define	POSITIONPUREMONTECARLO_H

#include <stdlib.h>
#include <time.h>
#include "Position.h"

//! Random Playouts from the Root-Position.
/*!
 * This class doesn't have an evaluation-function, but uses
 * a pure monte-carlo-method to calculate the best move
 * from a given position. Random playouts are started at the
 * root position. Move with best results is choosen for next
 * move.
*/
class PositionPureMonteCarlo : public Position {
public:
    string getName() {return "PureMonteCarlo"; };
    void go(unsigned int depth);
    static void pmcTest(const string &fen);
protected:
    int result[265][2];
    void playoutPosition(int moveIndex, int games);    
    int doRandomMove();
    bool isKingWhiteLost();
    bool isKingBlackLost();
};

#endif	/* POSITIONPUREMONTECARLO_H */