/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   PositionUCTSearch.h
 * Author: peter
 *
 * Created on 20. Februar 2014, 18:32
 */

#ifndef POSITIONUCTSEARCH_H
#define	POSITIONUCTSEARCH_H

#include <stdlib.h>
#include <time.h>
#include <fstream>
#include <string.h>
#include "PositionMonteCarloTreeSearch.h"

//! A Monte-Carlo-Tree-Search Game-Mode with UCT-Selection
/*!
 * This class doesn't have an evaluation-function, but uses
 * a monte-carlo-tree-search-method to calculate the best move
 * from a given position. Is uses a UCT-Method for selecting nodes
 * in the game-tree. This class uses the UCTNode-Class.
*/

extern string uctfile;

class PositionUCTSearch : public PositionMonteCarloTreeSearch {
public:
    PositionUCTSearch() {
        cout << "load uctfile: " << uctfile << endl;
        string line;
        ifstream myfile(uctfile);
        if (myfile.is_open()) {
            fromUCTFile=true;
            int lineNumber=0;
            while (getline(myfile,line)) {
                for (int i=0;i<256;i++) {
                    int posTo=line.find_first_of(";");
                    string sub=line.substr(0,posTo);
                    double rValue=atof(sub.c_str());
                    PositionUCTSearch::MOVESTATS[lineNumber][i]=rValue;
                    line=line.substr(posTo+1);
                }                
                lineNumber++;
            }
            cout << "uctfile read done" << endl;
            myfile.close();
        }
        else {
            cout << "Unable to open file" << endl;
            for (int i=0;i<256;i++) {
                for (int j=0;j<256;j++) {
                    PositionUCTSearch::MOVESTATS[i][j]=0;
                }
            }
        }
    }

    string getName() {
        return "UCTSearch";
    };
    
    string getNameUCTAppendix(string enginename) {
        if (fromUCTFile)
            return enginename+" "+uctfile;
        else
            return enginename;
    };
    void go(unsigned int depth);
    static double MOVESTATS[256][256];
protected:
    bool fromUCTFile=false;
};

#endif	/* POSITIONUCTEARCH_H */