/* 
 * Madlenka - Chess Engine for Linux
 * Copyright (C) 2012-2014 Peter Garscha
 * 
 * This File is part of Madlenka.
 * 
 * Madlenka is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Madlenka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Madlenka. If not, see <http://www.gnu.org/licenses/>.
 */

/* 
 * File:   Timer.h
 * Author: peter
 *
 * Created on 21. Dezember 2012, 15:26
 */

#ifndef TIMER_H
#define	TIMER_H

#include <sys/time.h>
#include <sys/timeb.h>

using namespace std;

//! Time-Management
/*!
 * This class handles the complete time-management for all game-modes.
*/
class Timer {
public:
    Timer() {};

    unsigned long getTime() {
        struct timeb timebuffer;
        ftime(&timebuffer);
        return (timebuffer.time*1000)+timebuffer.millitm;
    }
    
    void resetTimer(long timeForMove, int movestogo) {
        if (timeForMove>50)
            timeForMove-=50;
        startTime=getTime();
//        if (movestogo<10)
            movestogo+=2;
        tfmlong=timeForMove/movestogo;
        endTime=startTime+tfmlong;        
        timeout=false;        
        
//        string txt;
//        txt.append(long2string(timeForMove));
//        txt.append(", ");
//        txt.append(long2string(movestogo));
//        txt.append(", ");
//        txt.append(long2string(tfmlong));
//        
//        logger.log(txt);
    }
    
    unsigned long getEndTime() {
        return endTime;
    }
       
    void stopTimer() {
        endTime=getTime();
    }
    
    bool checkTimeout() {
        timeout=(getTime()>endTime);
        return timeout;
    }
    
    bool isTimeout() {
        return timeout;
    }
    
    double getStartEndDiffSeconds() {
        return double(endTime-startTime)/1000;
    }
    
    virtual ~Timer() {};
private:
    unsigned long startTime;
    unsigned long endTime;
    unsigned long tfmlong;
    bool timeout;
};

#endif	/* TIMER_H */